Package: FastGaSP
Type: Package
Title: Fast and Exact Computation of Gaussian Stochastic Process
Version: 0.6.4
Date: 2026-05-12
Authors@R: c(person(given="Mengyang", family="Gu", role=c("aut", "cre"),   email="mengyang@pstat.ucsb.edu"),
  person(given="Xinyi", family="Fang", role=c("aut"), 
  email="xinyifang@ucsb.edu"),
  person(given="Yizi", family="Lin", role=c("aut"), 
  email="lin768@ucsb.edu"))
Maintainer: Mengyang Gu <mengyang@pstat.ucsb.edu>
Author: Mengyang Gu [aut, cre],
  Xinyi Fang [aut],
  Yizi Lin [aut]
Description: Implements fast and exact computation of Gaussian stochastic process with the Matern kernel using forward filtering and backward smoothing algorithm. It includes efficient implementations of the inverse Kalman filter, with applications such as estimating particle interaction functions. These tools support models with or without noise. Additionally, the package offers algorithms for fast parameter estimation in latent factor models, where the factor loading matrix is orthogonal, and latent processes are modeled by Gaussian processes.  See the references: 1) Mengyang Gu and Yanxun Xu (2020), Journal of Computational and Graphical Statistics; 2) Xinyi Fang and Mengyang Gu (2024), <doi:10.48550/arXiv.2407.10089>; 3) Mengyang Gu and Weining Shen (2020), Journal of Machine Learning Research; 4) Yizi Lin, Xubo Liu, Paul Segall and Mengyang Gu (2025), <doi:10.48550/arXiv.2501.01324>.
License: GPL (>= 2)
Depends: methods
Imports: Rcpp,rstiefel
LinkingTo: Rcpp, RcppEigen
NeedsCompilation: yes
Repository: CRAN
Packaged: 2026-05-12 18:09:12 UTC; mengyanggu
RoxygenNote: 7.3.2
Date/Publication: 2026-05-13 07:30:56 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-13 10:16:05 UTC; unix
Archs: FastGaSP.so.dSYM

useDynLib(FastGaSP)

# Import
import("Rcpp", "methods")
importFrom("Rcpp", "evalCpp")
importFrom("stats","rnorm","qnorm","sd","optim","optimize","runif","var")
importFrom("rstiefel", "rustiefel")
# Exports
export(fgasp,predict.fgasp,Get_log_det_S2,Construct_W0_matern_5_2, Construct_W0_matern_3_2,Construct_W0_exp,Construct_G_matern_5_2, Construct_G_matern_3_2,Construct_G_exp,Construct_W_matern_5_2, Construct_W_matern_3_2,Construct_W_exp,log_lik,Sample_KF,Sample_KF_post,Get_C_R_K_Q,fmou,gppca,simulate_particle,trajectory_data,extract_time_window,unnormalized_Vicsek,f_Vicsek_variation,Get_Q_K,Get_R_y,IKF_CG_particle,A_times_x_particle,A_t_times_x_particle,get_consecutive_data,IKF_CG_particle_cell,IKF)
exportMethods(show,predict,fit.fmou,fit.gppca,predict.fmou,predict.gppca, fit)
exportClasses(fgasp,predictobj.fgasp,fmou,gppca,particle.data, particle.est)

##########################################################################
## example.R
## 
## FastGaSP Package
##
## This software is distributed under the terms of the GNU GENERAL
## PUBLIC LICENSE Version 3, April 2013.
##
## Copyright (C) 2018-present Mengyang Gu
##							  
##    
##########################################################################

################################################################
# Simulated examples to compare fast and slow computation by GaSP 
################################################################

library(FastGaSP)

#------------------------------------------------------------------------------
# Example 1: a simple example with noise to show fast computation algorithm 
#------------------------------------------------------------------------------

y_R<-function(x){
  cos(2*pi*x)
}

###let's test for 2000 observations
set.seed(1)
num_obs=2000
input=runif(num_obs)

output=y_R(input)+rnorm(num_obs,mean=0,sd=0.1)

##constucting the fgasp.model
fgasp.model=fgasp(input, output)

##range and noise-variance ratio (nugget) parameters 
param=c( log(1),log(.02))
## the log lik
log_lik(param,fgasp.model)
##time cost to compute the likelihood
time_cost=system.time(log_lik(param,fgasp.model))
time_cost[1]

##consider a nonparametric regression setting 
##estimate the parameter by maximum likelihood estimation

est_all<-optim(c(log(1),log(.02)),log_lik,object=fgasp.model,method="L-BFGS-B",
              control = list(fnscale=-1))  

##estimated log inverse range parameter and log nugget
est_all$par

##estimate variance 
est.var=Get_log_det_S2(est_all$par,fgasp.model@have_noise,fgasp.model@delta_x,
                          fgasp.model@output, fgasp.model@kernel_type)[[2]]/fgasp.model@num_obs
est.var

###1. Do some interpolation test 
num_test=5000
testing_input=runif(num_test) ##there are the input where you don't have observations
pred.model=predict(param=est_all$par,object=fgasp.model,testing_input=testing_input)

lb=pred.model@mean+qnorm(0.025)*sqrt(pred.model@var)
ub=pred.model@mean+qnorm(0.975)*sqrt(pred.model@var)

## calculate lb for the mean function
pred.model2=predict(param=est_all$par,object=fgasp.model,testing_input=testing_input,var_data=FALSE)
lb_mean_funct=pred.model2@mean+qnorm(0.025)*sqrt(pred.model2@var)
ub_mean_funct=pred.model2@mean+qnorm(0.975)*sqrt(pred.model2@var)

## plot the prediction
min_val=min(lb,output)
max_val=max(ub,output)

plot(pred.model@testing_input,pred.model@mean,type='l',col='blue',
     ylim=c(min_val,max_val),
     xlab='x',ylab='y')
polygon( c(pred.model@testing_input,rev(pred.model@testing_input)),
         c(lb,rev(ub)),col = "grey80", border = FALSE)
lines(pred.model@testing_input,pred.model@mean,type='l',col='blue')
lines(pred.model@testing_input,y_R(pred.model@testing_input),type='l',col='black')
lines(pred.model2@testing_input,lb_mean_funct,col='blue',lty=2)
lines(pred.model2@testing_input,ub_mean_funct,col='blue',lty=2)
lines(input,output,type='p',pch=16,col='black',cex=0.4) #one can plot data

legend("bottomleft", legend=c("predictive mean","95% predictive interval","truth"),
       col=c("blue","blue","black"), lty=c(1,2,1), cex=.8)


##mean square error for all inputs
mean((pred.model@mean- y_R(pred.model@testing_input))^2)
##coverage for the mean
length(which(y_R(pred.model@testing_input)>lb_mean_funct &
               y_R(pred.model@testing_input)<ub_mean_funct))/pred.model@num_testing
##average length of the invertal for the mean
mean(abs(ub_mean_funct-lb_mean_funct))
##average length of the invertal for the data
mean(abs(ub-lb))

#---------------------------------------------------------------------------------
# Example 2: numerical comparison with the GaSP by inverting the covariance matrix
#---------------------------------------------------------------------------------
##matern correlation with smoothness parameter being 2.5
matern_5_2_kernel<-function(d,beta){  
  res=(1+sqrt(5)*beta*d + 5*beta^2*d^2/3 )*exp(-sqrt(5)*beta*d)
  res
}

##A function for computing the likelihood by the GaSP in a straightforward way
log_lik_GaSP_slow<-function(param,have_noise=TRUE,input,output){
  n=length(output)
  beta=exp(param[1])
  eta=0
  if(have_noise){
    eta=exp(param[2])
  }
  R00=abs(outer(input,input,'-'))
  R=matern_5_2_kernel(R00,beta=beta)
  R_tilde=R+eta*diag(n)
  #use Cholesky and one backsolver and one forward solver so it is more stable
  L=t(chol(R_tilde))
  output_t_R.inv= t(backsolve(t(L),forwardsolve(L,output )))
  S_2=output_t_R.inv%*%output
  
  -sum(log(diag(L)))-n/2*log(S_2)
  ##or one can just inverting it using the following code, but it is less stable
  #R_tilde_inv=solve(R_tilde)
  #S_2=t(output)%*%R_tilde_inv%*%output
  #(-determinant(R_tilde,logarithm=T)$modulus[1]/2-length(output)/2*log(S_2))
}


pred_GaSP_slow<-function(param,have_noise=TRUE,input,output,testing_input){
  beta=exp(param[1])
  R00=abs(outer(input,input,'-'))
  eta=0
  if(have_noise){
    eta=exp(param[2])
  }
  R=matern_5_2_kernel(R00,beta=beta)
  R_tilde=R+eta*diag(length(output))
  
  ##I invert it here but one can still use cholesky to make it more stable
  R_tilde_inv=solve(R_tilde)

  r0=abs(outer(input,testing_input,'-'))
  r=matern_5_2_kernel(r0,beta=beta)

  S_2=t(output)%*%(R_tilde_inv%*%output)

  sigma_2_hat=as.numeric(S_2/num_obs)

  pred_mean=t(r)%*%(R_tilde_inv%*%output)
  pred_var=rep(0,length(testing_input))
  
  
  for(i in 1:length(testing_input)){
    pred_var[i]=-t(r[,i])%*%R_tilde_inv%*%r[,i]
  }
  pred_var=pred_var+1+eta
  list=list()
  list$mean=pred_mean
  list$var=pred_var*sigma_2_hat
  list
}

##let's test sin function
y_R<-function(x){
  sin(2*pi*x)
}


###let's test for 500 observations
set.seed(1)
num_obs=500
input=runif(num_obs)
output=y_R(input)+rnorm(num_obs,mean=0,sd=0.1)


##constucting the fgasp.model
fgasp.model=fgasp(input, output)

##range and noise-variance ratio (nugget) parameters 
param=c( log(1),log(.02))
## the log lik
log_lik(param,fgasp.model)
#log_lik_GaSP_slow(param,have_noise=TRUE,input=input,output=output)

##time cost to compute the likelihood
##More number of inputs mean larger differences
time_cost=system.time(log_lik(param,fgasp.model))
time_cost

#time_cost_slow=system.time(log_lik_GaSP_slow(param,have_noise=TRUE,input=input,output=output))
#time_cost_slow


est_all<-optim(c(log(1),log(.02)),log_lik,object=fgasp.model,method="L-BFGS-B",
               control = list(fnscale=-1))  

##estimated log inverse range parameter and log nugget
est_all$par


##Do some interpolation test for comparison
num_test=200
testing_input=runif(num_test) ##there are the input where you don't have observations


## two ways of prediction
pred.model=predict(param=est_all$par,object=fgasp.model,testing_input=testing_input)
#pred_slow=pred_GaSP_slow(param=est_all$par,have_noise=TRUE,input,output,sort(testing_input))
## look at the difference
#max(abs(pred.model@mean-pred_slow$mean))
#max(abs(pred.model@var-pred_slow$var))


#--------------------------------------------------------------
# Example 3: example that one does not have a noise in the data
#--------------------------------------------------------------
## Here is a function in the Sobolev Space with order 3
y_R<-function(x){
  j_seq=seq(1,200,1)
  record_y_R=0
  for(i_j in 1:200){
    record_y_R=record_y_R+2*j_seq[i_j]^{-2*3}*sin(j_seq[i_j])*cos(pi*(j_seq[i_j]-0.5)*x)

  }
  record_y_R
}


##generate some data without noise
num_obs=50
input=seq(0,1,1/(num_obs-1))

output=y_R(input)


##constucting the fgasp.model
fgasp.model=fgasp(input, output,have_noise=FALSE)

##range and noise-variance ratio (nugget) parameters 
param=c( log(1))
## the log lik
log_lik(param,fgasp.model)
#log_lik_GaSP_slow(param,have_noise=FALSE,input=input,output=output)



#if one does not have noise one may need to give a lower bound or use a penalty 
#(e.g. induced by a prior) to make the estimation more robust
est_all<-optimize(log_lik,interval=c(0,10),maximum=TRUE,fgasp.model)
  
##Do some interpolation test for comparison
num_test=1000
testing_input=runif(num_test) ##there are the input where you don't have observations


pred.model=predict(param=est_all$maximum,object=fgasp.model,testing_input=testing_input)



#This is the 95 posterior credible interval for the outcomes which contain the estimated 
#variance of the noise
#sometimes there are numerical instability is one does not have noise or error
lb=pred.model@mean+qnorm(0.025)*sqrt(abs(pred.model@var))
ub=pred.model@mean+qnorm(0.975)*sqrt(abs(pred.model@var))

## plot the prediction
min_val=min(lb,output)
max_val=max(ub,output)

plot(pred.model@testing_input,pred.model@mean,type='l',col='blue',
     ylim=c(min_val,max_val),
     xlab='x',ylab='y')
polygon( c(pred.model@testing_input,rev(pred.model@testing_input)),
         c(lb,rev(ub)),col = "grey80", border = FALSE)
lines(pred.model@testing_input,pred.model@mean,type='l',col='blue')
lines(pred.model@testing_input,y_R(pred.model@testing_input),type='l',col='black')
lines(input,output,type='p',pch=16,col='black')
legend("bottomleft", legend=c("predictive mean","95% predictive interval","truth"),
       col=c("blue","blue","black"), lty=c(1,2,1), cex=.8)


##mean square error for all inputs
mean((pred.model@mean- y_R(pred.model@testing_input))^2)

##########################################################################
## fgasp construction function
## 
## FastGaSP Package
##
## This software is distributed under the terms of the GNU GENERAL
## PUBLIC LICENSE Version 2, April 2013.
##
## Copyright (C) 2018-present Mengyang Gu
##							  
##    
##########################################################################


fgasp <- function(input, output, have_noise=TRUE, kernel_type='matern_5_2'){


  if( (kernel_type!='matern_5_2')&(kernel_type!='matern_3_2')&(kernel_type!='exp')){
    stop("the current version only support the Matern covariance with smoothness 
         parameter being 2.5, 1.5 or 0.5 (exponential kernel). \n")
  }
  
  ##if the input is not sorted so we will sort them
  object <- new("fgasp")
  object@num_obs=length(output)
  #object@param=param
  object@kernel_type=kernel_type
  object@have_noise=have_noise
  
  ## if the input is not sorted, then I sort it
  #if(sum(which(input[2:object@num_obs]-input[1:(object@num_obs-1)]<0 ))>0){
  sorting=sum((input[2:object@num_obs]-input[1:(object@num_obs-1)])<0)>0
  if(sorting){
      
    input_sorted_all=sort(input,index.return=TRUE)
    object@input=input_sorted_all[[1]]
    object@output=output[input_sorted_all[[2]]]
  }else{
    object@input=input
    object@output=output
  }
  object@delta_x=object@input[2:object@num_obs]-object@input[1:(object@num_obs-1)]
  if(sorting){
    object@input_sorted_ix=input_sorted_all$ix
  }else{
    object@input_sorted_ix=1:length(input)
  }
  if(length(object@delta_x)!=(length(object@output)-1)){
    stop("length(output) should be equal to length(delta_x)+1 \n")
  }
  ##the current version does not support duplicated inputs
  if(length(which(object@delta_x==0))>0){
    stop("Please delete the repeated inputs. \n")
  }
  return(object)
}

show.fgasp <- function(object) {	
  cat('Number of observations: ',object@num_obs,'\n')
  cat('Have noise: ',object@have_noise,'\n')
  cat('Type of kernel: ',object@kernel_type,'\n')
  
}


##likelihood function of GaSP using fast computation
##the Filtering algorithm is implemented using Rcpp and RcppEigen
log_lik<-function(param,object){
  param=as.vector(param)
  if(object@have_noise==T && length(param)!=2){
       stop("Please specify both the log inverse range parameter 
            and log nugget parameter if there is noise. \n")
  }
  
  log_det_S2=Get_log_det_S2(param,object@have_noise,object@delta_x,
                            object@output, object@kernel_type);
  ##log likelihood
  -log_det_S2[[1]]/2-(object@num_obs)/2*log(log_det_S2[[2]])
}

##prediction 
predict.fgasp<-function(param,object,testing_input=NULL,var_data=TRUE,sigma_2=NULL,
                        predict_input=F,index_return=F){
  param=as.vector(param)
  if(is.null(testing_input)&&predict_input==F){
    stop("Please give testing_input or let predict_input be true \n")
  }
  if(object@have_noise==T && length(param)!=2){
    stop("Please specify both the log inverse range parameter 
          and log nugget parameter. \n")
  }
  ## if sigma_2 is not given, I will estimate it
  if(is.null(sigma_2)){
  sigma_2=Get_log_det_S2(param,object@have_noise,object@delta_x,
                             object@output,object@kernel_type)[[2]]/object@num_obs
  }
  
  predictobj<- new("predictobj.fgasp")
  if(predict_input==T){
    testing_input=object@input
    delta_x_all_here=object@delta_x
    delta_x_all=object@delta_x
    index_x_all= rep(1,length(object@input))
    sorting=F
    testing_loc=rep(1,length(object@input))
  }else{
    predictobj@num_testing=length(testing_input)
    
    sorting=sum((testing_input[2:predictobj@num_testing]-testing_input[1:(predictobj@num_testing-1)])<0)>0
    
    if(sorting){
      testing_input_sorted_all=sort(testing_input,index.return=T)
      
      testing_input_sorted=testing_input_sorted_all$x
      predictobj@testing_input=testing_input_sorted
    }else{
      predictobj@testing_input=testing_input
      testing_input_sorted=testing_input
    }
    predictobj@param=param
    predictobj@var_data=var_data
    ## we don't support the repeated testing input in this version
    if(sum((testing_input_sorted[2:predictobj@num_testing]-
                    testing_input_sorted[1:(predictobj@num_testing-1)])==0)>0){
      stop("Please delete the repeated testing inputs. \n")
    }
    ##combine the input and testing input
    input_all_ori=c(object@input,testing_input_sorted)
    input_all_sort=sort(input_all_ori,index.return=TRUE)
  
    input_all=input_all_sort[[1]] ##all sorted input with and without the observations
    delta_x_all=input_all[2:length(input_all_ori)]-input_all[1:(length(input_all_ori)-1)]
    
    ##create a vector of inputs where the one mean there is an observation
    ##and zero means no observation
    index_x_all=rep(0, length(input_all))
    index_x_all[1:object@num_obs]=1
    index_x_all=index_x_all[input_all_sort[[2]]]
    index_x_all=as.integer(index_x_all)
    
    
  
    testing_loc=which(index_x_all==0)
    delta_x_all_here=delta_x_all
    
    if(length(which(delta_x_all==0))>0){
      #input_all=input_all[-which(delta_x_all==0)]
      index_x_all=index_x_all[-(which(delta_x_all==0)+1)]
      delta_x_all_here=delta_x_all[-which(delta_x_all==0)]
    }
  }

  
  KF_smoother_result=Kalman_smoother(param, object@have_noise,
                                     index_x_all, delta_x_all_here,object@output,sigma_2,object@kernel_type)
  if(length(which(delta_x_all==0))==0){
    predictobj@mean=KF_smoother_result[[1]][testing_loc]
    if(var_data==T | object@have_noise==F){
       predictobj@var=KF_smoother_result[[2]][testing_loc]
    }else{
      predictobj@var=KF_smoother_result[[2]][testing_loc]-sigma_2*exp(param[2])
    }
  }else{##enlarge it to the orginal long sequence that may contain knots. 
    res=rep(NA,length(input_all_ori))
    res[-(which(delta_x_all==0)+1)]=KF_smoother_result[[1]]
    res[which(delta_x_all==0)+1]= res[which(delta_x_all==0)]
    predictobj@mean=res[testing_loc]
    
    res[-(which(delta_x_all==0)+1)]=KF_smoother_result[[2]]
    res[which(delta_x_all==0)+1]= res[which(delta_x_all==0)]
    if(var_data==T | object@have_noise==F){
      predictobj@var=res[testing_loc]
    }else{
      predictobj@var=res[testing_loc]-sigma_2*exp(param[2])
    }
    
  }
  if(index_return==F){
    predictobj@index_x_all=NULL
    predictobj@delta_x_all=NULL
    predictobj@testing_loc=NULL
    predictobj@input_all_sorted_ix=NULL
    predictobj@testing_input_sorted_ix=NULL
  }else{
    #predictobj@index_x_all=index_x_all 
    predictobj@index_x_all=index_x_all
    predictobj@delta_x_all=delta_x_all_here 
    predictobj@testing_loc=predictobj@testing_loc
    if(predict_input==F){ ##if not predicting the index
      predictobj@input_all_sorted_ix=input_all_sort$ix
      if(sorting){
        predictobj@testing_input_sorted_ix=testing_input_sorted_all$ix
      }else{
        predictobj@testing_input_sorted_ix=1:length(testing_input)
      }
    }else{
      predictobj@input_all_sorted_ix=1:length(object@input)
      predictobj@testing_input_sorted_ix=1:length(object@input)
    }
  }
  
  return(predictobj)
  
}

###simulate
#simulate.fgasp <- function (param,object,...){
  
#}

## fmou class
fmou <- function(output, d=NULL, est_d=FALSE,est_U0=TRUE, est_sigma0_2=TRUE, 
                 U0 = NULL, sigma0_2=NULL){
  object <- new("fmou", output, d, est_d, est_U0, est_sigma0_2, 
                U0, sigma0_2)
  return(object)
}

#### fmou k-step-ahead prediction
predict_fmou <- function(param_est, step=1, interval=FALSE, interval_data=TRUE){
  n = ncol(param_est$output)
  U0 = param_est$U
  rho = param_est$rho
  z_n_hat = matrix(param_est$post_z_mean[,n], ncol=1)
  z_n_var = matrix(param_est$post_z_var[,n], ncol=1)
  sigma2 = param_est$sigma2
  sigma0_2 = param_est$sigma0_2
  
  pred_mean_obs = matrix(NA, nrow=nrow(U0), ncol=length(step))
  for(t in step){
    pred_mean_latent = diag(rho^{t}) %*% z_n_hat
    pred_mean_obs[,t] = U0 %*% pred_mean_latent
  }
  
  if(interval){
    pred_z_var = matrix(NA, nrow=length(rho), ncol=length(step))
    pre_var = z_n_var
    for(t in step){
      pred_z_var[,t] = rho^2 * pre_var + sigma2
      pre_var = pred_z_var[,t]
    }
    
    if(interval_data){
      data_var = matrix(NA, nrow=nrow(U0), ncol=length(step))
      for(t in step){
        data_var[,t] = rowSums(U0*t(matrix(pred_z_var[,t],dim(U0)[2],dim(U0)[1]))*U0) + sigma0_2
      }
      pred_interval_95lb = pred_mean_obs - 1.96 * sqrt(data_var)
      pred_interval_95ub = pred_mean_obs + 1.96 * sqrt(data_var)
    }
    else{
      mean_var = matrix(NA, nrow=nrow(U0), ncol=length(step))
      for(t in step){
        mean_var[,t] = rowSums(U0*t(matrix(pred_z_var[,t],dim(U0)[2],dim(U0)[1]))*U0)
      }
      pred_interval_95lb = pred_mean_obs - 1.96 * sqrt(mean_var)
      pred_interval_95ub = pred_mean_obs + 1.96 * sqrt(mean_var)
    }
  }
  
  res = list(pred_mean = pred_mean_obs)
  if(interval){
    res$pred_interval_95lb = pred_interval_95lb
    res$pred_interval_95ub = pred_interval_95ub
  }
  return(res)
}

## gppca class
neg_log_lik_shared_cov_FFBS<-function(param,kernel_type, output, delta_x,d){
  # compute negative likelihood when latent processes share parameters
  
  n = ncol(output)
  k = nrow(output)
  output_2=output%*%t(output)
  trace_output_2=sum(output^2)
  G_log_det_cov=Get_G_log_det_cov(param, output, delta_x,d=1,kernel_type=kernel_type)
  G=output_2-G_log_det_cov[[1]]
  
  eigen_G=eigen(G)
  
  res = -(-(sum(G_log_det_cov[[2]]))*d/2-(n*k)/2*log(trace_output_2-sum(eigen_G$values[1:d]) ))
  return(res)
}

neg_log_lik_diff_cov_FFBS<-function(param,A_ini,kernel_type='matern_5_2',output,delta_x){
  # compute negative likelihood when latent processes have distinct parameters
  
  d = ncol(A_ini)
  n = ncol(output)
  k = nrow(output)
  G_log_det_cov=Get_G_log_det_cov(param,output, delta_x,d,kernel_type)
  G=as.list(1:d)
  output_2=output%*%t(output)
  trace_output_2=sum(output^2)
  
  for(i in 1:d){
    G[[i]]=output_2-G_log_det_cov[[i]]
  }
  
  A_hat_here=Optimization_Stiefel_Manifold(A_ini, G=G,max_iter=200)
  
  S_2=trace_output_2+F_Funct(A_hat_here,G)
  neg_log_lik=-(-(sum(G_log_det_cov[[d+1]]))/2-(n*k)/2*log(S_2))
  return(neg_log_lik)
}

pred_FFBS_FastGaSP<-function(param,A_hat,input,testing_input, output_here,d,var_data=F,kernel_type){
  # predict posterior mean of observations
  
  beta=param[1]
  sigma_2=param[2]
  sigma_2_0=param[3]
  k = nrow(output_here)
  
  num_testing=length(testing_input)
  num_obs=length(input)
  
  predict_all_dlm=matrix(0, num_testing,d)
  var_all_dlm=matrix(0, num_testing,d)
  output_t_A=t(output_here)%*%A_hat
  var_est=0
  
  for(i_d in 1:d){
    
    m.model=fgasp(input,output_t_A[,i_d],have_noise=TRUE,kernel_type=kernel_type)
    m.pred=predict(param=c(log(beta),log(sigma_2_0/sigma_2)),object=m.model,
                   testing_input=testing_input,var_data=FALSE,sigma_2=sigma_2)
    predict_all_dlm[,i_d]=m.pred@mean
    
    var_all_dlm[,i_d]= m.pred@var  
    
    var_est= var_est+(A_hat[,i_d]^2)%*%t(as.matrix(var_all_dlm[,i_d],num_testing,1))
  }
  
  return.list=as.list(1:2)
  return.list[[1]]=A_hat%*%t(predict_all_dlm)
  if(var_data==F){
    return.list[[2]]=var_est
  }else{
    return.list[[2]]=var_est+rep(sigma_2_0,k)
  }
  
  return(return.list)
  
}

pred_FFBS_FastGaSP_diff_cov<-function(param,A_hat,input,testing_input, output_here,d,var_data=F,kernel_type='matern_5_2'){
  
  
  beta=param[1:d]
  sigma_2=param[(d+1):(2*d)]
  sigma_2_0=param[2*d+1]
  num_testing=length(testing_input)
  num_obs=length(input)
  k = nrow(output_here)
  
  predict_all_dlm=matrix(0, num_testing,d)
  var_all_dlm=matrix(0, num_testing,d)
  output_t_A=t(output_here)%*%A_hat
  var_est=0
  
  for(i_d in 1:d){
    
    m.model=fgasp(input,output_t_A[,i_d],have_noise=TRUE,kernel_type=kernel_type)
    m.pred=predict(param=c(log(beta[i_d]),log(sigma_2_0/sigma_2[i_d])),object=m.model,
                   testing_input=testing_input,var_data=FALSE,sigma_2=sigma_2[i_d])
    predict_all_dlm[,i_d]=m.pred@mean
    
    var_all_dlm[,i_d]= m.pred@var  
    
    var_est= var_est+(A_hat[,i_d]^2)%*%t(as.matrix(var_all_dlm[,i_d],num_testing,1))
  }
  
  return.list=as.list(1:2)
  return.list[[1]]=A_hat%*%t(predict_all_dlm)
  if(var_data==F){
    return.list[[2]]=var_est
  }else{
    return.list[[2]]=var_est+rep(sigma_2_0,k)
  }
  
  return.list
  
}

get_chol<-function(x,beta){
  # get cholesky decomposition
  R0_00=abs(outer(x,x,'-'))
  R=matern_5_2_funct(R0_00,beta)
  rcppeigen_get_chol(R)
}

gppca <- function(input, output, d,est_d=FALSE, shared_params = TRUE, kernel_type="matern_5_2"){
  object <- new("gppca", input, output, d, est_d, shared_params, kernel_type)
  return(object)
}

gppca_fit <- function(input, output, d, est_d, shared_params, kernel_type, sigma0_2, d_ub){
  input=sort(input)
  delta_x=input[2:length(input)]-input[1:(length(input)-1)]
  
  k = nrow(output)
  n = ncol(output)
  output_2=output%*%t(output)
  trace_output_2=sum(output^2)
  svd_output=svd(output)
  
  if(est_d){
    if(is.null(d_ub)){
      d_ub = k
    }
    if(is.null(sigma0_2)){
      # variance of noise if unknown, use IC to estimate d
      IC_vals = rep(NA, d_ub)
      for(i_d in 1:d_ub){
        criteria_val = log(mean((output - svd_output$u[,1:i_d]%*%t(svd_output$u[,1:i_d])%*%output)^2)) + i_d*(k+n)/(k*n)*log(k*n/(k+n))
        IC_vals[i_d] = criteria_val
      }
      d = which.min(IC_vals)
      print(d)
      
    }
    else{
      # variance is known, use binary search
      left = 1
      right = d_ub
      
      while(left <= right){
        mid = left + floor((right-left)/2)
        if(shared_params){
          param_ini=c(0,log(10))
          m=try(optim(param_ini,neg_log_lik_shared_cov_FFBS, kernel_type=kernel_type,output=output,
                      delta_x=delta_x, d=mid,method="L-BFGS-B"),silent=T)
          
          while(is.character(m)){
            param_ini=param_ini+runif(2)
            m=try(optim(param_ini,neg_log_lik_shared_cov_FFBS,kernel_type=kernel_type, output=output,
                        delta_x=delta_x,d=mid, method="L-BFGS-B"),silent=T)
          }
          G_log_det_cov_hat=Get_G_log_det_cov(m$par, output, delta_x,d=1,kernel_type=kernel_type)
          G_hat=output_2- G_log_det_cov_hat[[1]]
          eigen_G_hat=eigen(G_hat)
          sigma_2_0_hat=(trace_output_2-sum(eigen_G_hat$values[1:mid]))/(n*k)
        }else{
          G_hat=as.list(1:d)
          param_ini=c(rep(0,mid),rep(log(10),mid))
          m=try(optim(param_ini,neg_log_lik_diff_cov_FFBS,A_ini=A_ini,output=output,delta_x=delta_x,
                      lower=c(rep(log(10^{-10}),mid),rep(log(10^{-10}),mid)),
                      upper=c(rep(log(10^10),mid),rep(log(10^10),mid)),control = list(maxit = 100), method="L-BFGS-B"),silent=T)
          
          while(is.character(m)){
            param_ini=param_ini+runif(2*mid)
            #print(is.character(m))
            m=try(optim(param_ini,neg_log_lik_diff_cov_FFBS,A_ini=A_ini,output=output,delta_x=delta_x,
                        lower=c(rep(log(10^{-10}),mid),rep(log(10^{-10}),mid)),
                        upper=c(rep(log(10^10),mid),rep(log(10^10),mid)),control = list(maxit = 100), method="L-BFGS-B"),silent=T)
          }
          
          tau_hat=exp(m$par[(mid+1):(2*mid)])
          G_log_det_cov_hat=Get_G_log_det_cov(m$par, output, delta_x,d,kernel_type = kernel_type)
          for(i in 1:mid){
            G_hat[[i]]=output_2-G_log_det_cov_hat[[i]]
          }
          A_hat=Optimization_Stiefel_Manifold(A_ini, G=G_hat,max_iter=100)
          sigma_2_0_hat=(trace_output_2+F_Funct(A_hat,G_hat))/(n*k)
        }
        if(sigma_2_0_hat < sigma0_2){
          # overfit, reduce d
          right = mid-1
        }
        else{
          # underfit, increase d
          left = mid + 1
        }
      }
      d = mid
    }
    
  }
  
  ## start to implement GPPCA, shared_params=TRUE
  A_ini=svd_output$u[,1:d] 
  if(shared_params){
    param_ini=c(0,log(10))
    m=try(optim(param_ini,neg_log_lik_shared_cov_FFBS, kernel_type=kernel_type,output=output,
                delta_x=delta_x, d=d, method="L-BFGS-B"),silent=T)
    
    while(is.character(m)){
      #print(is.character(m))
      param_ini=param_ini+runif(2)
      m=try(optim(param_ini,neg_log_lik_shared_cov_FFBS,kernel_type=kernel_type, output=output,
                  delta_x=delta_x,d=d, method="L-BFGS-B"),silent=T)
    } 
    
    beta_hat=exp(m$par[1])
    tau_hat=exp(m$par[2])
    
    G_log_det_cov_hat=Get_G_log_det_cov(m$par, output, delta_x,d=1,kernel_type=kernel_type)
    G_hat=output_2- G_log_det_cov_hat[[1]]
    eigen_G_hat=eigen(G_hat)
    sigma_2_0_hat=(trace_output_2-sum(eigen_G_hat$values[1:d]))/(n*k)
    sigma2_hat = tau_hat*(sigma_2_0_hat)
    est_param = c(beta_hat,sigma2_hat, sigma_2_0_hat)
    
    
    AZ_posterior=pred_FFBS_FastGaSP(param=est_param,A_hat=eigen_G_hat$vectors[,1:d],input=input,testing_input=input, output_here=output,d=d,kernel_type=kernel_type)
    Y_hat=AZ_posterior[[1]]
    Y_var=AZ_posterior[[2]]
    Y_95_lower=Y_hat+sqrt(Y_var)*qnorm(0.025)
    Y_95_upper=Y_hat+sqrt(Y_var)*qnorm(0.975)
    
    res = list(est_A = eigen_G_hat$vectors[,1:d],
               est_beta = beta_hat,
               est_sigma0_2 = sigma_2_0_hat,
               est_sigma2 = sigma2_hat,
               mean_obs = Y_hat,
               mean_obs_95lb = Y_95_lower,
               mean_obs_95ub = Y_95_upper)
  }else{
    G_hat=as.list(1:d)
    param_ini=c(rep(0,d),rep(log(10),d))
    m=try(optim(param_ini,neg_log_lik_diff_cov_FFBS,A_ini=A_ini,output=output,delta_x=delta_x,
                lower=c(rep(log(10^{-10}),d),rep(log(10^{-10}),d)),
                upper=c(rep(log(10^10),d),rep(log(10^10),d)),control = list(maxit = 100), method="L-BFGS-B"),silent=T)
    
    while(is.character(m)){
      param_ini=param_ini+runif(2*d)
      #print(is.character(m))
      m=try(optim(param_ini,neg_log_lik_diff_cov_FFBS,A_ini=A_ini,output=output,delta_x=delta_x,
                  lower=c(rep(log(10^{-10}),d),rep(log(10^{-10}),d)),
                  upper=c(rep(log(10^10),d),rep(log(10^10),d)),control = list(maxit = 100), method="L-BFGS-B"),silent=T)
    }
    
    beta_hat=exp(m$par[1:d])
    tau_hat=exp(m$par[(d+1):(2*d)])
    G_log_det_cov_hat=Get_G_log_det_cov(m$par, output, delta_x,d,kernel_type = kernel_type)
    for(i in 1:d){
      G_hat[[i]]=output_2-G_log_det_cov_hat[[i]]
    }
    A_hat=Optimization_Stiefel_Manifold(A_ini, G=G_hat,max_iter=100)
    sigma_2_0_hat=(trace_output_2+F_Funct(A_hat,G_hat))/(n*k)
    sigma2_hat = tau_hat*(sigma_2_0_hat)
    
    est_param=c(beta_hat, sigma2_hat, sigma_2_0_hat)
    pred_GPPCA=pred_FFBS_FastGaSP_diff_cov(param=est_param,A_hat=A_hat,input=input,
                                  testing_input=input, output_here=output,
                                  d=d,var_data=F,kernel_type)
    
    Y_hat=pred_GPPCA[[1]]
    Y_var=pred_GPPCA[[2]]
    Y_95_lower=Y_hat+sqrt(Y_var)*qnorm(0.025)
    Y_95_upper=Y_hat+sqrt(Y_var)*qnorm(0.975)
    
    res = list(est_A = A_hat,
               est_beta = beta_hat,
               est_sigma0_2 = sigma_2_0_hat,
               est_sigma2 = sigma2_hat,
               mean_obs = Y_hat,
               mean_obs_95lb = Y_95_lower,
               mean_obs_95ub = Y_95_upper)
  }
  
  return(res)
}

predict_gppca <- function(param, A_hat, input, step=1, output,d, kernel_type, shared_param, interval=FALSE, interval_data=TRUE){
  n = ncol(output)
  testing_input = n + step
  if(shared_param){
    pred_gppca = pred_FFBS_FastGaSP(param,A_hat,input,testing_input=testing_input, output_here=output,d=d,kernel_type=kernel_type,var_data=interval_data)
  }
  else{
    pred_GPPCA=pred_FFBS_FastGaSP_diff_cov(param,A_hat,input,testing_input=testing_input, output_here=output,d,var_data=interval_data,kernel_type=kernel_type)
  }
  res = list(pred_mean = pred_gppca[[1]])
  Y_hat = pred_gppca[[1]]
  Y_var=pred_gppca[[2]]
  if(interval){
    res$pred_interval_95lb = Y_hat+sqrt(Y_var)*qnorm(0.025)
    res$pred_interval_95ub = Y_hat+sqrt(Y_var)*qnorm(0.975)
  }
  return(res)
}

## fgasp Class
setClass("fgasp", 		
         representation( 
           num_obs = "integer",          ## observations number
           have_noise="logical",            ## whether the data contain noise
           kernel_type="character",           ## information of the kernel
           ## data
           input="vector",              ## location of the sort input
           delta_x="vector",            ## distance between each sorted input
           output = "vector",           ## the observations 
           input_sorted_ix = "vector"           ## sorted input index
         ),
)

## show
if(!isGeneric("show")) {
  setGeneric(name = "show",
             def = function(object) standardGeneric("show")
  )
}

# if (!isGeneric("fit")) {
#   setGeneric("fit", function(object, ...) standardGeneric("fit"))
# }
setGeneric("fit",
           def = function(object, ...) {
             standardGeneric("fit")
           }
)

setMethod("show", "fgasp",
          function(object){
            show.fgasp(object)
          }
)


# pred.fgasp class
setClassUnion("vectorOrNULL", c("vector", "NULL"))

setClass("predictobj.fgasp", representation(
  num_testing="vector",     ##the number of testing input
  testing_input="vector",   ##sorted testing input
  param="vector",            ##param
  mean = "vector",          ##predictive mean 
  var="vector",              ##predictive variance
  var_data="logical",        ##whether to calculate the predictive variance of the data.
  index_x_all = "vectorOrNULL",        # a vector of all index locations
  delta_x_all = "vectorOrNULL",        # a vector of distances between all locations
  testing_loc = "vectorOrNULL",        # a vector oflocation for  the test inputs in the combined sorted inputs
  input_all_sorted_ix = "vectorOrNULL",        # a vector of sorted indices for all inputs
  testing_input_sorted_ix = "vectorOrNULL"        # a vector of sorted indiced for test inputs
  
),
)

if(!isGeneric("predict")) {
  setGeneric(name = "predict",
             def = function(object,...) standardGeneric("predict")
  )
}

setMethod("predict", "fgasp",
          definition=function(param=NA,object, testing_input=NULL, var_data=TRUE,sigma_2=NULL,
                              predict_input=F,index_return=F){
            predict.fgasp(param=param,object=object,testing_input=testing_input,var_data=var_data,
                          sigma_2=sigma_2,predict_input=predict_input,index_return=index_return)
          }
)


# fmou class
setClassUnion("numericOrNULL", c("numeric", "NULL"))
setClassUnion("matrixOrNULL", c("matrix", "NULL"))
setClassUnion("listOrNULL", c("list", "NULL"))

setClass(
  "fmou",
  representation  = list(
    output = "matrix",          # observation data matrix
    d = "numericOrNULL",        # number of latent factor
    est_d = "logical",          # whether to estimate d
    est_U0 = "logical",         # whether to estimate factor loading matrix 
    est_sigma0_2 = "logical",   # whether to estimate variance of noise
    U0 = "matrixOrNULL",        # fixed factor loading matrix
    sigma0_2 = "numericOrNULL" # fixed variance of noise
  )
)

setMethod(
  "initialize", "fmou",
  function(.Object, output, d = NULL, est_d = FALSE,
           est_U0 = TRUE, est_sigma0_2 = TRUE, U0 = NULL, 
           sigma0_2 = NULL) {
    
    .Object@output <- output
    .Object@d <- d
    .Object@est_d <- est_d
    .Object@est_U0 <- est_U0
    .Object@est_sigma0_2 <- est_sigma0_2
    .Object@U0 <- U0
    .Object@sigma0_2 <- sigma0_2
    
    # Add validation checks here (from the original function)
    if (.Object@est_d == FALSE & is.null(.Object@d)) {
      stop("d should be provided if est_d=FALSE")
    }
    
    if (.Object@est_U0 == FALSE & is.null(.Object@U0)) {
      stop("U0 must be provided if it is fixed.")
    }
    
    if (!is.null(.Object@U0)) {
      if (!all(abs(t(.Object@U0) %*% .Object@U0 - diag(ncol(.Object@U0))) < 1e-10)) {
        stop("U0 must be an orthogonal matrix.")
      }
    }
    
    if(.Object@est_sigma0_2==F & is.null(.Object@sigma0_2)){
      stop("sigma0_2 must be provided if it is fixed.")
    }
    if(.Object@est_sigma0_2==F & (!is.null(.Object@sigma0_2))){
      if(.Object@sigma0_2<0){
        stop("sigma0_2 must be nonnegative.")
      }
    }
    
    if(!is.null(.Object@d)){
      if(!is.null(.Object@U0)){
        if(nrow(.Object@U0)!=nrow(.Object@output) | ncol(.Object@U0)!=.Object@d){
          stop("U0 must be a k*d matrix.")
        }
      }
    }
    return(.Object)
  }
)


setGeneric("fit.fmou", function(object,...) standardGeneric("fit.fmou"))

setMethod(
  "fit.fmou", "fmou",
  function(object,M=50, threshold=1e-4,track_iterations=FALSE,track_neg_log_lik=FALSE,
           U_init=NULL, rho_init=NULL, sigma2_init=NULL, d_ub=NULL) {
    k <- nrow(object@output)
    n <- ncol(object@output)
    
    if(!is.null(U_init) & !is.null(object@d)){
      if(nrow(U_init)!=k | ncol(U_init)!=object@d){
        stop("U_init must be a k*d matrix.")
      }
    }
    
    if(!is.null(rho_init) & !is.null(object@d)){
      if(length(rho_init)!=object@d){
        stop("rho_init is a d-dimensional vector.")
      }
    }
    if(!is.null(sigma2_init) & !is.null(object@d)){
      if(length(sigma2_init)!=object@d){
        stop("sigma2_init is a d-dimensional vector.")
      }
      if(any(sigma2_init<0)){
        stop("sigma2_init is a non-negative vector.")
      }
    }
    
    # estimate d
    if(object@est_d & is.null(object@d)){ 
      if(object@est_sigma0_2){
        # noise level is unknown, perform IC
        if(is.null(d_ub)){
          d_ub = k
        }
        IC_vals = rep(NA, d_ub)
        U_IC = svd(object@output)$u
        
        for(i_d in 1:d_ub){
          criteria_val = log(mean((object@output - U_IC[,1:i_d]%*%t(U_IC[,1:i_d])%*%object@output)^2)) + i_d*(k+n)/(k*n)*log(k*n/(k+n))
          IC_vals[i_d] = criteria_val
        }
        object@d = which.min(IC_vals)
      }
      else{
        # noise level is known, perform variance matching (VM) with binary search
        if(is.null(d_ub)){
          d_ub=k
        }
        if(!is.null(U_init)){
          d_ub = min(d_ub, ncol(U_init))
        }
        if(!object@est_U0){
          d_ub = min(d_ub, ncol(object@U0))
        }
        if(!is.null(rho_init)){
          d_ub = min(d_ub, length(rho_init))
        }
        if(!is.null(sigma2_init)){
          d_ub = min(d_ub, length(sigma2_init))
        }
        left = 1
        right = d_ub
        
        while(left <= right){
          mid = left + floor((right-left)/2)
          em_fit = fmou_cpp(object@output, d=mid, M=M, threshold=threshold,
                            est_U0=object@est_U0, est_sigma0_2=T, 
                            track_iterations=F, 
                            track_neg_log_lik=F,
                            U0 = object@U0[,1:mid], U_init=U_init[,1:mid], 
                            rho_init=rho_init[1:mid],
                            sigma2_init=sigma2_init[1:mid],sigma0_2=NULL)
          est_noise = em_fit$sigma0_2
          if(est_noise < object@sigma0_2){
            # overfit, reduce d
            right = mid-1
          }
          else{
            # underfit, increase d
            left = mid + 1
          }
        }
        object@d = mid
      }
    }
    # estimate parameters in fmou
    fmou_fit <- fmou_cpp(
      object@output, 
      d = object@d, 
      M = M, 
      threshold = threshold,
      est_U0 = object@est_U0, 
      est_sigma0_2 = object@est_sigma0_2, 
      track_iterations = track_iterations, 
      track_neg_log_lik = track_neg_log_lik,
      U0 = object@U0[,1:object@d], 
      U_init = U_init[,1:object@d], 
      rho_init = rho_init[1:object@d],
      sigma2_init = sigma2_init[1:object@d], 
      sigma0_2 = object@sigma0_2
    )
    return(fmou_fit)
  }
)

setGeneric("predict.fmou", function(object,...) standardGeneric("predict.fmou"))

setMethod(
  "predict.fmou", "fmou",
  function(object, param_est, step=1, interval=FALSE, interval_data=TRUE) {
    d = object@d
    res = predict_fmou(param_est, step=1, interval=interval, interval_data=interval_data)
    return(res)
  }
)


#### GPPCA


setClass(
  "gppca",
  representation  = list(
    input = "numeric",           # input
    output = "matrix",          # observation matrix
    d = "numericOrNULL",        # number of latent factor
    est_d = "logical",          # whether to estimate d
    shared_params = "logical",  # whether latent processes share parameters
    kernel_type = "character"  # kernel type of latent processes
  )
)

setMethod(
  "initialize", "gppca",
  function(.Object, input, output, d,est_d=FALSE, shared_params = TRUE, 
           kernel_type="matern_5_2") {
    
    .Object@input <- input
    .Object@output <- output
    .Object@d <- d
    .Object@est_d <- est_d
    .Object@shared_params <- shared_params
    .Object@kernel_type <- kernel_type
    return(.Object)
  }
)

setGeneric("fit.gppca", function(object,...) standardGeneric("fit.gppca"))

setMethod(
  "fit.gppca", "gppca",
  function(object, sigma0_2=NULL, d_ub=NULL) {
    res <- gppca_fit(object@input, object@output, object@d, object@est_d,
                     object@shared_params, object@kernel_type, sigma0_2=NULL,
                     d_ub=NULL)
    return(res)
  }
)


setGeneric("predict.gppca", function(object,...) standardGeneric("predict.gppca"))

setMethod(
  "predict.gppca", "gppca",
  function(object, param, A_hat, step=1, interval=FALSE, interval_data=TRUE) {
    res <- predict_gppca(param=param, A_hat=A_hat, input=object@input, step=step,
                         output=object@output, d=object@d, kernel_type=object@kernel_type,
                         shared_param=object@shared_params,interval=interval, interval_data=interval_data)
    return(res)
  }
)

### particle simulations

# Define particle simulation class
# setClass("particle.data",
#          slots = list(
#            positions = "matrix",  # Matrix of particle positions
#            velocities = "matrix",  # Matrix of particle velocities
#            angles = "matrixOrNULL", # Matrix of angles
#            n_particles = "numeric",  # Number of particles
#            T_time = "numeric", # Time steps
#            sigma_0 = "numeric",  # Noise variance sigma_0
#            radius = "numeric", # Interaction radius
#            model = "character",  # Model type
#            D_y = "numeric" # Dimension of output
#          ))



# setMethod("initialize", "particle.data",
#           function(.Object, 
#                    positions = matrix(),velocities = matrix(),angles = NULL, # only needed for Vicsek
#                    n_particles = numeric(),T_time = numeric(),
#                    sigma_0 = numeric(),radius = numeric(),model = "Vicsek",  # Default model
#                    D_y = numeric(), 
#                    ...) {
#             # Assign values to slots
#             .Object@positions <- positions
#             .Object@velocities <- velocities
#             .Object@angles <- angles
#             .Object@n_particles <- n_particles
#             .Object@T_time <- T_time
#             .Object@sigma_0 <- sigma_0
#             .Object@radius <- radius
#             .Object@model <- model
#             .Object@D_y <- D_y
#             
#             # Validation
#             if (!identical(dim(positions), dim(velocities))) {
#               stop("Dimensions of positions and velocities must match")
#             }
#             
#             validObject(.Object)
#             return(.Object)
#           })
# 
# 
# setMethod("show", "particle.data",
#           function(object) {
#             cat("Model:", object@model, "\n")
#             cat("Number of particles:", object@n_particles, "\n")
#             cat("Number of time frames:", object@T_time, "\n")
#             cat("sigma_0 (noise variance):", object@sigma_0, "\n")
#             cat("Interaction radius:", object@radius, "\n")
#           })


setClassUnion("characterOrNULL", c("character", "NULL"))
setClassUnion("listOrNULL", c("list", "NULL"))
setClass("particle.data",
         slots = list(
           # Core data storage
           px_list = "list",  # List of x positions at each time
           py_list = "list",  # List of y positions at each time
           vx_list = "list",     # List of x velocities at each time
           vy_list = "list",     # List of y velocities at each time
           theta_list = "listOrNULL",  # Optional list of angles at each time
           
           # Tracking information (primarily for experimental data)
           particle_tracking = "listOrNULL",  # List of data frames for particle mappings
           
           # Metadata
           data_type = "character",    # "simulation" or "experiment"
           n_particles = "numeric",    # Number of particles (constant for simulation)
           T_time = "numeric",         # Total time steps
           D_y = "numeric",           # Dimension of output
           
           # Simulation-specific parameters (NULL for experimental)
           model = "characterOrNULL",      # Model type (e.g., "Vicsek")
           sigma_0 = "numericOrNULL",      # Noise variance
           radius = "numericOrNULL"        # Interaction radius
         ))

# Initialize method
setMethod("initialize", "particle.data",
          function(.Object,
                   # For both types
                   px_list = list(), py_list = list(),
                   vx_list = list(), vy_list = list(),
                   theta_list = NULL,
                   n_particles = numeric(),
                   T_time = numeric(),
                   D_y = numeric(),  # Default dimension
                   
                   # For experimental data
                   particle_tracking = NULL,
                   data_type = c("simulation", "experiment"),
                   
                   # For simulation data
                   model = NULL,
                   sigma_0 = NULL,
                   radius = NULL,
                   ...) {
            
            # Match data_type argument
            data_type <- match.arg(data_type)
            
            # Assign core data
            .Object@px_list <- px_list
            .Object@py_list <- py_list
            .Object@vx_list <- vx_list
            .Object@vy_list <- vy_list
            .Object@theta_list <- theta_list
            
            # Assign metadata
            .Object@data_type <- data_type
            .Object@n_particles <- n_particles
            .Object@T_time <- T_time
            .Object@D_y <- D_y
            
            # Assign tracking information for experimental data
            .Object@particle_tracking <- particle_tracking
            
            # Assign simulation-specific parameters
            .Object@model <- model
            .Object@sigma_0 <- sigma_0
            .Object@radius <- radius
            
            # Validation
            if (length(px_list) != length(py_list) ||
                length(px_list) != length(vx_list) ||
                length(px_list) != length(vy_list)) {
              stop("All list lengths must match")
            }
            
            # if (data_type == "simulation") {
            #   # For simulation, check constant particle numbers
            #   n_particles_per_frame <- vapply(px_list, length, numeric(1))
            #   if (!all(n_particles_per_frame == n_particles)) {
            #     stop("Simulation must have constant number of particles")
            #   }
            #   
            #   # Check simulation parameters
            #   if (is.null(model) || is.null(sigma_0) || is.null(radius)) {
            #     stop("Simulation data requires model, sigma_0, and radius parameters")
            #   }
            # } else {
            #   # For experimental, check particle tracking if provided
            #   if (!is.null(particle_tracking) && length(particle_tracking) != (length(px_list) - 1)) {
            #     stop("particle_tracking length must be T_time - 1")
            #   }
            # }
            
            # Additional validation for theta if provided
            if (!is.null(theta_list) && length(theta_list) != length(px_list)) {
              stop("theta_list length must match other lists")
            }
            
            validObject(.Object)
            return(.Object)
          })

# Show method
setMethod("show", "particle.data",
          function(object) {
            cat("Particle Data Object\n")
            cat("-------------------\n")
            cat("Data type:", object@data_type, "\n")
            if (object@data_type == "simulation") {
              cat("Model:", object@model, "\n")
              cat("Time steps:", object@T_time, "\n")
              cat("Number of particles:", object@n_particles, "\n")
              cat("Sigma_0 (noise variance):", object@sigma_0, "\n")
              cat("Interaction radius:", object@radius, "\n")
            } else {
              cat("Time steps:", object@T_time, "\n")
              cat("Average number of particles:", 
                  round(mean(vapply(object@px_list, length, numeric(1))),2), "\n")
            }
            # cat("Dimension:", object@D_y, "\n")
            # if (!is.null(object@theta_list)) {
            #   cat("Theta information: Available\n")
            # }
            # if (!is.null(object@particle_tracking)) {
            #   cat("Particle tracking: Available\n")
            # }
          })


# Define particle estimation class
setClass("particle.est",
         slots = list(
           data_type = "character",
           model = "characterOrNULL",  # Model type
           D_y = "numeric", # Dimension of output
           num_interaction = "numeric", # Number of interactions
           # data = "list",  # Input data list
           parameters = "numeric",   # Estimated parameters (beta, tau, radius)
           sigma_2_0_est = "numeric",
           predictions = "listOrNULL",   # Predictions and intervals
           training_data = "list",  # Training data used in GP model
           gp_weights = "matrix"    # GP model weights for prediction
         ))

# Initialize method
setMethod("initialize", "particle.est",
          function(.Object,
                   data_type = c("simulation", "experiment"),
                   model = NULL,
                   D_y = numeric(),
                   num_interaction = numeric(),
                   parameters = numeric(),
                   sigma_2_0_est = numeric(),
                   predictions = NULL,
                   training_data = list(),
                   gp_weights = matrix(),
                   ...) {
            
            # Match data_type argument
            data_type <- match.arg(data_type)
            
            # Assign slots
            .Object@data_type <- data_type
            .Object@model <- model
            .Object@D_y <- D_y
            .Object@num_interaction <- num_interaction
            .Object@parameters <- parameters
            .Object@sigma_2_0_est <- sigma_2_0_est
            .Object@predictions <- predictions
            .Object@training_data <- training_data
            .Object@gp_weights <- gp_weights
            
            # # Validation
            # if (data_type == "simulation" && 
            #     !model %in% c("Vicsek", "two_interactions_Vicsek")) {
            #   stop("For simulation data, model must be either 'Vicsek' or 'two_interactions_Vicsek'")
            # }
            # 
            # if (data_type == "experiment" && !is.null(model)) {
            #   stop("For experimental data, model should be NULL")
            # }
            
            # if (length(parameters) != 2*num_interaction + 1) {
            #   stop("Length of parameters vector must be 2*num_interaction + 1")
            # }
            
            validObject(.Object)
            return(.Object)
          })

# Show method
setMethod("show", "particle.est",
          function(object) {
            cat("Particle Model Estimation\n")
            cat("-------------------------\n")
            cat("Data type:", object@data_type, "\n")
            
            if (object@data_type == "simulation") {
              cat("Model type:", object@model, "\n")
            }
            
            cat("Output dimension:", object@num_interaction, "\n")
            cat("\nEstimated parameters:\n")
            cat("  beta (inverse range):", object@parameters[1:object@num_interaction], "\n")
            cat("  tau (variance-noise ratio):", 
                object@parameters[(object@num_interaction + 1):(2*object@num_interaction)], "\n")
            cat("  radius:", object@parameters[2*object@num_interaction + 1], "\n")
          })



setMethod("fit", "particle.data",
          definition=function(object, param, cut_r_max=1, est_param = TRUE, nx=NULL, ny=NULL, 
                              # px_list = object@px_list, py_list = object@py_list, 
                              # vx_list = object@vx_list, vy_list = object@vy_list, theta_list = object@theta_list,
                              # n_t = object@n_particles,T_time = object@T_time, D_y=object@D_y, 
                              kernel_type = "matern_5_2", tilde_nu=0.1, tol=1e-6, maxIte=1000,
                              output=NULL, ho_output = NULL, 
                              testing_inputs=NULL, compute_CI = TRUE, num_interaction = (length(param)-1)/2, 
                              data_type = object@data_type, model= object@model,apolar_vicsek=FALSE, direction = NULL){
            
            fit.particle.data(data=object, param = param, cut_r_max = cut_r_max, est_param = est_param, nx = nx, ny = ny,
                              # px_list = px_list,py_list = py_list, 
                              # vx_list = vx_list, vy_list = vy_list, theta_list = theta_list,
                              # n_t = n_t,T_time = T_time, D_y=D_y, 
                              kernel_type = kernel_type, tilde_nu=tilde_nu, tol=tol, maxIte=maxIte,
                              output=output, ho_output = ho_output, 
                              testing_inputs = testing_inputs, compute_CI=compute_CI, num_interaction = num_interaction,
                              data_type = data_type, model = model, apolar_vicsek=apolar_vicsek, direction = direction)
          }
)

IKF = function(beta, tilde_nu, delta_x, output, kernel_type='matern_5_2'){
  if(kernel_type=='matern_5_2'){
    lambda=sqrt(5)*beta
    
    W0=Construct_W0_matern_5_2(1,lambda)  
    GG=Construct_G_matern_5_2(delta_x,lambda) 
    W=Construct_W_matern_5_2(1.0,delta_x,lambda,W0)
  }else if(kernel_type=='exp'){
    lambda=beta
    W0=Construct_W0_exp(1,lambda)
    GG=Construct_G_exp(delta_x,lambda) 
    W=Construct_W_exp(1,delta_x,lambda,W0)
  }
  Q_K=Get_Q_K(GG,W,W0,tilde_nu)
  res=Get_R_y(GG,pmax(Q_K[[1]],0),Q_K[[2]],output)#-tilde_nu*output
  
  return(res)
}


initialization_Vicsek = function(n_t,v_abs){ ##2D initial position, uniform 
  p0=sqrt(n_t)*runif(n_t*2) ##uniform initial positions, enlarge the box size proportional to the number of particles 
  theta0=2*pi*runif(n_t)#-pi ###uniform direction 
  v0=rep(NA,2*n_t)
  for(i in 1:n_t){
    v0[(i-1)*2+1:2]=c(v_abs*cos(theta0[i]),v_abs*sin(theta0[i]))
  }
  return(list(p0 = p0, v0 = v0, theta0=theta0))
}


#
Vicsek = function(p0,v0,theta0,v_abs,n_t,T_sim,h,cut_r,sigma_0,noise_type='Gaussian'){ ##position, velocity, particle number, time, time interval, sigma is the noise
  pos=matrix(NA,2*n_t,T_sim+1)
  v=matrix(NA,2*n_t,T_sim+1)
  theta=matrix(NA,n_t,T_sim+1)
  
  v[,1]=v0
  pos[,1]=p0
  theta[,1]=theta0
  
  for(t in 1:T_sim){
    input_here=matrix(pos[,t],2,n_t)
    v_vec_here=matrix(v[,t],2,n_t)
    v_here=sqrt(colSums(v_vec_here^2))
    
    for(i in 1:n_t){
      input_i=as.vector(pos[(i-1)*2+1:2,t])
      d_vec_here_all=(input_i-input_here)
      
      d_here=sqrt(colSums(d_vec_here_all^2))
      
      index_neighbor=which(d_here<cut_r&v_here>0)
      theta_mean = atan2(mean(v_vec_here[2,index_neighbor]),mean(v_vec_here[1,index_neighbor]))
      
      if(noise_type=='Gaussian'){
        theta_i=theta_mean+sigma_0*rnorm(1)  ##only learn this function
      }else if(noise_type=='Uniform'){
        theta_i=theta_mean+sigma_0*(runif(1)-0.5)  ##only learn this function
      }
      
      v[(i-1)*2+1:2,t+1]=c(v_abs*cos(theta_i),v_abs*sin(theta_i))
      pos[(i-1)*2+1:2,t+1]=pos[(i-1)*2+1:2,t]+v[(i-1)*2+1:2,t]*h
      theta[i,t+1]=theta_i
      
    }
  }
  
  
  return(list(pos = pos, v=v, theta=theta))
}


simulate_Vicsek = function(v_abs, n_t, T_sim, h, cut_r, sigma_0, noise_type = 'Gaussian') {
  # Initialize the system
  initial_all = initialization_Vicsek(n_t = n_t, v_abs = v_abs)
  p0 = initial_all$p0 # Initial location
  v0 = initial_all$v0 # Initial speed
  theta0 = initial_all$theta0 # Initial velocity angle
  
  # Adjust theta0 to be within [-pi, pi]
  theta0[which(theta0 > pi)] = theta0[which(theta0 > pi)] - 2 * pi
  
  # Simulate the trajectory
  m_Vicsek = Vicsek(
    p0 = p0, v0 = v0, theta0 = theta0, 
    v_abs = v_abs, n_t = n_t, 
    T_sim = T_sim, h = h, cut_r = cut_r, 
    sigma_0 = sigma_0, noise_type = noise_type
  )
  
  
  
  # Split position matrix into x and y lists
  px_list = split(m_Vicsek$pos[seq(1, nrow(m_Vicsek$pos), 2), ], 
                  col(m_Vicsek$pos[seq(1, nrow(m_Vicsek$pos), 2), ]))
  py_list = split(m_Vicsek$pos[seq(2, nrow(m_Vicsek$pos), 2), ], 
                  col(m_Vicsek$pos[seq(2, nrow(m_Vicsek$pos), 2), ]))
  
  # Split velocity matrix into x and y lists
  vx_list = split(m_Vicsek$v[seq(1, nrow(m_Vicsek$v), 2), ], 
                  col(m_Vicsek$v[seq(1, nrow(m_Vicsek$v), 2), ]))
  vy_list = split(m_Vicsek$v[seq(2, nrow(m_Vicsek$v), 2), ], 
                  col(m_Vicsek$v[seq(2, nrow(m_Vicsek$v), 2), ]))
  
  # Split theta matrix into list
  theta_list = split(m_Vicsek$theta, col(m_Vicsek$theta))
  
  new("particle.data",
      px_list = px_list,
      py_list = py_list,
      vx_list = vx_list,
      vy_list = vy_list,
      theta_list = theta_list,
      data_type = "simulation",
      n_particles = n_t,
      T_time = T_sim,
      sigma_0 = sigma_0,
      radius = cut_r,
      model = "Vicsek",
      D_y = 1)
  
  
}


get_boundary_grid=function(px_min,px_max,py_min,py_max,nx,ny){
  grid_boundary_mat=matrix(NA,nx*ny,4)
  colnames(grid_boundary_mat)=c('pos_x_min','pos_x_max','pos_y_min','pos_y_max')
  
  ##get slightly larger grid 
  len_x_ori=px_max-px_min
  len_y_ori=py_max-py_min
  delta_x=len_x_ori/nx*0.1
  delta_y=len_y_ori/ny*0.1
  
  px_seq=seq(px_min-delta_x,px_max+delta_x,(px_max-px_min+2*delta_x)/(nx))
  py_seq=seq(py_min-delta_y,py_max+delta_y,(py_max-py_min+2*delta_y)/(ny))
  grid_boundary_mat[,1]=rep(px_seq[1:nx],ny)
  grid_boundary_mat[,2]=rep(px_seq[2:(nx+1)],ny)
  grid_boundary_mat[,3]=as.numeric(t(matrix(py_seq[1:ny],ny,nx)))
  grid_boundary_mat[,4]=as.numeric(t(matrix(py_seq[2:(ny+1)],ny,nx)))
  
  my_grid=list()
  
  
  Lx_min = min(grid_boundary_mat[,1:2]);
  Lx_max = max(grid_boundary_mat[,1:2]);
  Ly_min = min(grid_boundary_mat[,3:4]);
  Ly_max = max(grid_boundary_mat[,3:4]);
  
  len_x=  (max(grid_boundary_mat[,1:2])-min(grid_boundary_mat[,1:2]))/nx
  len_y=  (max(grid_boundary_mat[,3:4])-min(grid_boundary_mat[,3:4]))/ny
  
  
  grid_boundary_info=list()
  grid_boundary_info$grid_boundary_mat=grid_boundary_mat
  grid_boundary_info$grid_info=as.matrix(c(Lx_min,Lx_max,Ly_min,Ly_max,nx,ny,len_x,len_y),8,1)
  rownames(  grid_boundary_info$grid_info)=c('Lx_min','Lx_max','Ly_min','Ly_max','nx','ny','len_x','len_y')
  
  return(grid_boundary_info)
  
}



initiate_grid=function(grid_boundary_info){
  #include_theta = !is.null(theta) #if provided theta, then update theta to grid
  
  # Lx_min=unname(grid_boundary_info$grid_info['Lx_min',])
  # Lx_max=unname(grid_boundary_info$grid_info['Lx_max',])
  # Ly_min=unname(grid_boundary_info$grid_info['Ly_min',])
  # Ly_max=unname(grid_boundary_info$grid_info['Ly_max',])
  nx=unname(grid_boundary_info$grid_info['nx',])
  ny=unname(grid_boundary_info$grid_info['ny',])
  # len_x=unname(grid_boundary_info$grid_info['len_x',])
  # len_y=unname(grid_boundary_info$grid_info['len_y',])
  
  
  ##form neighboring particle
  ##from x first
  neighbor_index_list=as.list(1:(nx*ny))
  ##exterior
  for(i in 1:(nx*ny)){
    i_x=(i%%nx)
    if(i_x==0){
      i_x=nx
    }
    i_y=ceiling(i/nx)
    
    
    if((i_x-1)>0&(i_y-1)>0){
      neighbor_index_list[[i]]=c( neighbor_index_list[[i]],(i_y-2)*nx+(i_x-1))
    }
    
    if((i_y-1)>0){
      neighbor_index_list[[i]]=c( neighbor_index_list[[i]],(i_y-2)*nx+(i_x))
    }
    
    if((i_x+1)<=nx&(i_y-1)>0){
      neighbor_index_list[[i]]=c( neighbor_index_list[[i]],(i_y-2)*nx+(i_x+1))
    }
    
    if((i_x-1)>0){
      neighbor_index_list[[i]]=c( neighbor_index_list[[i]],(i_y-1)*nx+(i_x-1))
    }
    
    if((i_x+1)<=nx){
      neighbor_index_list[[i]]=c( neighbor_index_list[[i]],(i_y-1)*nx+(i_x+1))
    }
    
    if((i_x-1)>0&(i_y+1)<=ny){
      neighbor_index_list[[i]]=c( neighbor_index_list[[i]],(i_y)*nx+(i_x-1))
    }
    
    if((i_y+1)<=ny){
      neighbor_index_list[[i]]=c( neighbor_index_list[[i]],(i_y)*nx+(i_x))
    }
    if((i_x+1)<=nx&(i_y+1)<=ny){
      neighbor_index_list[[i]]=c( neighbor_index_list[[i]],(i_y)*nx+(i_x+1))
    }
  }
  
  
  
  
  #return(list(m_grid = m_grid, neighbor_index_list = neighbor_index_list))
  return(neighbor_index_list)
}

create_particle_grid=function(grid_boundary_info, pos_x,pos_y,vel_x,vel_y,neighbor_index_list){ #,return_theta=FALSE
  Lx_min=unname(grid_boundary_info$grid_info['Lx_min',])
  Lx_max=unname(grid_boundary_info$grid_info['Lx_max',])
  Ly_min=unname(grid_boundary_info$grid_info['Ly_min',])
  Ly_max=unname(grid_boundary_info$grid_info['Ly_max',])
  nx=unname(grid_boundary_info$grid_info['nx',])
  ny=unname(grid_boundary_info$grid_info['ny',])
  len_x=unname(grid_boundary_info$grid_info['len_x',])
  len_y=unname(grid_boundary_info$grid_info['len_y',])
  
  m_grid=vector(mode = 'list', nx*ny) #as.list(rep(NA,nx*ny))
  n_t=length(pos_x)
  
  for(i in 1:(nx*ny)){
    m_grid[[i]]=list(particle_pos = NULL, neighbor_pos=NULL, particle_vel=NULL, neighbor_vel=NULL)
    # if(return_theta){
    #   m_grid[[i]]=list(particle_pos = NULL, neighbor_pos=NULL, particle_vel=NULL, neighbor_vel=NULL,
    #                    particle_theta=NULL, neighbor_theta=NULL)
    # }else{
    #   
    # }
  }
  
  
  for(i in 1:n_t){
    i_x=ceiling((pos_x[i]-Lx_min)/len_x)
    i_y=ceiling((pos_y[i]-Ly_min)/len_y)
    
    index_grid=(i_y-1)*nx+i_x
    
    ##update pos and vel
    m_grid[[index_grid]]$particle_pos=cbind((m_grid[[index_grid]]$particle_pos),c(pos_x[i],pos_y[i]))
    m_grid[[index_grid]]$particle_vel=cbind((m_grid[[index_grid]]$particle_vel),c(vel_x[i],vel_y[i]))
    
    #if(return_theta) m_grid[[index_grid]]$particle_theta = c(m_grid[[index_grid]]$particle_theta,atan2(vel_y[i],vel_x[i]))
    
  }
  
  for(i in 1:(nx*ny)){
    #print(i)
    if(!is.null(m_grid[[i]]$particle_pos)){# only care about the grid that has particles
      neighbor_index = neighbor_index_list[[i]]
      for(idx in neighbor_index){
        if(!is.null(m_grid[[idx]]$particle_pos)){
          m_grid[[i]]$neighbor_pos=cbind(m_grid[[i]]$neighbor_pos,m_grid[[idx]]$particle_pos)
          m_grid[[i]]$neighbor_vel=cbind(m_grid[[i]]$neighbor_vel,m_grid[[idx]]$particle_vel)
          
          #if(return_theta) m_grid[[i]]$neighbor_theta = c(m_grid[[i]]$neighbor_theta, m_grid[[idx]]$particle_theta)
        }
      }
    }
    
  }
  
  return(m_grid)
  
}

find_grid_neighbors=function(pos_x_list,pos_y_list, vel_x_list,vel_y_list, time_range, #n_t,T_time,
                             grid_boundary_info){ #
  
  
  # Lx_min=grid_boundary_info$grid_info[1]
  # Lx_max=grid_boundary_info$grid_info[2]
  # Ly_min=grid_boundary_info$grid_info[3]
  # Ly_max=grid_boundary_info$grid_info[4]
  # nx=grid_boundary_info$grid_info[5]
  # ny=grid_boundary_info$grid_info[6]
  # len_x=grid_boundary_info$grid_info[7]
  # len_y=grid_boundary_info$grid_info[8]
  
  if(min(time_range)<1 | max(time_range)>length(pos_x_list)){
    stop('invalid time_range')
  }
  
  
  neighbor_index_list=initiate_grid(grid_boundary_info=grid_boundary_info)
  
  neighbors_info = vector(length(time_range), mode = "list")
  names(neighbors_info) <- paste0("time", time_range)
  for(t in time_range){
    #print(t)
    
    m_grid_here=create_particle_grid(grid_boundary_info=grid_boundary_info,
                                     pos_x=pos_x_list[[t]],pos_y=pos_y_list[[t]],
                                     vel_x=vel_x_list[[t]],vel_y=vel_y_list[[t]],
                                     neighbor_index_list=neighbor_index_list)
    
    neighbors_info[[t]] = m_grid_here
    
    
  }
  
  
  return(neighbors_info)
}


#### unnormalized Vicsek #####
unnormalized_Vicsek <- function(p0,v0,n_t,T_sim,h,cut_r,sigma_0,noise_type='Gaussian'){ ##position, velocity, particle number, time, time interval, sigma is the noise
  pos=matrix(NA,2*n_t,T_sim+1)
  v=matrix(NA,2*n_t,T_sim+1)
  
  v[,1]=v0
  pos[,1]=p0
  
  for(t in 1:T_sim){
    input_here=matrix(pos[,t],2,n_t)
    v_vec_here=matrix(v[,t],2,n_t)
    #v_here=sqrt(colSums(v_vec_here^2))
    
    for(i in 1:n_t){
      input_i=as.vector(pos[(i-1)*2+1:2,t])
      d_vec_here_all=(input_i-input_here) 
      
      d_here=sqrt(colSums(d_vec_here_all^2))
      
      index_neighbor=which(d_here<cut_r) #&v_here>0
      vy_neighbor=v_vec_here[2,index_neighbor]
      vx_neighbor=v_vec_here[1,index_neighbor]
      #theta_neighbor=atan2(v_vec_here[2,index_neighbor],v_vec_here[1,index_neighbor])
      
      v[(i-1)*2+1,t+1]=mean(vx_neighbor)
      v[(i-1)*2+2,t+1]=mean(vy_neighbor)
      
      ##add noise
      if(noise_type=='Gaussian'){
        v[(i-1)*2+1,t+1]=  v[(i-1)*2+1,t+1]+sigma_0*rnorm(1)  ##
        v[(i-1)*2+2,t+1]= v[(i-1)*2+2,t+1]+sigma_0*rnorm(1)  ##
      }else if(noise_type=='Uniform'){
        v[(i-1)*2+1,t+1]=  v[(i-1)*2+1,t+1]+sigma_0*(runif(1)-0.5)  ##
        v[(i-1)*2+2,t+1]= v[(i-1)*2+2,t+1]+sigma_0*(runif(1)-0.5)
      }
      
      pos[(i-1)*2+1:2,t+1]=pos[(i-1)*2+1:2,t]+v[(i-1)*2+1:2,t]*h
      
    }
  }
  
  
  return(list(pos = pos, v=v))
}

simulate_unnormalized_Vicsek = function(v_abs, n_t, T_sim, h, cut_r, sigma_0, noise_type = 'Gaussian') {
  # Initialize the system
  initial_all = initialization_Vicsek(n_t = n_t, v_abs = v_abs)
  p0 = initial_all$p0 # Initial location
  v0 = initial_all$v0 # Initial speed
  
  
  # Simulate the trajectory
  m_unnormalized_Vicsek = unnormalized_Vicsek(p0=p0,v0=v0,n_t=n_t,T_sim=T_sim,
                                              h=h,cut_r=cut_r,sigma_0=sigma_0,noise_type=noise_type)
  
  # input_pos_all = m_Vicsek$pos
  # v_all = m_Vicsek$v
  # theta_all = m_Vicsek$theta
  
  # Split position matrix into x and y lists
  px_list = split(m_unnormalized_Vicsek$pos[seq(1, nrow(m_unnormalized_Vicsek$pos), 2), ], 
                  col(m_unnormalized_Vicsek$pos[seq(1, nrow(m_unnormalized_Vicsek$pos), 2), ]))
  py_list = split(m_unnormalized_Vicsek$pos[seq(2, nrow(m_unnormalized_Vicsek$pos), 2), ], 
                  col(m_unnormalized_Vicsek$pos[seq(2, nrow(m_unnormalized_Vicsek$pos), 2), ]))
  
  # Split velocity matrix into x and y lists
  vx_list = split(m_unnormalized_Vicsek$v[seq(1, nrow(m_unnormalized_Vicsek$v), 2), ], 
                  col(m_unnormalized_Vicsek$v[seq(1, nrow(m_unnormalized_Vicsek$v), 2), ]))
  vy_list = split(m_unnormalized_Vicsek$v[seq(2, nrow(m_unnormalized_Vicsek$v), 2), ], 
                  col(m_unnormalized_Vicsek$v[seq(2, nrow(m_unnormalized_Vicsek$v), 2), ]))
  
  new("particle.data",
      px_list = px_list,
      py_list = py_list,
      vx_list = vx_list,
      vy_list = vy_list,
      data_type = "simulation",
      n_particles = n_t,
      T_time = T_sim,
      sigma_0 = sigma_0,
      radius = cut_r,
      model = "unnormalized_Vicsek",
      D_y = 2)
}



form_neighbors_unnormalized_Vicsek_with_r=function(threshold_r,pos_x_list,pos_y_list,vel_x_list,vel_y_list,
                                                   time_range,grid_boundary_info,neighbors_info,D_y){
  
  T_total = length(time_range)
  n_t = length(pos_x_list[[1]])
  
  Lx_min=unname(grid_boundary_info$grid_info['Lx_min',])
  Lx_max=unname(grid_boundary_info$grid_info['Lx_max',])
  Ly_min=unname(grid_boundary_info$grid_info['Ly_min',])
  Ly_max=unname(grid_boundary_info$grid_info['Ly_max',])
  nx=unname(grid_boundary_info$grid_info['nx',])
  ny=unname(grid_boundary_info$grid_info['ny',])
  len_x=unname(grid_boundary_info$grid_info['len_x',])
  len_y=unname(grid_boundary_info$grid_info['len_y',])
  
  
  A_vec=rep(NA,n_t*T_total*15*D_y)
  v_vec=rep(NA,n_t*T_total*15*D_y)
  num_neighbors_vec=rep(NA,n_t*T_total) ##here N neighbor each particle (obs) has
  
  count=0
  
  for(i_t in 1:length(time_range)){
    t = time_range[i_t]
    pos_x_t=pos_x_list[[t]]
    pos_y_t=pos_y_list[[t]]
    vel_x_t=vel_x_list[[t]]
    vel_y_t=vel_y_list[[t]]
    m_grid_here = neighbors_info[[t]]
    
    for(i in 1:n_t){
      input_pos_i=as.vector(c(pos_x_t[i],pos_y_t[i]))
      
      i_x=ceiling((input_pos_i[1]-Lx_min)/len_x)
      i_y=ceiling((input_pos_i[2]-Ly_min)/len_y)
      
      index_grid=(i_y-1)*nx+i_x
      d_vec_here_all=input_pos_i-as.matrix(m_grid_here[[index_grid]]$neighbor_pos)
      
      d_here=sqrt(colSums(d_vec_here_all^2))
      
      index_neighbor=which(d_here<threshold_r)
      # if(apolar_vicsek==F){
      #   index_neighbor=which(d_here<threshold_r)
      # }else{
      #   index_neighbor=which(d_here<threshold_r)
      #   index_same_v_direction=which(colSums(m_grid_here[[index_grid]]$neighbor_vel*input_vel_i)>=0)
      #   index_neighbor=intersect(index_neighbor,index_same_v_direction)
      # }
      n_neighbor=length(index_neighbor)
      num_neighbors_vec[n_t*(i_t-1)+i]=n_neighbor
      
      v_vec[count*D_y+(1:(n_neighbor*D_y))] = c(m_grid_here[[index_grid]]$neighbor_vel[1,index_neighbor],
                                                m_grid_here[[index_grid]]$neighbor_vel[2,index_neighbor])
      A_vec[2*count*D_y+(1:(2*n_neighbor*D_y))]=c(rep(c(1/n_neighbor,0),n_neighbor),rep(c(0,1/n_neighbor),n_neighbor))
      
      
      
      
      count=count+n_neighbor
      
      
    }
  }
  
  
  A_vec=A_vec[1:(2*count*D_y)]
  v_vec=v_vec[1:(count*D_y)]
  
  return(list(A_vec=A_vec, v_vec=v_vec, num_neighbors_vec=num_neighbors_vec))
}



pred_ho_output_unnormalized_Vicsek_log_RMSE=function(param, kernel_type, neighbors_info,grid_boundary_info,
                                                     pos_x_list, pos_y_list, vel_x_list, vel_y_list,
                                                     T_index_train, T_index_ho, output, ho_output, 
                                                     D_y, cut_r_max, tilde_nu, tol, maxIte){
  beta=exp(param[1])
  tau=exp(param[2])  ###sigma_2/sigma_2_0
  threshold_r=exp(param[3])/(1+exp(param[3])) * cut_r_max
  
  T_train=length(T_index_train) ##length of training data
  T_ho=length(T_index_ho) ##hold out prediction
  n_t = length(pos_x_list[[1]])
  
  
  ans_neighbors_train=form_neighbors_unnormalized_Vicsek_with_r(threshold_r=threshold_r,pos_x_list=pos_x_list,pos_y_list=pos_y_list,
                                                                vel_x_list=vel_x_list,vel_y_list=vel_y_list,time_range=T_index_train,
                                                                grid_boundary_info=grid_boundary_info,neighbors_info=neighbors_info,D_y=D_y)
  
  ans_neighbors_ho=form_neighbors_unnormalized_Vicsek_with_r(threshold_r=threshold_r,pos_x_list=pos_x_list,pos_y_list=pos_y_list,
                                                             vel_x_list=vel_x_list,vel_y_list=vel_y_list,time_range=T_index_ho,
                                                             grid_boundary_info=grid_boundary_info,neighbors_info=neighbors_info,D_y=D_y)
  
  N_train=n_t*T_train*D_y 
  N_ho=n_t*T_ho*D_y
  
  ##sort d_train
  sort_d_train=sort(ans_neighbors_train$v_vec,index.return=T) ###sorted d
  N_tilde=length(ans_neighbors_train$v_vec) ###N_tilde, distances  
  delta_x_train=sort_d_train$x[-1]-sort_d_train$x[-N_tilde]
  sort_d_train_ix=sort_d_train$ix
  
  ##form augmented samples for cross-validation
  N_ho_tilde=length(ans_neighbors_ho$v_vec)  #number of hold out distance
  d_aug=c(ans_neighbors_ho$v_vec,(ans_neighbors_train$v_vec))
  d_aug_sort=sort(d_aug,index.return=T) ##sort augmented samples, this will have N_aug_tilde log(N_aug_tilde) order?
  d_aug_sort_x=d_aug_sort$x
  d_aug_sort_rev_ix=sort(d_aug_sort$ix,index.return=T)$ix ###this is to reverse the previous sort 
  
  delta_x_aug=d_aug_sort_x[2:length(d_aug_sort_x)]-d_aug_sort_x[1:(length(d_aug_sort_x)-1)]
  
  ###finish construction, now start to predict
  
  m_CG=IKF_CG_particle( param,  kernel_type,   delta_x_train,   output, ans_neighbors_train$A_vec, # [[1]], 
                        sort_d_train_ix,  ans_neighbors_train$num_neighbors_vec*2, tilde_nu,
                        D_y,   N_train,   tol=tol,  maxIte = maxIte)
  
  ans_CG=m_CG[[1]] 
  
  ##this change this back to original parameterization
  ans_CG_tilde=ans_CG*tau ##this gives R_inv_y 
  
  ###z=A_t_sparse_times_x, 
  w_CG=A_t_times_x_particle(output=ans_CG_tilde, A_all_v=ans_neighbors_train$A_vec,  num_neighbors_vec=2*ans_neighbors_train$num_neighbors_vec,  
                            D_y=D_y, N_tilde=N_tilde)
  w_aug=c(rep(0,N_ho_tilde),w_CG)
  
  # param_here=log(c(beta,tilde_nu)) ##tilde nu is one to stablize the computation
  # pred_mean_aug=R_times_z(param_here, have_noise=T, delta_x=delta_x_aug, z=w_aug[d_aug_sort$ix],
  #                         kernel_type=kernel_type)-tilde_nu*w_aug[d_aug_sort$ix]
  pred_mean_aug = IKF(beta=beta, tilde_nu=tilde_nu, 
                      delta_x=delta_x_aug, output=w_aug[d_aug_sort$ix], kernel_type=kernel_type) - tilde_nu*w_aug[d_aug_sort$ix]
  
  pred_mean_fast=pred_mean_aug[d_aug_sort_rev_ix][1:N_ho_tilde]
  
  ##can only observes output so cross-validation on output
  pred_mean_ho_output=A_times_x_particle(output= pred_mean_fast, A_all_v= ans_neighbors_ho$A_vec,  num_neighbors_vec = 2*ans_neighbors_ho$num_neighbors_vec,  
                                         D_y,  N_ho)
  
  log_RMSE_ho=1/2*log(mean( (ho_output-pred_mean_ho_output)^2)) ##many pars should work as it contains noises
  
  #print(c(beta,tau,threshold_r,log_RMSE_ho))
  
  return(log_RMSE_ho)
  
  
}



particle_interaction_est_unnormalized_Vicsek = function(data_obj, param, cut_r_max, est_param=TRUE, nx=NULL, ny=NULL,
                                                        kernel_type='matern_5_2', tilde_nu=0.1, tol=1e-6, maxIte=1000, 
                                                        output=NULL, ho_output=NULL, testing_input=NULL, compute_CI = TRUE){ 
  
  px_list = data_obj@px_list
  py_list = data_obj@py_list
  vx_list = data_obj@vx_list
  vy_list = data_obj@vy_list
  n_t = data_obj@n_particles
  T_sim = data_obj@T_time
  D_y = data_obj@D_y
  
  N=n_t*T_sim*D_y 
  
  
  T_index_time = 1:T_sim
  
  
  px_min=min(unlist(px_list))
  px_max=max(unlist(px_list))
  py_min=min(unlist(py_list))
  py_max=max(unlist(py_list))
  
  if(is.null(nx)){
    nx=floor((px_max-px_min)/cut_r_max)
  }else{
    if(cut_r_max>(px_max-px_min)/nx) nx=floor((px_max-px_min)/cut_r_max)
  }
  
  if(is.null(ny)){
    ny=floor((py_max-py_min)/cut_r_max)
  }else{
    if(cut_r_max>(py_max-py_min)/ny) ny=floor((py_max-py_min)/cut_r_max)
  }
  
  grid_boundary_info=get_boundary_grid(px_min=px_min,px_max=px_max,
                                       py_min=py_min,py_max=py_max,nx=nx,ny=ny)
  #print(grid_boundary_info)
  
  
  neighbors_info = find_grid_neighbors(pos_x_list=px_list,pos_y_list=py_list,
                                       vel_x_list=vx_list,vel_y_list=vy_list, 
                                       time_range=T_index_time, grid_boundary_info=grid_boundary_info)
  
  
  
  
  if(est_param){
    ## split train and hold-out validation
    T_index_ho=seq(5,T_sim,5) ##every 5 use the last one as holdout
    T_index_train=(1:T_sim)[-T_index_ho]
    
    T_train=length(T_index_train) ##length of training data
    T_ho=length(T_index_ho) ##hold out prediction
    
    if(is.null(output)){
      output=as.vector(rbind(
        unlist(vx_list[T_index_train + 1]),
        unlist(vy_list[T_index_train + 1])
      ))#as.vector(v_all[,1+T_index_train])  
    } 
    if(is.null(ho_output)){
      ho_output=as.vector(rbind(
        unlist(vx_list[T_index_ho + 1]),
        unlist(vy_list[T_index_ho + 1])
      ))# as.vector(v_all[,1+T_index_ho])  
    }
    
    
    m_IKF=optim(param,pred_ho_output_unnormalized_Vicsek_log_RMSE, control=list(maxit=200),
                #lower=c(-8,-8,-8), upper=c(5,1,2),
                kernel_type=kernel_type, neighbors_info=neighbors_info,grid_boundary_info=grid_boundary_info,
                pos_x_list=px_list, pos_y_list=py_list, vel_x_list=vx_list, vel_y_list=vy_list,
                T_index_train=T_index_train, T_index_ho=T_index_ho, output=output, ho_output=ho_output, 
                D_y=D_y, cut_r_max=cut_r_max, tilde_nu=tilde_nu, tol=tol, maxIte=maxIte)
    
    while(m_IKF$par[2]>log(10^6)){
      param[2]=param[2]-log(10)+runif(1)
      m_IKF=optim(param,pred_ho_output_unnormalized_Vicsek_log_RMSE, control=list(maxit=200),
                  #lower=c(-8,-8,-8), upper=c(5,1,2),
                  kernel_type=kernel_type, neighbors_info=neighbors_info,grid_boundary_info=grid_boundary_info,
                  pos_x_list=px_list, pos_y_list=py_list, vel_x_list=vx_list, vel_y_list=vy_list,
                  T_index_train=T_index_train, T_index_ho=T_index_ho, output=output, ho_output=ho_output, 
                  D_y=D_y, cut_r_max=cut_r_max, tilde_nu=tilde_nu, tol=tol, maxIte=maxIte)
      
    }
    
    param=m_IKF$par
  }
  
  beta=exp(param[1])
  tau=exp(param[2])  ###sigma_2/sigma_2_0
  threshold_r=exp(param[3])/(1+exp(param[3])) * cut_r_max
  
  
  parameters = c(beta, tau, threshold_r)
  names(parameters) = c('beta', 'tau', 'radius')
  
  
  # # prediction
  output_all=as.vector(rbind(
    unlist(vx_list[1+T_index_time]),
    unlist(vy_list[1+T_index_time])
  ))
  
  ans_neighbors_all=form_neighbors_unnormalized_Vicsek_with_r(threshold_r=threshold_r,pos_x_list=px_list,pos_y_list=py_list,
                                                              vel_x_list=vx_list,vel_y_list=vy_list,time_range=T_index_time,
                                                              grid_boundary_info=grid_boundary_info,neighbors_info=neighbors_info,D_y=D_y)
  
  A_all_vec=ans_neighbors_all$A_vec
  v_all_vec=ans_neighbors_all$v_vec
  num_neighbors_all_vec=ans_neighbors_all$num_neighbors_vec
  sort_v_all=sort(v_all_vec,index.return=T)
  N_tilde_all=length(v_all_vec) ###this is N_j in the paper
  
  
  delta_x_all=sort_v_all$x[-1]-sort_v_all$x[-N_tilde_all]
  
  
  m_CG=IKF_CG_particle(param=log(c(beta,tau)), kernel_type=kernel_type, delta_x_all=delta_x_all, output=output_all, 
                       A_all_v = A_all_vec, sort_d_all_ix=sort_v_all$ix,  num_neighbors_vec=2*num_neighbors_all_vec, tilde_nu=tilde_nu,
                       D=D_y,  N=N,   tol=tol,  maxIte = maxIte)
  
  ans_CG=m_CG[[1]] 
  
  ans_CG_tilde=ans_CG*tau
  
  sigma_2_0_est = output_all%*%ans_CG/length(output_all) ##sometimes negative? solved
  
  ###z=A_t_sparse_times_x, get the weight; maybe write this in C++
  w_CG=A_t_times_x_particle(output=ans_CG_tilde, A_all_v=A_all_vec,  num_neighbors_vec=2*num_neighbors_all_vec,  
                            D_y=D_y, N_tilde=N_tilde_all)
  
  
  if(!is.null(testing_input)){
    testing_n = length(testing_input)
    
    
    sigma_2_est=sigma_2_0_est*tau
    
    param=log(c(beta,tau))
    
    d_aug=c(testing_input,(v_all_vec))
    d_aug_sort=sort(d_aug,index.return=T)
    d_aug_sort_x=d_aug_sort$x
    d_aug_sort_rev_ix=sort(d_aug_sort$ix,index.return=T)$ix ###this is to reverse the previous sort 
    
    delta_x_aug=d_aug_sort_x[2:length(d_aug_sort_x)]-d_aug_sort_x[1:(length(d_aug_sort_x)-1)]
    
    
    w_aug=c(rep(0,testing_n),w_CG)
    
    ###this should go back to nu
    # param_tilde=log(c(beta,tilde_nu)) 
    # pred_mean_aug=R_times_z(param_tilde, have_noise=T, delta_x=delta_x_aug, z=w_aug[d_aug_sort$ix],
    #                         kernel_type=kernel_type)-tilde_nu*w_aug[d_aug_sort$ix]
    pred_mean_aug = IKF(beta=beta, tilde_nu=tilde_nu, 
                        delta_x=delta_x_aug, output=w_aug[d_aug_sort$ix], kernel_type=kernel_type)-tilde_nu*w_aug[d_aug_sort$ix]
    # if(kernel_type=='matern_5_2'){
    #   lambda=sqrt(5)*beta
    #   
    #   W0=Construct_W0_matern_5_2(1,lambda)  
    #   GG=Construct_G_matern_5_2(delta_x_aug,lambda) 
    #   W=Construct_W_matern_5_2(1.0,delta_x_aug,lambda,W0)
    # }else if(kernel_type=='exp'){
    #   lambda=beta
    #   W0=Construct_W0_exp(1,lambda)
    #   GG=Construct_G_exp(delta_x_aug,lambda) 
    #   W=Construct_W_exp(1,delta_x_aug,lambda,W0)
    # }
    # Q_K=Get_Q_K(GG,W,W0,tilde_nu)
    # pred_mean_aug=Get_R_y(GG,Q_K[[1]],Q_K[[2]],w_aug[d_aug_sort$ix])-tilde_nu*w_aug[d_aug_sort$ix]
    
    
    pred_mean_fast=pred_mean_aug[d_aug_sort_rev_ix][1:testing_n]
    
    
    #NRMSE = mean( (pred_mean_fast-testing_output)^2)/sd(testing_output)
    if(compute_CI){
      #predictive variance
      c_star=rep(NA,testing_n)
      r0=abs(outer(testing_input,(v_all_vec),'-'))
      if(kernel_type=='exp'){
        r = exp(-beta*r0)
      }else if(kernel_type=='matern_5_2'){
        r = matern_5_2_funct(r0, beta)
      }
      
      #system.time(
      print("Computing the predictive variance ...")
      for(i in 1:testing_n ){
        #print(i)
        
        A_r_i=A_times_x_particle(output=r[i,], A_all_v=A_all_vec,  num_neighbors_vec=2*num_neighbors_all_vec,
                                 D=D_y, N)
        
        #tol=sd(A_r_i)^2*0.01*N_tilde ##can make it smaller
        tol_interval=tol*10^{-4}
        R_inv_r_all=IKF_CG_particle( param=param,  kernel_type=kernel_type,   delta_x_all=delta_x_all,   output=A_r_i,
                                     A_all_v=A_all_vec, sort_v_all$ix,  num_neighbors_vec=2*num_neighbors_all_vec, tilde_nu,
                                     D=D_y,   N=N,   tol=tol_interval,  maxIte = maxIte)
        R_inv_r=R_inv_r_all[[1]]*tau
        r_R_inv_r=A_r_i%*%R_inv_r
        
        c_star[i]=1-r_R_inv_r
        
      }
      c_star = abs(c_star)
      ##95 intervals
      LB95=    pred_mean_fast+sqrt(as.numeric(sigma_2_est)*c_star)*qnorm(0.025)
      UB95=    pred_mean_fast+sqrt(as.numeric(sigma_2_est)*c_star)*qnorm(0.975)
    }
    
    # coverage95 = length(which(testing_output<UB95 & testing_output>LB95))/length(testing_output)
    # length95=mean(UB95-LB95)
    
    #est_par_val=c(exp(m_IKF$par), exp(m_IKF$value))
    
    #return(list(NRMSE=NRMSE,coverage95=coverage95,length95=length95,est_par_val=est_par_val,pred_mean_fast=pred_mean_fast,LB95=LB95,UB95=UB95))
    #return(list(est_par_val=est_par_val,pred_mean_fast=pred_mean_fast,LB95=LB95,UB95=UB95))
    
  }
  
  
  
  # return(list(input_pos_all=input_pos_all,v_all = v_all,theta_all = theta_all,
  #             parameters = parameters, 
  #             pred_mean_fast = pred_mean_fast, LB95 = LB95, UB95 = UB95))
  new("particle.est",
      D_y=D_y,
      parameters = parameters,  # This contains the estimated parameters
      sigma_2_0_est = sigma_2_0_est[1,1],
      predictions = if(!is.null(testing_input)) {
        if(compute_CI) {
          list(mean = pred_mean_fast, lower95 = LB95, upper95 = UB95)
        } else {
          list(mean = pred_mean_fast)
        }
      } else {
        NULL
      },
      training_data = list(
        training_velocity = v_all_vec,
        A_v = A_all_vec,
        num_neighbors = num_neighbors_all_vec
        
      ),
      gp_weights = matrix(w_CG)  # Reshape weights if necessary
  )
}



#### Vicsek variation #####

f_Vicsek_variation=function(r,a=0.02,b=1,r_min=0.01,r_max=0.8, beta=20){
  fr=-a/(r+r_min)-b*(r-r_max)+a/r_max
  return(beta*fr)
}
##here D has to be 2
Vicsek_variation = function(p0,v0,n_t,D=2,T_sim,h,cut_r,sigma_0,
                            noise_type='Gaussian'){ ##position, velocity, particle number, time, time interval, sigma is the noise
  pos=matrix(NA,D*n_t,T_sim+1)
  v=matrix(NA,D*n_t,T_sim+1)
  #theta=matrix(NA,n_t,T_sim+1)
  
  v[,1]=v0
  pos[,1]=p0
  #ans_list=as.list(1:2)
  #set.seed(0)
  for(t in 1:T_sim){
    input_here=matrix(pos[,t],D,n_t)
    v_vec_here=matrix(v[,t],D,n_t)
    
    for(i in 1:n_t){
      input_i=input_here[,i]#as.vector(pos[(i-1)*D+1:D,t])
      d_vec_here_all=(input_here-input_i) ###same as in the paper
      
      
      d_here=sqrt(colSums(d_vec_here_all^2))
      v_here=sqrt(colSums(v_vec_here^2))
      ##need to delete itself
      index_neighbor=which(d_here<cut_r&d_here>0) #v_here>0
      
      
      if(length(index_neighbor)>0){
        #n_neighbor=length(index_neighbor)
        vx_sum_neighbor=v_vec_here[1,i]+sum(v_vec_here[1,index_neighbor])
        vy_sum_neighbor=v_vec_here[2,i]+sum(v_vec_here[2,index_neighbor])
        
        f_neighbor=f_Vicsek_variation(d_here[index_neighbor],r_max=cut_r)
        d_norm_x=d_vec_here_all[1,index_neighbor]/d_here[index_neighbor] ##normalized version
        d_norm_y=d_vec_here_all[2,index_neighbor]/d_here[index_neighbor] ##normalized version
        
        fx_sum_neighbor=sum(f_neighbor*(d_norm_x))
        fy_sum_neighbor=sum(f_neighbor*(d_norm_y))
        
        v_norm= (length(index_neighbor))
        
        ##D is 2
        v[(i-1)*D+1,t+1]= (vx_sum_neighbor/(v_norm+1)+fx_sum_neighbor/(v_norm))
        v[(i-1)*D+2,t+1]=(vy_sum_neighbor/(v_norm+1)+fy_sum_neighbor/(v_norm))
        
      }else{
        
        v[(i-1)*D+1,t+1]=v_vec_here[1,i]
        v[(i-1)*D+2,t+1]=v_vec_here[2,i]
        
      }
      
      #v_norm= sqrt(v[(i-1)*D+1,t+1]^2+ v[(i-1)*D+2,t+1]^2)
      #v_norm=1
      
      ##add noise
      if(noise_type=='Gaussian'){
        v[(i-1)*D+1,t+1]=  v[(i-1)*D+1,t+1]+sigma_0*rnorm(1)  ##
        v[(i-1)*D+2,t+1]= v[(i-1)*D+2,t+1]+sigma_0*rnorm(1)  ##
      }else if(noise_type=='Uniform'){
        v[(i-1)*D+1,t+1]=  v[(i-1)*D+1,t+1]+sigma_0*(runif(1)-0.5)  ##
        v[(i-1)*D+2,t+1]= v[(i-1)*D+2,t+1]+sigma_0*(runif(1)-0.5)
      }
      
      pos[(i-1)*D+1:D,t+1]=pos[(i-1)*D+1:D,t]+v[(i-1)*D+1:D,t]*h
      
      #theta[i,t+1]=atan2(v[(i-1)*D+1,t+1],v[(i-1)*D+2,t+1]) ##
      
      #c(v_abs*cos(theta_i),v_abs*sin(theta_i))
      
    }
  }
  ans_list = list(pos=pos, v=v) #, theta=theta
  
  return(ans_list)
  
}



form_neighbors_Vicsek_variation_with_r=function(threshold_r,pos_x_list,pos_y_list,vel_x_list,vel_y_list,
                                                time_range,grid_boundary_info,neighbors_info,D_y){
  T_total = length(time_range)
  n_t = length(pos_x_list[[1]])
  
  Lx_min=unname(grid_boundary_info$grid_info['Lx_min',])
  Lx_max=unname(grid_boundary_info$grid_info['Lx_max',])
  Ly_min=unname(grid_boundary_info$grid_info['Ly_min',])
  Ly_max=unname(grid_boundary_info$grid_info['Ly_max',])
  nx=unname(grid_boundary_info$grid_info['nx',])
  ny=unname(grid_boundary_info$grid_info['ny',])
  len_x=unname(grid_boundary_info$grid_info['len_x',])
  len_y=unname(grid_boundary_info$grid_info['len_y',])
  
  A_v_vec=rep(NA,n_t*T_total*15*D_y)
  A_f_vec=rep(NA,n_t*T_total*15*D_y)
  d_pos_vec=rep(NA,n_t*T_total*15)
  v_vec=rep(NA,n_t*T_total*15*D_y)
  num_neighbors_vec=rep(NA,n_t*T_total) 
  
  count=0
  
  #count_neighbor_start=0
  
  for(i_t in 1:length(time_range)){
    t = time_range[i_t]
    pos_x_t=pos_x_list[[t]]
    pos_y_t=pos_y_list[[t]]
    vel_x_t=vel_x_list[[t]]
    vel_y_t=vel_y_list[[t]]
    m_grid_here = neighbors_info[[t]]
    
    for(i in 1:n_t){
      input_pos_i=as.vector(c(pos_x_t[i],pos_y_t[i]))
      input_vel_i=as.vector(c(vel_x_t[i],vel_y_t[i]))
      
      i_x=ceiling((input_pos_i[1]-Lx_min)/len_x)
      i_y=ceiling((input_pos_i[2]-Ly_min)/len_y)
      
      index_grid=(i_y-1)*nx+i_x
      d_vec_here_all=as.matrix(m_grid_here[[index_grid]]$neighbor_pos)-input_pos_i
      
      d_here=sqrt(colSums(d_vec_here_all^2))
      
      #index_neighbor=which(d_here<cut_r_max)
      index_neighbor=which(d_here<threshold_r&d_here>0) # add i itself (d_here == 0) at the end later
      index_itself = which(d_here==0)
      
      n_neighbor=length(index_neighbor)+1
      num_neighbors_vec[n_t*(i_t-1)+i]=n_neighbor
      
      v_vec[count*D_y+(1:(n_neighbor*D_y))] = c(m_grid_here[[index_grid]]$neighbor_vel[1,c(index_neighbor,index_itself)],
                                                m_grid_here[[index_grid]]$neighbor_vel[2,c(index_neighbor,index_itself)])
      A_v_vec[2*count*D_y+(1:(2*n_neighbor*D_y))]=c(rep(c(1/n_neighbor,0),n_neighbor),rep(c(0,1/n_neighbor),n_neighbor))
      
      if(n_neighbor>1){
        e_vec_here=as.vector(d_vec_here_all[,index_neighbor]/rep(d_here[index_neighbor],each=D_y))
        A_f_vec[count*D_y+(1:(n_neighbor*D_y))]=c(e_vec_here/(n_neighbor-1),0,0)
      }else{
        #e_vec_here=c(0,0)
        A_f_vec[count*D_y+(1:D_y)]=c(0,0)
      }
      
      d_pos_vec[count+(1:n_neighbor)]=c(d_here[index_neighbor],0)
      
      count=count+n_neighbor
      
      
    }
  }
  A_v_vec=A_v_vec[1:(2*count*D_y)]
  A_f_vec=A_f_vec[1:(count*D_y)]
  d_pos_vec=d_pos_vec[1:count]
  v_vec=v_vec[1:(count*D_y)]
  
  ans_list = list(A_v_vec=A_v_vec,A_f_vec=A_f_vec,
                  v_vec=v_vec,d_pos_vec=d_pos_vec,
                  num_neighbors_vec=num_neighbors_vec)
  return(ans_list)
}

pred_ho_output_Vicsek_variation_log_RMSE=function(param, kernel_type, neighbors_info,grid_boundary_info,
                                                  pos_x_list, pos_y_list, vel_x_list, vel_y_list,
                                                  T_index_train, T_index_ho, output, ho_output, 
                                                  D_y, cut_r_max, tilde_nu, tol, maxIte){
  beta_v=exp(param[1])
  beta_f=exp(param[2])
  tau_v=exp(param[3])  
  tau_f=exp(param[4])  
  threshold_r=exp(param[5])/(1+exp(param[5])) * cut_r_max
  
  T_train=length(T_index_train) ##length of training data
  T_ho=length(T_index_ho) ##hold out prediction
  n_t = length(pos_x_list[[1]])
  
  ##form neighbor of train
  # ans_neighbors_train=form_neighbors_Vicsek_variation_with_max_cut_r(T_time=T_train,n_t=n_t,D_y=D_y,threshold_r=threshold_r,
  #                                                                    d_pos_vec_max=max_neighbors_train$d_pos_vec,
  #                                                                    v_vec_max=max_neighbors_train$v_vec,
  #                                                                    e_vec_max=max_neighbors_train$e_vec,
  #                                                                    num_neighbors_vec_max=max_neighbors_train$num_neighbors_vec)
  
  ans_neighbors_train=form_neighbors_Vicsek_variation_with_r(threshold_r=threshold_r,pos_x_list=pos_x_list,pos_y_list=pos_y_list,
                                                             vel_x_list=vel_x_list,vel_y_list=vel_y_list,time_range=T_index_train,
                                                             grid_boundary_info=grid_boundary_info,neighbors_info=neighbors_info,D_y=D_y)
  
  ##form neighbor of hold-out cross-validation
  # ans_neighbors_ho=form_neighbors_Vicsek_variation_with_max_cut_r(T_time=T_ho,n_t=n_t,D_y=D_y,threshold_r=threshold_r,
  #                                                                 d_pos_vec_max=max_neighbors_ho$d_pos_vec,
  #                                                                 v_vec_max=max_neighbors_ho$v_vec,
  #                                                                 e_vec_max=max_neighbors_ho$e_vec,
  #                                                                 num_neighbors_vec_max=max_neighbors_ho$num_neighbors_vec)
  
  ans_neighbors_ho=form_neighbors_Vicsek_variation_with_r(threshold_r=threshold_r,pos_x_list=pos_x_list,pos_y_list=pos_y_list,
                                                          vel_x_list=vel_x_list,vel_y_list=vel_y_list,time_range=T_index_ho,
                                                          grid_boundary_info=grid_boundary_info,neighbors_info=neighbors_info,D_y=D_y)
  
  
  
  N_train=n_t*T_train*D_y ##this one is N in training, the output y dimension
  N_ho=n_t*T_ho*D_y ##this one is N in testing
  
  ######### process for vx ###########
  ##sort vx_train
  sort_v_train=sort(ans_neighbors_train$v_vec,index.return=T) ###sorted vx
  N_v_tilde=length(ans_neighbors_train$v_vec) ###N_tilde, distances
  delta_v_train=sort_v_train$x[-1]-sort_v_train$x[-N_v_tilde]
  sort_v_train_ix=sort_v_train$ix
  
  ##form augmented samples for cross-validation
  N_v_ho_tilde=length(ans_neighbors_ho$v_vec)  #number of hold out distance
  v_aug=c(ans_neighbors_ho$v_vec,(ans_neighbors_train$v_vec))
  v_aug_sort=sort(v_aug,index.return=T) ##sort augmented samples, this will have N_aug_tilde log(N_aug_tilde) order?
  v_aug_sort_x=v_aug_sort$x
  v_aug_sort_rev_ix=sort(v_aug_sort$ix,index.return=T)$ix ###this is to reverse the previous sort
  
  delta_v_aug=v_aug_sort_x[2:length(v_aug_sort_x)]-v_aug_sort_x[1:(length(v_aug_sort_x)-1)]
  
  ######### process for f ###########
  ##sort f_train
  sort_f_train=sort(ans_neighbors_train$d_pos_vec,index.return=T) ###sorted f
  N_f_tilde=length(ans_neighbors_train$d_pos_vec) ###N_tilde, distances
  delta_f_train=sort_f_train$x[-1]-sort_f_train$x[-N_f_tilde]
  sort_f_train_ix=sort_f_train$ix
  
  ##form augmented samples for cross-validation
  N_f_ho_tilde=length(ans_neighbors_ho$d_pos_vec)  #number of hold out distance
  f_aug=c(ans_neighbors_ho$d_pos_vec,(ans_neighbors_train$d_pos_vec))
  f_aug_sort=sort(f_aug,index.return=T) ##sort augmented samples, this will have N_aug_tilde log(N_aug_tilde) order?
  f_aug_sort_x=f_aug_sort$x
  f_aug_sort_rev_ix=sort(f_aug_sort$ix,index.return=T)$ix ###this is to reverse the previous sort
  
  delta_f_aug=f_aug_sort_x[2:length(f_aug_sort_x)]-f_aug_sort_x[1:(length(f_aug_sort_x)-1)]
  
  ###finish construction, now start to predict
  m_CG=IKF_CG_particle_two_interact(param1=log(c(beta_v,tau_v)), param2=log(c(beta_f,tau_f)), 
                                    kernel_type1=kernel_type, kernel_type2=kernel_type, 
                                    delta_x_all1=delta_v_train, delta_x_all2=delta_f_train, 
                                    A_all_v1=ans_neighbors_train$A_v_vec, A_all_v2=ans_neighbors_train$A_f_vec, 
                                    sort_d_all_ix1=sort_v_train_ix, sort_d_all_ix2=sort_f_train_ix, 
                                    num_neighbors_vec1=2*ans_neighbors_train$num_neighbors_vec, num_neighbors_vec2=ans_neighbors_train$num_neighbors_vec, 
                                    output=output, tilde_nu=tilde_nu, 
                                    D=D_y, N=N_train, tol = tol, maxIte = maxIte)
  
  
  ans_CG=m_CG[[1]]
  
  ##this change this back to original parameterization
  ans_CG_v_tilde=ans_CG*tau_v ##this gives R_inv_y
  
  ###z=A_t_sparse_times_x,
  w_CG_v=A_t_times_x_particle(output=ans_CG_v_tilde, A_all_v=ans_neighbors_train$A_v_vec,  num_neighbors_vec=2*ans_neighbors_train$num_neighbors_vec,
                              D_y=D_y, N_tilde=N_v_tilde)
  w_v_aug=c(rep(0,N_v_ho_tilde),w_CG_v)
  
  # param_here=log(c(beta_v,tilde_nu)) ##tilde nu is one to stablize the computation
  # pred_mean_v_aug=R_times_z(param_here, have_noise=T, delta_x=delta_v_aug, z=w_v_aug[v_aug_sort$ix],
  #                           kernel_type=kernel_type)-tilde_nu*w_v_aug[v_aug_sort$ix]
  
  pred_mean_v_aug = IKF(beta=beta_v, tilde_nu=tilde_nu, 
                        delta_x=delta_v_aug, output=w_v_aug[v_aug_sort$ix], kernel_type=kernel_type)-tilde_nu*w_v_aug[v_aug_sort$ix]
  
  pred_mean_v_fast=pred_mean_v_aug[v_aug_sort_rev_ix][1:N_v_ho_tilde]
  
  ##can only observes output so cross-validation on output
  pred_mean_v_ho_output=A_times_x_particle( pred_mean_v_fast,  ans_neighbors_ho$A_v_vec,  2*ans_neighbors_ho$num_neighbors_vec,
                                            D_y,  N_ho)
  
  ##this change this back to original parameterization
  ans_CG_f_tilde=ans_CG*tau_f ##this gives R_inv_y
  
  ###z=A_t_sparse_times_x,
  w_CG_f=A_t_times_x_particle(output=ans_CG_f_tilde, A_all_v=ans_neighbors_train$A_f_vec,  num_neighbors_vec=ans_neighbors_train$num_neighbors_vec,
                              D_y=D_y, N_tilde=N_f_tilde)
  w_f_aug=c(rep(0,N_f_ho_tilde),w_CG_f)
  
  # param_here=log(c(beta_f,tilde_nu)) ##tilde nu is one to stablize the computation
  # pred_mean_f_aug=R_times_z(param_here, have_noise=T, delta_x=delta_f_aug, z=w_f_aug[f_aug_sort$ix],
  #                           kernel_type=kernel_type)-tilde_nu*w_f_aug[f_aug_sort$ix]
  pred_mean_f_aug = IKF(beta=beta_f, tilde_nu=tilde_nu, 
                        delta_x=delta_f_aug, output=w_f_aug[f_aug_sort$ix], kernel_type=kernel_type)-tilde_nu*w_f_aug[f_aug_sort$ix]
  
  pred_mean_f_fast=pred_mean_f_aug[f_aug_sort_rev_ix][1:N_f_ho_tilde]
  
  ##can only observes output so cross-validation on output
  pred_mean_f_ho_output=A_times_x_particle( pred_mean_f_fast,  ans_neighbors_ho$A_f_vec,  ans_neighbors_ho$num_neighbors_vec,
                                            D_y,  N_ho)
  
  
  pred_mean_ho_output=pred_mean_v_ho_output+pred_mean_f_ho_output
  
  log_RMSE_ho=1/2*log(mean( (ho_output-pred_mean_ho_output)^2)) ##many pars should work as it contains noises
  
  #print(c(beta_v,beta_f,tau_v,tau_f,threshold_r,log_RMSE_ho))
  
  return(log_RMSE_ho)
  
  
}


simulate_Vicsek_variation = function(v_abs, n_t, T_sim, h, cut_r, sigma_0, noise_type = 'Gaussian'){
  initial_all=initialization_Vicsek(n_t=n_t,v_abs=v_abs)
  p0=initial_all$p0 # initial location
  v0=initial_all$v0 # initial speed
  
  ##simulate the trajectory
  m_Vicsek_variation=Vicsek_variation(
    p0=p0,v0=v0,n_t=n_t, D = 2,
    T_sim=T_sim,h=h,cut_r=cut_r,
    sigma_0=sigma_0,noise_type=noise_type
  )
  # input_pos_all = m_Vicsek_variation$pos 
  # v_all = m_Vicsek_variation$v
  
  # Split position matrix into x and y lists
  px_list = split(m_Vicsek_variation$pos[seq(1, nrow(m_Vicsek_variation$pos), 2), ], 
                  col(m_Vicsek_variation$pos[seq(1, nrow(m_Vicsek_variation$pos), 2), ]))
  py_list = split(m_Vicsek_variation$pos[seq(2, nrow(m_Vicsek_variation$pos), 2), ], 
                  col(m_Vicsek_variation$pos[seq(2, nrow(m_Vicsek_variation$pos), 2), ]))
  
  # Split velocity matrix into x and y lists
  vx_list = split(m_Vicsek_variation$v[seq(1, nrow(m_Vicsek_variation$v), 2), ], 
                  col(m_Vicsek_variation$v[seq(1, nrow(m_Vicsek_variation$v), 2), ]))
  vy_list = split(m_Vicsek_variation$v[seq(2, nrow(m_Vicsek_variation$v), 2), ], 
                  col(m_Vicsek_variation$v[seq(2, nrow(m_Vicsek_variation$v), 2), ]))
  
  new("particle.data",
      px_list = px_list,
      py_list = py_list,
      vx_list = vx_list,
      vy_list = vy_list,
      data_type = "simulation",
      n_particles = n_t,
      T_time = T_sim,
      sigma_0 = sigma_0,
      radius = cut_r,
      model = "two_interactions_Vicsek",
      D_y = 2)
}

particle_interaction_est_Vicsek_variation=function(data_obj, param, cut_r_max, est_param=TRUE, nx=NULL, ny=NULL,#px_list, py_list, vx_list, vy_list, n_t, T_sim, D_y=2, 
                                                   kernel_type='matern_5_2',tilde_nu=0.1, tol=1e-6, maxIte=1000,
                                                   output=NULL, ho_output=NULL, testing_v_input=NULL, testing_d_input=NULL, compute_CI = TRUE){
  
  px_list = data_obj@px_list
  py_list = data_obj@py_list
  vx_list = data_obj@vx_list
  vy_list = data_obj@vy_list
  n_t = data_obj@n_particles
  T_sim = data_obj@T_time
  D_y = data_obj@D_y
  
  N=n_t*T_sim*D_y 
  
  T_index_time = 1:T_sim
  
  
  px_min=min(unlist(px_list))
  px_max=max(unlist(px_list))
  py_min=min(unlist(py_list))
  py_max=max(unlist(py_list))
  
  if(is.null(nx)){
    nx=floor((px_max-px_min)/cut_r_max)
  }else{
    if(cut_r_max>(px_max-px_min)/nx) nx=floor((px_max-px_min)/cut_r_max)
  }
  
  if(is.null(ny)){
    ny=floor((py_max-py_min)/cut_r_max)
  }else{
    if(cut_r_max>(py_max-py_min)/ny) ny=floor((py_max-py_min)/cut_r_max)
  }
  
  grid_boundary_info=get_boundary_grid(px_min=px_min,px_max=px_max,
                                       py_min=py_min,py_max=py_max,nx=nx,ny=ny)
  
  
  neighbors_info = find_grid_neighbors(pos_x_list=px_list,pos_y_list=py_list,
                                       vel_x_list=vx_list,vel_y_list=vy_list, 
                                       time_range=T_index_time, grid_boundary_info=grid_boundary_info)
  
  
  
  if(est_param){
    T_index_ho=seq(5,T_sim,5) ##every 5 use the last one as holdout
    T_index_train=(1:T_sim)[-T_index_ho]
    
    T_train=length(T_index_train) ##length of training data
    T_ho=length(T_index_ho) ##hold out prediction
    
    if(is.null(output)){
      output=as.vector(rbind(
        unlist(vx_list[T_index_train + 1]),
        unlist(vy_list[T_index_train + 1])
      ))#as.vector(v_all[,1+T_index_train])  
    } 
    if(is.null(ho_output)){
      ho_output=as.vector(rbind(
        unlist(vx_list[T_index_ho + 1]),
        unlist(vy_list[T_index_ho + 1])
      ))# as.vector(v_all[,1+T_index_ho])  
    }
    
    m_IKF=optim(param,pred_ho_output_Vicsek_variation_log_RMSE, control=list(maxit=200),
                #lower=c(-8,-8,-8), upper=c(5,1,2),
                kernel_type=kernel_type, neighbors_info=neighbors_info,grid_boundary_info=grid_boundary_info,
                pos_x_list=px_list, pos_y_list=py_list, vel_x_list=vx_list, vel_y_list=vy_list,
                T_index_train=T_index_train, T_index_ho=T_index_ho, output=output, ho_output=ho_output, 
                D_y=D_y, cut_r_max=cut_r_max, tilde_nu=tilde_nu, tol=tol, maxIte=maxIte)
    
    #while(m_IKF$par[3]>log(10^6) | m_IKF$par[4]>log(10^6)){
    ##10^8 for the second one
    while(m_IKF$par[3]>log(10^6) | m_IKF$par[4]>log(10^6)){
      if(m_IKF$par[3]>log(10^6)){
        param[3]=param[3]+runif(1)
      }
      if(m_IKF$par[4]>log(10^6)){
        param[4]=param[4]+runif(1)
      }
      
      m_IKF=optim(param,pred_ho_output_Vicsek_variation_log_RMSE, control=list(maxit=200),
                  #lower=c(-8,-8,-8), upper=c(5,1,2),
                  kernel_type=kernel_type, neighbors_info=neighbors_info,grid_boundary_info=grid_boundary_info,
                  pos_x_list=px_list, pos_y_list=py_list, vel_x_list=vx_list, vel_y_list=vy_list,
                  T_index_train=T_index_train, T_index_ho=T_index_ho, output=output, ho_output=ho_output, 
                  D_y=D_y, cut_r_max=cut_r_max, tilde_nu=tilde_nu, tol=tol, maxIte=maxIte)
    }
    
    
    param = m_IKF$par
  }
  
  beta_v=exp(param[1])
  beta_f=exp(param[2])
  tau_v=exp(param[3])  
  tau_f=exp(param[4])  
  threshold_r=exp(param[5])/(1+exp(param[5])) * cut_r_max
  
  parameters = c(beta_v, beta_f, tau_v, tau_f, threshold_r)
  names(parameters) = c('beta_v', 'beta_f', 'tau_v', 'tau_f', 'radius')
  
  
  
  ###for predicting
  # ans_neighbors_all=find_neighbors_Vicsek_variation_fast_grid(pos_x_list=px_list[T_index_time],pos_y_list=py_list[T_index_time], 
  #                                                             vel_x_list=vx_list[T_index_time],vel_y_list=vy_list[T_index_time], 
  #                                                             n_t=n_t,T_time=T_sim,D_y=D_y,grid_boundary_info=grid_boundary_info,cut_r=threshold_r)
  
  ans_neighbors_all=form_neighbors_Vicsek_variation_with_r(threshold_r=threshold_r,pos_x_list=px_list,pos_y_list=py_list,
                                                           vel_x_list=vx_list,vel_y_list=vy_list,time_range=T_index_time,
                                                           grid_boundary_info=grid_boundary_info,neighbors_info=neighbors_info,D_y=D_y)
  
  A_v_all_vec=ans_neighbors_all$A_v_vec
  A_f_all_vec=ans_neighbors_all$A_f_vec
  v_all_vec=ans_neighbors_all$v_vec
  d_pos_all_vec=ans_neighbors_all$d_pos_vec
  num_neighbors_all_vec=ans_neighbors_all$num_neighbors_vec
  
  sort_v_all=sort(v_all_vec,index.return=T)
  N_tilde_v_all=length(v_all_vec) ###this is N_j in the paper
  delta_x_v_all=sort_v_all$x[-1]-sort_v_all$x[-N_tilde_v_all]
  
  
  sort_f_all=sort(d_pos_all_vec,index.return=T)
  N_tilde_f_all=length(d_pos_all_vec) ###this is N_j in the paper
  delta_x_f_all=sort_f_all$x[-1]-sort_f_all$x[-N_tilde_f_all]
  
  output_all=as.vector(rbind(
    unlist(vx_list[2:(T_sim+1)]),
    unlist(vy_list[2:(T_sim+1)])
  ))#as.vector(v_all[,2:(T_sim+1)]) 
  
  m_CG=IKF_CG_particle_two_interact(param1=log(c(beta_v,tau_v)), param2=log(c(beta_f,tau_f)), 
                                    kernel_type1=kernel_type, kernel_type2=kernel_type, 
                                    delta_x_all1=delta_x_v_all, delta_x_all2=delta_x_f_all, 
                                    A_all_v1=A_v_all_vec, A_all_v2=A_f_all_vec, 
                                    sort_d_all_ix1=sort_v_all$ix, sort_d_all_ix2=sort_f_all$ix, 
                                    num_neighbors_vec1=2*num_neighbors_all_vec, num_neighbors_vec2=num_neighbors_all_vec, 
                                    output=output_all, tilde_nu=tilde_nu, 
                                    D=D_y, N=N, tol = tol, maxIte = maxIte)
  ans_CG=m_CG[[1]]
  sigma_2_0_est = output_all%*%ans_CG/length(output_all) ##sometimes negative? solved
  
  #first interaction
  ans_CG_v_tilde=ans_CG*tau_v ##this gives R_inv_y
  w_CG_v=A_t_times_x_particle(output=ans_CG_v_tilde, A_all_v=A_v_all_vec,  num_neighbors_vec=2*num_neighbors_all_vec,
                              D_y=D_y, N_tilde=N_tilde_v_all)
  #second interaction
  ans_CG_f_tilde=ans_CG*tau_f ##this gives R_inv_y
  w_CG_f=A_t_times_x_particle(output=ans_CG_f_tilde, A_all_v=A_f_all_vec,  num_neighbors_vec=num_neighbors_all_vec,
                              D_y=D_y, N_tilde=N_tilde_f_all)
  
  
  if(!is.null(testing_v_input) & !is.null(testing_d_input)){
    testing_n = length(testing_v_input)
    
    
    sigma_2_v_est=sigma_2_0_est*tau_v
    sigma_2_f_est=sigma_2_0_est*tau_f
    
    
    
    v_aug=c(testing_v_input,v_all_vec)
    v_aug_sort=sort(v_aug,index.return=T) ##sort augmented samples, this will have N_aug_tilde log(N_aug_tilde) order?
    v_aug_sort_x=v_aug_sort$x
    v_aug_sort_rev_ix=sort(v_aug_sort$ix,index.return=T)$ix ###this is to reverse the previous sort
    
    delta_v_aug=v_aug_sort_x[2:length(v_aug_sort_x)]-v_aug_sort_x[1:(length(v_aug_sort_x)-1)]
    
    
    w_v_aug=c(rep(0,testing_n),w_CG_v)
    
    # param_tilde=log(c(beta_v,tilde_nu)) ##tilde nu is one to stablize the computation
    # pred_mean_v_aug=R_times_z(param_tilde, have_noise=T, delta_x=delta_v_aug, z=w_v_aug[v_aug_sort$ix],
    #                           kernel_type=kernel_type)-tilde_nu*w_v_aug[v_aug_sort$ix]
    
    pred_mean_v_aug = IKF(beta=beta_v, tilde_nu=tilde_nu, 
                          delta_x=delta_v_aug, output=w_v_aug[v_aug_sort$ix], kernel_type=kernel_type)-tilde_nu*w_v_aug[v_aug_sort$ix]
    
    pred_mean_v_fast=pred_mean_v_aug[v_aug_sort_rev_ix][1:testing_n]
    
    #NRMSE_v=sqrt(mean((pred_mean_v_fast-testing_v_output)^2))/sd(testing_v_output)
    
    
    
    f_aug=c(testing_d_input,d_pos_all_vec)
    f_aug_sort=sort(f_aug,index.return=T) ##sort augmented samples, this will have N_aug_tilde log(N_aug_tilde) order?
    f_aug_sort_x=f_aug_sort$x
    f_aug_sort_rev_ix=sort(f_aug_sort$ix,index.return=T)$ix ###this is to reverse the previous sort
    
    delta_f_aug=f_aug_sort_x[2:length(f_aug_sort_x)]-f_aug_sort_x[1:(length(f_aug_sort_x)-1)]
    
    
    
    w_f_aug=c(rep(0,testing_n),w_CG_f)
    
    # param_tilde=log(c(beta_f,tilde_nu)) ##tilde nu is one to stablize the computation
    # pred_mean_f_aug=R_times_z(param_tilde, have_noise=T, delta_x=delta_f_aug, z=w_f_aug[f_aug_sort$ix],
    #                           kernel_type=kernel_type)-tilde_nu*w_f_aug[f_aug_sort$ix]
    
    pred_mean_f_aug = IKF(beta=beta_f, tilde_nu=tilde_nu, 
                          delta_x=delta_f_aug, output=w_f_aug[f_aug_sort$ix], kernel_type=kernel_type)-tilde_nu*w_f_aug[f_aug_sort$ix]
    
    pred_mean_f_fast=pred_mean_f_aug[f_aug_sort_rev_ix][1:testing_n]
    
    
    if(compute_CI){
      #######variance
      #first kernel
      c_v_star=rep(NA,testing_n)
      r0_v=abs(outer(testing_v_input,(v_all_vec),'-'))
      if(kernel_type=='exp'){
        r_v = exp(-beta*r0_v)
      }else if(kernel_type=='matern_5_2'){
        r_v = matern_5_2_funct(r0_v, beta_v)
      }
      
      print("Computing the predictive variance for the first interaction ...")
      #system.time(
      for(i in 1:testing_n){
        #print(i)
        #A_r_i=A_times_x_particle(output= r[i,],  A_all_v=A_all_v,  num_neighbors_vec=num_neighbors_vec,  
        #                       D=D, N_tilde=N_tilde,  T_sim=T_sim,  S=S)
        
        A_r_v_i=A_times_x_particle(output=r_v[i,], A_all_v=A_v_all_vec,  num_neighbors_vec=2*num_neighbors_all_vec,  
                                   D=D_y, N)
        
        #tol=sd(A_r_i)^2*0.01*N_tilde ##can make it smaller
        tol_interval=tol*10^{-4}
        R_inv_r_v_all=IKF_CG_particle_two_interact(param1=log(c(beta_v,tau_v)), param2=log(c(beta_f,tau_f)), 
                                                   kernel_type1=kernel_type, kernel_type2=kernel_type, 
                                                   delta_x_all1=delta_x_v_all, delta_x_all2=delta_x_f_all, 
                                                   A_all_v1=A_v_all_vec, A_all_v2=A_f_all_vec, 
                                                   sort_d_all_ix1=sort_v_all$ix, sort_d_all_ix2=sort_f_all$ix, 
                                                   num_neighbors_vec1=2*num_neighbors_all_vec, num_neighbors_vec2=num_neighbors_all_vec, 
                                                   output=A_r_v_i, tilde_nu=tilde_nu, 
                                                   D=D_y, N=N, tol = tol_interval, maxIte = maxIte)
        
        # R_inv_r_all=fast_pred_sparse_CG( param=param,  kernel_type=kernel_type,   delta_x_all=delta_x_all,   output=A_r_i, 
        #                                  A_all_v=A_all_vec, sort_d_all$ix,  num_neighbors_vec=num_neighbors_all_vec, tilde_nu,
        #                                  D=D_y,   N=N,   tol=tol_interval,  maxIte = maxIte)
        R_inv_r_v=R_inv_r_v_all[[1]]*tau_v
        r_R_inv_r_v=A_r_v_i%*%R_inv_r_v
        
        c_v_star[i]=1-r_R_inv_r_v
        ##Ur=sparse_A_times_z( A,   P=permutation_A, r[i,])
        
      }
      c_v_star=abs(c_v_star)
      
      LB95_v=    pred_mean_v_fast+sqrt(as.numeric(sigma_2_v_est)*c_v_star)*qnorm(0.025)
      UB95_v=    pred_mean_v_fast+sqrt(as.numeric(sigma_2_v_est)*c_v_star)*qnorm(0.975)
      
      
      #second kernel
      c_f_star=rep(NA,testing_n)
      r0_f=abs(outer(testing_d_input,(d_pos_all_vec),'-'))
      if(kernel_type=='exp'){
        r_f = exp(-beta*r0_f)
      }else if(kernel_type=='matern_5_2'){
        r_f = matern_5_2_funct(r0_f, beta_f)
      }
      
      print("Computing the predictive variance for the second interaction ...")
      #system.time(
      for(i in 1:testing_n){
        #print(i)
        #A_r_i=A_times_x_particle(output= r[i,],  A_all_v=A_all_v,  num_neighbors_vec=num_neighbors_vec,  
        #                       D=D, N_tilde=N_tilde,  T_sim=T_sim,  S=S)
        
        A_r_f_i=A_times_x_particle(output=r_f[i,], A_all_v=A_f_all_vec,  num_neighbors_vec=num_neighbors_all_vec,  
                                   D=D_y, N)
        
        #tol=sd(A_r_i)^2*0.01*N_tilde ##can make it smaller
        tol_interval=tol*10^{-4}
        R_inv_r_f_all=IKF_CG_particle_two_interact(param1=log(c(beta_v,tau_v)), param2=log(c(beta_f,tau_f)), 
                                                   kernel_type1=kernel_type, kernel_type2=kernel_type, 
                                                   delta_x_all1=delta_x_v_all, delta_x_all2=delta_x_f_all, 
                                                   A_all_v1=A_v_all_vec, A_all_v2=A_f_all_vec, 
                                                   sort_d_all_ix1=sort_v_all$ix, sort_d_all_ix2=sort_f_all$ix, 
                                                   num_neighbors_vec1=2*num_neighbors_all_vec, num_neighbors_vec2=num_neighbors_all_vec, 
                                                   output=A_r_f_i, tilde_nu=tilde_nu, 
                                                   D=D_y, N=N, tol = tol_interval, maxIte = maxIte)
        
        # R_inv_r_all=fast_pred_sparse_CG( param=param,  kernel_type=kernel_type,   delta_x_all=delta_x_all,   output=A_r_i, 
        #                                  A_all_v=A_all_vec, sort_d_all$ix,  num_neighbors_vec=num_neighbors_all_vec, tilde_nu,
        #                                  D=D_y,   N=N,   tol=tol_interval,  maxIte = maxIte)
        R_inv_r_f=R_inv_r_f_all[[1]]*tau_f
        r_R_inv_r_f=A_r_f_i%*%R_inv_r_f
        
        c_f_star[i]=1-r_R_inv_r_f
        ##Ur=sparse_A_times_z( A,   P=permutation_A, r[i,])
        
      }
      
      c_f_star=abs(c_f_star)
      
      LB95_f=    pred_mean_f_fast+sqrt(as.numeric(sigma_2_f_est)*c_f_star)*qnorm(0.025)
      UB95_f=    pred_mean_f_fast+sqrt(as.numeric(sigma_2_f_est)*c_f_star)*qnorm(0.975)
      
    }
    
    #est_par_val=c(exp(m_IKF$par), exp(m_IKF$value))
    
  }
  
  
  
  
  
  
  # return(list(input_pos_all=input_pos_all,v_all = v_all,
  #             parameters = parameters, 
  #             pred_mean_v_fast = pred_mean_v_fast, pred_mean_f_fast = pred_mean_f_fast,
  #             LB95_v = LB95_v, UB95_v = UB95_v, LB95_f=LB95_f, UB95_f = UB95_f))
  new("particle.est",
      D_y = D_y,
      # data = list(
      #   positions = input_pos_all,
      #   velocities = v_all
      # ),
      parameters = parameters,  # This contains the estimated parameters
      sigma_2_0_est = sigma_2_0_est[1,1], 
      predictions = if(!is.null(testing_v_input) & !is.null(testing_d_input)){
        if(compute_CI) {
          list(mean_v = pred_mean_v_fast,lower95_v = LB95_v,upper95_v = UB95_v,
               mean_f = pred_mean_f_fast,lower95_f = LB95_f,upper95_f = UB95_f)
        } else {
          list(mean_v = pred_mean_v_fast,
               mean_f = pred_mean_f_fast)
        }
      }else{
        NULL
      },
      training_data = list(
        training_velocity = v_all_vec,
        training_distance = d_pos_all_vec,
        A_v = A_v_all_vec,
        A_f = A_f_all_vec,
        num_neighbors =num_neighbors_all_vec
      ),
      gp_weights = cbind(w_v=w_CG_v,w_f=w_CG_f) #matrix(w_CG, ncol = D_y)
  )
  
}


### general functions

simulate_particle = function(v_abs, n_t=100, T_sim=5, h=0.1, 
                             cut_r=0.5, sigma_0=0.1, noise_type = 'Gaussian', model = "Vicsek"){
  
  if (!model %in% c("Vicsek", "unnormalized_Vicsek", "two_interactions_Vicsek")){
    stop("Invalid model specified. Model must be either 'Vicsek', 'unnormalized_Vicsek', or 'two_interactions_Vicsek'")
  }
  if(model == "Vicsek"){
    sim = simulate_Vicsek(v_abs = v_abs, n_t = n_t, T_sim = T_sim, h = h,
                          cut_r = cut_r, sigma_0 = sigma_0, noise_type = noise_type)
  }else if(model == "unnormalized_Vicsek"){
    sim = simulate_unnormalized_Vicsek(v_abs = v_abs, n_t = n_t, T_sim = T_sim, h = h,
                                       cut_r = cut_r, sigma_0 = sigma_0, noise_type = noise_type)
  }else if(model == "two_interactions_Vicsek"){
    sim = simulate_Vicsek_variation(v_abs = v_abs, n_t = n_t, T_sim = T_sim, h = h, 
                                    cut_r = cut_r, sigma_0 = sigma_0, noise_type = noise_type)
  }
  return(sim)
}



### cell
trajectory_data <- function(particle_data) {
  # Get time range
  T_start <- min(particle_data$time)
  T_end <- max(particle_data$time)
  T_time <- T_end - T_start 
  
  # Initialize lists
  px_list <- vector("list", T_time + 1)
  py_list <- vector("list", T_time + 1)
  vx_list <- vector("list", T_time + 1)
  vy_list <- vector("list", T_time + 1)
  theta_list <- vector("list", T_time + 1)
  n_record <- numeric(T_time)
  particle_tracking <- vector("list", T_time)
  
  # Pre-compute time indices for faster lookup
  time_indices <- split(seq_len(nrow(particle_data)), particle_data$time)
  
  # First, store all particle data for each time point
  for(t in T_start:T_end) {
    current_idx <- t - T_start + 1
    index_current <- time_indices[[as.character(t)]]
    
    current_data <- particle_data[index_current, ]
    px_list[[current_idx]] <- current_data$px
    py_list[[current_idx]] <- current_data$py
    vx_list[[current_idx]] <- current_data$vx
    vy_list[[current_idx]] <- current_data$vy
    theta_list[[current_idx]] <- atan2(current_data$vy, current_data$vx)
    if(t < T_end) {
      n_record[current_idx] <- length(index_current)
    }
    
  }
  
  # Then, create tracking information between consecutive frames
  for(t in T_start:(T_end-1)) {
    current_idx <- t - T_start + 1
    
    # Get indices for current and next frame
    index_current <- time_indices[[as.character(t)]]
    index_next <- time_indices[[as.character(t + 1)]]
    
    
    # Find shared particles
    particles_current <- particle_data$particleID[index_current]
    particles_next <- particle_data$particleID[index_next]
    shared_particleID <- intersect(particles_current, particles_next)
    
    if(length(shared_particleID) == 0) next
    
    # Get indices of shared particles in both frames
    match_current <- match(shared_particleID, particles_current)
    match_next <- match(shared_particleID, particles_next)
    
    # Store tracking information
    particle_tracking[[current_idx]] <- data.frame(
      t_idx = current_idx,
      particle_ids = shared_particleID,
      current_indices = match_current,
      next_indices = match_next
    )
  }
  
  # Create and return particle.data object
  new("particle.data",
      px_list = px_list,
      py_list = py_list,
      vx_list = vx_list,
      vy_list = vy_list,
      theta_list = theta_list,
      particle_tracking = particle_tracking,
      data_type = "experiment",
      n_particles = n_record,
      T_time = T_time,
      D_y = 1)
}

# # Utility function to get paired data from consecutive frames
get_consecutive_data <- function(data_obj, variable = c("vx", "vy", "px", "py", "theta")) {
  variable <- match.arg(variable)
  
  # Get the appropriate list based on the variable
  data_list <- switch(variable,
                      "vx" = data_obj@vx_list,
                      "vy" = data_obj@vy_list,
                      "px" = data_obj@px_list,
                      "py" = data_obj@py_list,
                      "theta" = data_obj@theta_list)
  
  if(is.null(data_list)) {
    stop(paste(variable, "data not available"))
  }
  
  T_time <- data_obj@T_time
  start_list <- vector("list", T_time)
  end_list <- vector("list", T_time)
  
  # For simulation data, it's straightforward
  if(data_obj@data_type == "simulation") {
    for(t in 1:T_time) {
      start_list[[t]] <- data_list[[t]]
      end_list[[t]] <- data_list[[t + 1]]
    }
  } else {
    # For experimental data, use particle tracking
    for(t in 1:T_time) {
      tracking <- data_obj@particle_tracking[[t]]
      if(!is.null(tracking)) {
        # Use tracking indices to get corresponding data
        current_frame_data <- data_list[[t]]
        next_frame_data <- data_list[[t + 1]]
        
        # Extract data in the correct order using tracking indices
        start_list[[t]] <- current_frame_data[tracking$current_indices]
        end_list[[t]] <- next_frame_data[tracking$next_indices]
      }
    }
  }
  
  list(start = start_list, end = end_list)
}


form_neighbors_cell_with_r = function(threshold_r,pos_x_list,pos_y_list,vel_x_list,vel_y_list,
                                      time_range,grid_boundary_info,neighbors_info,
                                      direction,apolar_vicsek = FALSE){
  
  n_record = sapply(pos_x_list, length)
  
  Lx_min=unname(grid_boundary_info$grid_info['Lx_min',])
  Lx_max=unname(grid_boundary_info$grid_info['Lx_max',])
  Ly_min=unname(grid_boundary_info$grid_info['Ly_min',])
  Ly_max=unname(grid_boundary_info$grid_info['Ly_max',])
  nx=unname(grid_boundary_info$grid_info['nx',])
  ny=unname(grid_boundary_info$grid_info['ny',])
  len_x=unname(grid_boundary_info$grid_info['len_x',])
  len_y=unname(grid_boundary_info$grid_info['len_y',])
  
  
  
  A_v_neighbor_record=rep(NA,sum(n_record)*15)
  v_neighbor_record=rep(NA,sum(n_record)*15)
  num_neighbors_vec=rep(NA,sum(n_record))  
  
  count=0
  
  for(i_t in 1:length(time_range)){
    t = time_range[i_t]
    pos_x_t=pos_x_list[[t]]
    pos_y_t=pos_y_list[[t]]
    vel_x_t=vel_x_list[[t]]
    vel_y_t=vel_y_list[[t]]
    m_grid_here = neighbors_info[[t]]
    n_t = n_record[t]
    
    index_start = ifelse(i_t==1, 0, sum(n_record[time_range][1:(i_t-1)]))
    
    for(i in 1:n_t){
      input_pos_i=as.vector(c(pos_x_t[i],pos_y_t[i]))
      input_vel_i=as.vector(c(vel_x_t[i],vel_y_t[i]))
      
      i_x=ceiling((input_pos_i[1]-Lx_min)/len_x)
      i_y=ceiling((input_pos_i[2]-Ly_min)/len_y)
      
      index_grid=(i_y-1)*nx+i_x
      d_vec_here_all=input_pos_i-as.matrix(m_grid_here[[index_grid]]$neighbor_pos)
      
      d_here=sqrt(colSums(d_vec_here_all^2))
      if(apolar_vicsek){
        index_neighbor=which(d_here<threshold_r)
        #index_same_v_direction=which(colSums(m_grid_here[[index_grid]]$neighbor_vel*input_vel_i)>0)
        #if(length(index_same_v_direction)==0){  ##if v is (0,0), the above can be integer(0)
        #  index_same_v_direction=1:(length(m_grid_here[[index_grid]]$neighbor_vel)/2) ##then all may be counted as same direction
        #}
        index_same_v_direction=which(colSums(m_grid_here[[index_grid]]$neighbor_vel*input_vel_i)>=0)
        
        index_neighbor=intersect(index_neighbor,index_same_v_direction)
      }else{
        index_neighbor=which(d_here<threshold_r)
      }
      
      n_neighbor = length(index_neighbor)
      num_neighbors_vec[index_start+i]=n_neighbor
      
      
      if(direction == "x"){
        v_neighbor_record[count+(1:n_neighbor)]=m_grid_here[[index_grid]]$neighbor_vel[1,index_neighbor]
      }else if(direction == "y"){
        v_neighbor_record[count+(1:n_neighbor)]=m_grid_here[[index_grid]]$neighbor_vel[2,index_neighbor]
      }
      
      A_v_neighbor_record[count+(1:n_neighbor)]=rep(1/n_neighbor,n_neighbor)
      
      
      
      count=count+n_neighbor
      
      
      
    }
    
  }
  
  v_neighbor_record=v_neighbor_record[1:count]
  A_v_neighbor_record=A_v_neighbor_record[1:count]
  
  ans_list = list(A_v_neighbor_record=A_v_neighbor_record,v_neighbor_record=v_neighbor_record,
                  num_neighbors_vec=num_neighbors_vec)
  return(ans_list)
}



pred_ho_output_cell_log_RMSE = function(param, kernel_type, neighbors_info,grid_boundary_info,
                                        pos_x_list, pos_y_list, vel_x_list, vel_y_list,sigma_2_record,
                                        T_index_train, T_index_ho, output, ho_output, direction, apolar_vicsek,
                                        D_y, cut_r_max, tilde_nu, tol, maxIte){
  beta=exp(param[1]) 
  tau=exp(param[2])  
  threshold_r=exp(param[3])/(1+exp(param[3])) * cut_r_max
  
  n_record = sapply(pos_x_list, length)
  
  # ##form neighbor of train
  # ans_neighbors_train=form_neighbors_cell_with_max_neighbor(v_max_neighbor_record=v_max_neighbor_record_train,
  #                                                           d_pos_max_vec=max_neighbors_apolar_train$d_pos_vec,
  #                                                           num_neighbors_max_vec=max_neighbors_apolar_train$num_neighbors_vec,
  #                                                           n_t_record=n_record[T_index_train],T_time=T_train,cut_r=threshold_r)
  
  ans_neighbors_train=form_neighbors_cell_with_r(threshold_r=threshold_r,pos_x_list=pos_x_list,pos_y_list=pos_y_list,
                                                 vel_x_list=vel_x_list,vel_y_list=vel_y_list,time_range=T_index_train,
                                                 grid_boundary_info=grid_boundary_info,neighbors_info=neighbors_info,
                                                 direction=direction,apolar_vicsek = apolar_vicsek)
  
  ##form neighbor of hold-out 
  # ans_neighbors_ho=form_neighbors_cell_with_max_neighbor(v_max_neighbor_record=v_max_neighbor_record_ho,
  #                                                        d_pos_max_vec=max_neighbors_apolar_ho$d_pos_vec,
  #                                                        num_neighbors_max_vec=max_neighbors_apolar_ho$num_neighbors_vec,
  #                                                        n_t_record=n_record[T_index_ho],T_time=T_ho,cut_r=threshold_r)
  
  ans_neighbors_ho=form_neighbors_cell_with_r(threshold_r=threshold_r,pos_x_list=pos_x_list,pos_y_list=pos_y_list,
                                              vel_x_list=vel_x_list,vel_y_list=vel_y_list,time_range=T_index_ho,
                                              grid_boundary_info=grid_boundary_info,neighbors_info=neighbors_info,
                                              direction=direction,apolar_vicsek = apolar_vicsek)
  
  N_train=D_y*sum(n_record[T_index_train])
  N_ho=D_y*sum(n_record[T_index_ho]) ##this one is N in testing
  
  ##sort vx_train
  sort_v_train=sort(ans_neighbors_train$v_neighbor_record,index.return=T) ###sorted vx
  N_v_tilde=length(ans_neighbors_train$v_neighbor_record) ###N_tilde, distances
  delta_v_train=sort_v_train$x[-1]-sort_v_train$x[-N_v_tilde]
  sort_v_train_ix=sort_v_train$ix
  
  ##form augmented samples for cross-validation
  N_v_ho_tilde=length(ans_neighbors_ho$v_neighbor_record)  #number of hold out distance
  v_aug=c(ans_neighbors_ho$v_neighbor_record,(ans_neighbors_train$v_neighbor_record))
  v_aug_sort=sort(v_aug,index.return=T) ##sort augmented samples, this will have N_aug_tilde log(N_aug_tilde) order?
  v_aug_sort_x=v_aug_sort$x
  v_aug_sort_rev_ix=sort(v_aug_sort$ix,index.return=T)$ix ###this is to reverse the previous sort
  
  delta_v_aug=v_aug_sort_x[2:length(v_aug_sort_x)]-v_aug_sort_x[1:(length(v_aug_sort_x)-1)]
  
  ###finish construction, now start to predict
  
  m_CG=IKF_CG_particle_cell(param=param, kernel_type=kernel_type, delta_x_all=delta_v_train, output=output, 
                            A_all_v=ans_neighbors_train$A_v_neighbor_record, sort_d_all_ix=sort_v_train_ix, 
                            sigma_2_vec=sigma_2_record[T_index_train], num_neighbors_vec=ans_neighbors_train$num_neighbors_vec, 
                            tilde_nu=tilde_nu, 
                            D=D_y, n_t_record=n_record[T_index_train], tol = tol, maxIte = maxIte)
  
  
  
  ans_CG=m_CG[[1]]
  
  ##this change this back to original parameterization
  ans_CG_v_tilde=ans_CG*tau ##this gives R_inv_y
  
  ###z=A_t_sparse_times_x,
  w_CG_v=A_t_times_x_particle(output=ans_CG_v_tilde, A_all_v=ans_neighbors_train$A_v_neighbor_record,  num_neighbors_vec=ans_neighbors_train$num_neighbors_vec,
                              D_y=D_y, N_tilde=N_v_tilde)
  w_v_aug=c(rep(0,N_v_ho_tilde),w_CG_v)
  
  # param_here=log(c(beta,tilde_nu)) ##tilde nu is one to stablize the computation
  # pred_mean_v_aug=R_times_z(param_here, have_noise=T, delta_x=delta_v_aug, z=w_v_aug[v_aug_sort$ix],
  #                           kernel_type=kernel_type)-tilde_nu*w_v_aug[v_aug_sort$ix]
  pred_mean_v_aug = IKF(beta=beta, tilde_nu=tilde_nu, 
                        delta_x=delta_v_aug, output=w_v_aug[v_aug_sort$ix], kernel_type=kernel_type)-tilde_nu*w_v_aug[v_aug_sort$ix]
  
  
  pred_mean_v_fast=pred_mean_v_aug[v_aug_sort_rev_ix][1:N_v_ho_tilde]
  
  
  ##can only observes output so cross-validation on output
  pred_mean_v_ho_output=A_times_x_particle( pred_mean_v_fast,  ans_neighbors_ho$A_v_neighbor_record,  ans_neighbors_ho$num_neighbors_vec,
                                            D_y,  N_ho)
  
  
  log_RMSE_ho=1/2*log(mean( (ho_output-pred_mean_v_ho_output)^2)) ##many pars should work as it contains noises
  
  #print(c(beta,tau,threshold_r,log_RMSE_ho))
  
  return(log_RMSE_ho)
  
}


particle_interaction_est_cell = function(data_obj, param, cut_r_max, est_param=TRUE, nx=NULL, ny=NULL, direction, 
                                         kernel_type='matern_5_2',tilde_nu=0.1, tol=1e-6, maxIte=1000, 
                                         output=NULL, ho_output = NULL, testing_input=NULL, compute_CI = TRUE, apolar_vicsek=FALSE){
  
  px_list = get_consecutive_data(data_obj, "px")$start
  py_list = get_consecutive_data(data_obj, "py")$start
  vx_pairs = get_consecutive_data(data_obj, "vx")
  vx_list = vx_pairs$start
  vx_end_list = vx_pairs$end
  vy_pairs = get_consecutive_data(data_obj, "vy")
  vy_list = vy_pairs$start
  vy_end_list = vy_pairs$end
  D_y = data_obj@D_y
  
  
  T_time = data_obj@T_time
  n_record = sapply(px_list,length)
  
  
  T_index_time = 1:T_time
  
  
  
  px_min=min(unlist(px_list))
  px_max=max(unlist(px_list))
  py_min=min(unlist(py_list))
  py_max=max(unlist(py_list))
  
  
  if(is.null(nx)){
    nx=floor((px_max-px_min)/cut_r_max)
  }else{
    if(cut_r_max>(px_max-px_min)/nx) nx=floor((px_max-px_min)/cut_r_max)
  }
  
  if(is.null(ny)){
    ny=floor((py_max-py_min)/cut_r_max)
  }else{
    if(cut_r_max>(py_max-py_min)/ny) ny=floor((py_max-py_min)/cut_r_max)
  }
  
  grid_boundary_info=get_boundary_grid(px_min=px_min,px_max=px_max,
                                       py_min=py_min,py_max=py_max,nx=nx,ny=ny)
  
  neighbors_info = find_grid_neighbors(pos_x_list=px_list,pos_y_list=py_list,
                                       vel_x_list=vx_list,vel_y_list=vy_list, 
                                       time_range=T_index_time, grid_boundary_info=grid_boundary_info)
  
  
  #sigma_2_record = rep(NA,T_time)
  
  
  if(est_param){
    T_index_ho=seq(5,T_time,5) ##every 5 use the last one as holdout
    T_index_train=(1:T_time)[-T_index_ho]
    
    T_train=length(T_index_train) ##length of training data
    T_ho=length(T_index_ho) ##hold out prediction
    
    if(direction == "x"){
      if(is.null(output)) output=unlist(vx_end_list[T_index_train])
      if(is.null(ho_output)) ho_output=unlist(vx_end_list[T_index_ho])
      sigma_2_record=sapply(vx_list, var)
      #testing_input=seq(min(unlist(vx_list[T_index_time])),max(unlist(vx_list[T_index_time])),length.out=testing_n)
    }else if(direction == "y"){
      if(is.null(output)) output=unlist(vy_end_list[T_index_train])
      if(is.null(ho_output)) ho_output=unlist(vy_end_list[T_index_ho])
      sigma_2_record=sapply(vy_list, var)
      #testing_input=seq(min(unlist(vy_list[T_index_time])),max(unlist(vy_list[T_index_time])),length.out=testing_n)
    }
    
    #param=log(c(1,50,50)) 
    m_IKF=optim(param,pred_ho_output_cell_log_RMSE, control=list(maxit=200),
                #lower=c(-8,-8,-8), upper=c(5,1,2),
                kernel_type=kernel_type, neighbors_info=neighbors_info,grid_boundary_info=grid_boundary_info,
                pos_x_list=px_list, pos_y_list=py_list, vel_x_list=vx_list, vel_y_list=vy_list,sigma_2_record=sigma_2_record,
                T_index_train=T_index_train, T_index_ho=T_index_ho, output=output, ho_output=ho_output, direction=direction,
                apolar_vicsek=apolar_vicsek, D_y=D_y, cut_r_max=cut_r_max, tilde_nu=tilde_nu, tol=tol, maxIte=maxIte)
    
    param = m_IKF$par
  }
  
  
  beta=exp(param[1])
  tau=exp(param[2]) 
  threshold_r=exp(param[3])/(1+exp(param[3])) * cut_r_max
  
  parameters = c(beta, tau, threshold_r)
  names(parameters) = c('beta', 'tau', 'radius')
  
  # # prediction
  # ans_neighbors_all=form_neighbors_cell_fast_grid(pos_x_list=px_list[T_index_time],pos_y_list=py_list[T_index_time], 
  #                                                 vel_x_list=vx_list[T_index_time],vel_y_list=vy_list[T_index_time],direction,
  #                                                 n_record=n_record[T_index_time],T_time,grid_boundary_info,cut_r=threshold_r,apolar_vicsek=apolar_vicsek)
  
  if(direction == "x"){
    output_all=unlist(vx_end_list[T_index_time])
    sigma_2_record=sapply(vx_list, var)
    #testing_input=seq(min(unlist(vx_list[T_index_time])),max(unlist(vx_list[T_index_time])),length.out=testing_n)
  }else if(direction == "y"){
    output_all=unlist(vy_end_list[T_index_time])
    sigma_2_record=sapply(vy_list, var)
    #testing_input=seq(min(unlist(vy_list[T_index_time])),max(unlist(vy_list[T_index_time])),length.out=testing_n)
  }
  
  ans_neighbors_all=form_neighbors_cell_with_r(threshold_r=threshold_r,pos_x_list=px_list,pos_y_list=py_list,
                                               vel_x_list=vx_list,vel_y_list=vy_list,time_range=T_index_time,
                                               grid_boundary_info=grid_boundary_info,neighbors_info=neighbors_info,
                                               direction=direction,apolar_vicsek = apolar_vicsek)
  
  A_v_all_vec=ans_neighbors_all$A_v_neighbor_record
  v_all_vec=ans_neighbors_all$v_neighbor_record
  #d_pos_vec=ans_neighbors_all$d_pos_vec
  num_neighbors_all_vec=ans_neighbors_all$num_neighbors_vec
  #mean(num_neighbors_all_vec)
  
  N=D_y*sum(n_record[T_index_time])
  sort_v_all=sort(v_all_vec,index.return=T)
  N_tilde_all=length(v_all_vec) ###this is N_j in the paper
  
  delta_x_all=sort_v_all$x[-1]-sort_v_all$x[-N_tilde_all]
  
  m_CG=IKF_CG_particle_cell(param=log(c(beta,tau)), kernel_type=kernel_type, delta_x_all=delta_x_all, output=output_all, 
                            A_all_v=A_v_all_vec, sort_d_all_ix=sort_v_all$ix, 
                            sigma_2_vec=sigma_2_record[T_index_time], num_neighbors_vec=num_neighbors_all_vec, tilde_nu=tilde_nu, 
                            D=D_y, n_t_record=n_record[T_index_time], tol = tol, maxIte = maxIte)
  ans_CG=m_CG[[1]]
  ans_CG_tilde=ans_CG*tau
  
  sigma_2_0_prop_est = output_all%*%m_CG[[1]]/length(output_all)
  
  w_CG=A_t_times_x_particle(output=ans_CG_tilde, A_all_v=A_v_all_vec,  num_neighbors_vec=num_neighbors_all_vec,  
                            D_y=D_y, N_tilde=N_tilde_all)
  
  
  if(!is.null(testing_input)){
    testing_n = length(testing_input)
    
    sigma_2_0_est=(as.numeric(sigma_2_0_prop_est)*sigma_2_record[T_index_time])
    sigma_2_est=sigma_2_0_prop_est*tau
    
    v_aug=c(testing_input,(v_all_vec))
    v_aug_sort=sort(v_aug,index.return=T)
    v_aug_sort_x=v_aug_sort$x
    v_aug_sort_rev_ix=sort(v_aug_sort$ix,index.return=T)$ix ###this is to reverse the previous sort 
    
    delta_x_aug=v_aug_sort_x[2:length(v_aug_sort_x)]-v_aug_sort_x[1:(length(v_aug_sort_x)-1)]
    
    
    w_aug=c(rep(0,testing_n),w_CG)
    
    ###this should go back to nu
    # param_tilde=log(c(beta,tilde_nu)) 
    # pred_mean_aug=R_times_z(param_tilde, have_noise=T, delta_x=delta_x_aug, z=w_aug[v_aug_sort$ix],
    #                         kernel_type=kernel_type)-tilde_nu*w_aug[v_aug_sort$ix]
    pred_mean_aug = IKF(beta=beta, tilde_nu=tilde_nu, 
                        delta_x=delta_x_aug, output=w_aug[v_aug_sort$ix], kernel_type=kernel_type)-tilde_nu*w_aug[v_aug_sort$ix]
    
    
    
    pred_mean_fast=pred_mean_aug[v_aug_sort_rev_ix][1:testing_n]
    #plot(testing_input, pred_mean_fast, type='l', main = paste(round(beta,2), round(tau,2), round(threshold_r,2)))
    
    
    if(compute_CI){
      c_v_star=rep(NA,testing_n)
      r0_v=abs(outer(testing_input,(v_all_vec),'-'))
      if(kernel_type=='exp'){
        r_v = exp(-beta*r0_v)
      }else if(kernel_type=='matern_5_2'){
        r_v = matern_5_2_funct(r0_v, beta)
      }
      
      N=length(num_neighbors_all_vec)
      A_r_v_rec=matrix(NA,N,testing_n)
      R_inv_A_r_v_rec=matrix(NA,N,testing_n)
      
      print("Computing the predictive variance ...")
      # #system.time(
      for(i in 1:testing_n){
        #print(i)
        
        A_r_v_i=A_times_x_particle(output=r_v[i,], A_all_v=A_v_all_vec,  num_neighbors_vec=num_neighbors_all_vec,
                                   D=D_y, N)
        A_r_v_rec[,i]=A_r_v_i
        #tol=sd(A_r_i)^2*0.01*N_tilde ##can make it smaller
        tol_interval=tol*10^{-14}
        R_inv_r_v_all=IKF_CG_particle_cell(param=log(c(beta,tau)), kernel_type=kernel_type, delta_x_all=delta_x_all, output=A_r_v_i,
                                           A_all_v=A_v_all_vec, sort_d_all_ix=sort_v_all$ix,
                                           sigma_2_vec=sigma_2_record, num_neighbors_vec=num_neighbors_all_vec, tilde_nu=tilde_nu,
                                           D=D_y, n_t_record=n_record[T_index_time], tol = tol_interval, maxIte = maxIte)
        R_inv_A_r_v_rec[,i]=R_inv_r_v_all[[1]]
        
        
        R_inv_r_v=R_inv_r_v_all[[1]]*tau
        r_R_inv_r_v=A_r_v_i%*%R_inv_r_v
        
        c_v_star[i]=1-r_R_inv_r_v
        
      }
      c_v_star=abs(c_v_star)
      
      #credible interval of the mean
      LB95 = pred_mean_fast+sqrt(as.numeric(sigma_2_est)*c_v_star)*qnorm(0.025)
      UB95 = pred_mean_fast+sqrt(as.numeric(sigma_2_est)*c_v_star)*qnorm(0.975)
    }
    
    
  }
  
  
  
  new("particle.est",
      D_y=D_y,
      # data = list(
      #   positions = input_pos_all,
      #   velocities = v_all,
      #   angles = theta_all
      # ),
      parameters = parameters,  # This contains the estimated parameters
      sigma_2_0_est = sigma_2_0_prop_est[1,1], 
      predictions = if(!is.null(testing_input)) {
        if(compute_CI) {
          list(mean = pred_mean_fast, lower95 = LB95, upper95 = UB95)
        } else {
          list(mean = pred_mean_fast)
        }
      } else {
        NULL
      },
      training_data =list(
        training_velocity = v_all_vec,
        A_v = A_v_all_vec,
        num_neighbors =num_neighbors_all_vec
      ),
      gp_weights = matrix(w_CG)
  )
  
}


extract_time_window <- function(data_obj, first_frame, last_frame) {
  # Input validation
  if (!inherits(data_obj, "particle.data")) {
    stop("Input must be a particle.data object")
  }
  
  # We need one more frame than the number of pairs
  if (first_frame < 1 || (last_frame + 1) > length(data_obj@px_list)) {
    stop("Time indices out of bounds")
  }
  
  if (first_frame >= last_frame) {
    stop("first_frame must be less than last_frame")
  }
  
  # Extract time window for each list
  # We need frames from first_frame to last_frame + 1 to have complete pairs
  px_list_new <- data_obj@px_list[first_frame:(last_frame + 1)]
  py_list_new <- data_obj@py_list[first_frame:(last_frame + 1)]
  vx_list_new <- data_obj@vx_list[first_frame:(last_frame + 1)]
  vy_list_new <- data_obj@vy_list[first_frame:(last_frame + 1)]
  
  # Handle optional theta_list
  theta_list_new <- if (!is.null(data_obj@theta_list)) {
    data_obj@theta_list[first_frame:(last_frame + 1)]
  } else {
    NULL
  }
  
  # Handle particle tracking
  particle_tracking_new <- if (!is.null(data_obj@particle_tracking)) {
    data_obj@particle_tracking[first_frame:last_frame]
  } else {
    NULL
  }
  
  # Extract relevant n_particles
  if (length(data_obj@n_particles) == 1) {
    # For simulation data
    n_particles_new <- data_obj@n_particles
  } else {
    # For experimental data
    n_particles_new <- data_obj@n_particles[first_frame:last_frame]  # Don't include last frame in n_particles
  }
  
  # Create new particle.data object
  new("particle.data",
      px_list = px_list_new,
      py_list = py_list_new,
      vx_list = vx_list_new,
      vy_list = vy_list_new,
      theta_list = theta_list_new,
      particle_tracking = particle_tracking_new,
      data_type = data_obj@data_type,
      n_particles = n_particles_new,
      T_time = last_frame - first_frame + 1,  # Number of pairs
      model = data_obj@model,
      sigma_0 = data_obj@sigma_0,
      radius = data_obj@radius,
      D_y = data_obj@D_y)
}



fit.particle.data = function(data, param, cut_r_max=1, est_param = TRUE, nx = NULL, ny = NULL,
                             # px_list = data@px_list, py_list = data@py_list, 
                             # vx_list = data@vx_list, vy_list = data@vy_list, theta_list = data@theta_list,
                             # n_t = data@n_particles,T_time = data@T_time, D_y=data@D_y, 
                             kernel_type = "matern_5_2", tilde_nu=0.1, tol=1e-6, maxIte=1000,
                             output=NULL, ho_output = NULL, 
                             testing_inputs, compute_CI = TRUE, num_interaction = (length(param)-1)/2,
                             data_type = data@data_type, model = data@model,  apolar_vicsek=FALSE, direction = NULL){
  if(data_type == "simulation"){
    if (!model %in% c("unnormalized_Vicsek", "two_interactions_Vicsek")){
      stop("Invalid model specified. Model must be either 'unnormalized_Vicsek' or 'two_interactions_Vicsek'")
    }
  }
  
  if(data_type == "experiment"){
    if (is.null(direction)){
      stop("Please specify the modeling direction ('x' or 'y')")
    }
  }
  
  
  if (!is.null(testing_inputs)) testing_inputs = matrix(testing_inputs, ncol = num_interaction)
  
  if(data_type == "simulation"){
    if(model == "unnormalized_Vicsek"){
      
      if (!is.null(testing_inputs)){
        testing_input = testing_inputs[,1]
      } else {
        testing_input = NULL
      }
      est = particle_interaction_est_unnormalized_Vicsek(data_obj=data, param=param, cut_r_max=cut_r_max, est_param=est_param, nx = nx, ny = ny,
                                                         kernel_type=kernel_type, tilde_nu=tilde_nu, tol=tol, maxIte=maxIte, 
                                                         output=output, ho_output = ho_output, testing_input = testing_input, compute_CI = compute_CI)
      est@model = model
    }else if(model == "two_interactions_Vicsek"){
      if (!is.null(testing_inputs)){
        testing_v_input = testing_inputs[,1]
        testing_d_input = testing_inputs[,2]
      } else{
        testing_v_input = NULL
        testing_d_input = NULL
      }
      est = particle_interaction_est_Vicsek_variation(data_obj=data, param=param, cut_r_max=cut_r_max, est_param=est_param, nx = nx, ny = ny,
                                                      kernel_type=kernel_type,tilde_nu=tilde_nu, tol=tol, maxIte=maxIte,
                                                      output=output, ho_output = ho_output, testing_v_input=testing_v_input, testing_d_input=testing_d_input, compute_CI = compute_CI)
      est@model = model
    }
  }else if(data_type == "experiment"){
    if (!is.null(testing_inputs)) {
      testing_input = testing_inputs[,1]
    }else{
      testing_input = NULL
    }
    est = particle_interaction_est_cell(data_obj=data, param=param, cut_r_max=cut_r_max, est_param=est_param, nx = nx, ny = ny, direction=direction, 
                                        kernel_type=kernel_type,tilde_nu=tilde_nu, tol=tol, maxIte=maxIte, 
                                        output=output, ho_output = ho_output, testing_input=testing_input, compute_CI = compute_CI, apolar_vicsek=apolar_vicsek)
  }
  
  est@data_type = data_type
  est@num_interaction = num_interaction
  
  return(est)
}

#' Multiply two numbers
#'
#' @param param a vector of parameters. The first parameter is the natural logarithm of the inverse range parameter in the kernel function. If the data contain noise, the second parameter is the logarithm of the nugget-variance ratio parameter.
#' @param have_noise a bool value. If it is true, it means the model contains a noise.
#' @param delta_x a vector with dimension (num_obs-1) x 1 for the differences between the sorted input locations.
#' @param output a vector with dimension num_obs x 1 for the observations at the sorted input locations.
#' @param kernel_type A character specifying the type of kernel.
#' @return A list where the first value is the natural logarithm of the determinant of the correlation matrix and the second value is the estimated sum of squares.
#' @keywords internal
#' @export

Get_log_det_S2_r <- function(param,have_noise,delta_x,output,kernel_type){
  return(Get_log_det_S2(param,have_noise,delta_x,output,kernel_type))
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

Construct_W0_matern_5_2 <- function(sigma2, lambda) {
    .Call('_FastGaSP_Construct_W0_matern_5_2', PACKAGE = 'FastGaSP', sigma2, lambda)
}

Construct_W0_matern_3_2 <- function(sigma2, lambda) {
    .Call('_FastGaSP_Construct_W0_matern_3_2', PACKAGE = 'FastGaSP', sigma2, lambda)
}

Construct_W0_exp <- function(sigma2, lambda) {
    .Call('_FastGaSP_Construct_W0_exp', PACKAGE = 'FastGaSP', sigma2, lambda)
}

Construct_G_matern_5_2 <- function(delta_x, lambda) {
    .Call('_FastGaSP_Construct_G_matern_5_2', PACKAGE = 'FastGaSP', delta_x, lambda)
}

Construct_G_matern_3_2 <- function(delta_x, lambda) {
    .Call('_FastGaSP_Construct_G_matern_3_2', PACKAGE = 'FastGaSP', delta_x, lambda)
}

Construct_G_exp <- function(delta_x, lambda) {
    .Call('_FastGaSP_Construct_G_exp', PACKAGE = 'FastGaSP', delta_x, lambda)
}

Construct_W_matern_5_2 <- function(sigma2, delta_x, lambda, W0) {
    .Call('_FastGaSP_Construct_W_matern_5_2', PACKAGE = 'FastGaSP', sigma2, delta_x, lambda, W0)
}

Construct_W_matern_3_2 <- function(sigma2, delta_x, lambda, W0) {
    .Call('_FastGaSP_Construct_W_matern_3_2', PACKAGE = 'FastGaSP', sigma2, delta_x, lambda, W0)
}

Construct_W_exp <- function(sigma2, delta_x, lambda, W0) {
    .Call('_FastGaSP_Construct_W_exp', PACKAGE = 'FastGaSP', sigma2, delta_x, lambda, W0)
}

Get_Q_K <- function(GG, W, C0, VV) {
    .Call('_FastGaSP_Get_Q_K', PACKAGE = 'FastGaSP', GG, W, C0, VV)
}

Get_log_det_S2 <- function(param, have_noise, delta_x, output, kernel_type) {
    .Call('_FastGaSP_Get_log_det_S2', PACKAGE = 'FastGaSP', param, have_noise, delta_x, output, kernel_type)
}

Get_L_inv_y <- function(GG, Q, K, output) {
    .Call('_FastGaSP_Get_L_inv_y', PACKAGE = 'FastGaSP', GG, Q, K, output)
}

Get_C_R_K_Q <- function(index, GG, W, C0, VV) {
    .Call('_FastGaSP_Get_C_R_K_Q', PACKAGE = 'FastGaSP', index, GG, W, C0, VV)
}

Get_m_a_pred <- function(index, output_vec, GG, K) {
    .Call('_FastGaSP_Get_m_a_pred', PACKAGE = 'FastGaSP', index, output_vec, GG, K)
}

Get_S_KK <- function(index, GG, C, R) {
    .Call('_FastGaSP_Get_S_KK', PACKAGE = 'FastGaSP', index, GG, C, R)
}

Get_s_1st <- function(m, a, C, KK) {
    .Call('_FastGaSP_Get_s_1st', PACKAGE = 'FastGaSP', m, a, C, KK)
}

Kalman_smoother <- function(param, have_noise, index_obs, delta_x_all, output, sigma_2_hat, kernel_type) {
    .Call('_FastGaSP_Kalman_smoother', PACKAGE = 'FastGaSP', param, have_noise, index_obs, delta_x_all, output, sigma_2_hat, kernel_type)
}

Sample_KF <- function(GG, W, C0, VV, kernel_type, sample_type) {
    .Call('_FastGaSP_Sample_KF', PACKAGE = 'FastGaSP', GG, W, C0, VV, kernel_type, sample_type)
}

Sample_KF_post <- function(index_obs, C_R_K_Q, W0, GG, W, VV, output, kernel_type, sample_type) {
    .Call('_FastGaSP_Sample_KF_post', PACKAGE = 'FastGaSP', index_obs, C_R_K_Q, W0, GG, W, VV, output, kernel_type, sample_type)
}

Get_L_t_y <- function(GG, Q, K, output) {
    .Call('_FastGaSP_Get_L_t_y', PACKAGE = 'FastGaSP', GG, Q, K, output)
}

Get_L_y <- function(GG, Q, K, output) {
    .Call('_FastGaSP_Get_L_y', PACKAGE = 'FastGaSP', GG, Q, K, output)
}

Get_L_t_inv_y <- function(GG, Q, K, output) {
    .Call('_FastGaSP_Get_L_t_inv_y', PACKAGE = 'FastGaSP', GG, Q, K, output)
}

Get_R_y <- function(GG, Q, K, output) {
    .Call('_FastGaSP_Get_R_y', PACKAGE = 'FastGaSP', GG, Q, K, output)
}

Get_Y_minus_a_1_scaled_matrix_2d <- function(output_KF, GG, Q, K) {
    .Call('_FastGaSP_Get_Y_minus_a_1_scaled_matrix_2d', PACKAGE = 'FastGaSP', output_KF, GG, Q, K)
}

F_Funct <- function(A_cur, G) {
    .Call('_FastGaSP_F_Funct', PACKAGE = 'FastGaSP', A_cur, G)
}

Get_G_log_det_cov <- function(param, output, delta_x, d, kernel_type) {
    .Call('_FastGaSP_Get_G_log_det_cov', PACKAGE = 'FastGaSP', param, output, delta_x, d, kernel_type)
}

KF_cpp_eigen <- function(F_t, V_t, G_t, W_t, y) {
    .Call('_FastGaSP_KF_cpp_eigen', PACKAGE = 'FastGaSP', F_t, V_t, G_t, W_t, y)
}

cubic_solver <- function(p) {
    .Call('_FastGaSP_cubic_solver', PACKAGE = 'FastGaSP', p)
}

fmou_cpp <- function(output, d, M = 50L, threshold = 1e-4, est_U0 = TRUE, est_sigma0_2 = TRUE, track_iterations = FALSE, track_neg_log_lik = FALSE, U0 = NULL, U_init = NULL, rho_init = NULL, sigma2_init = NULL, sigma0_2 = NULL) {
    .Call('_FastGaSP_fmou_cpp', PACKAGE = 'FastGaSP', output, d, M, threshold, est_U0, est_sigma0_2, track_iterations, track_neg_log_lik, U0, U_init, rho_init, sigma2_init, sigma0_2)
}

matern_5_2_funct <- function(d, beta_i) {
    .Call('_FastGaSP_matern_5_2_funct', PACKAGE = 'FastGaSP', d, beta_i)
}

rcppeigen_get_chol <- function(R) {
    .Call('_FastGaSP_rcppeigen_get_chol', PACKAGE = 'FastGaSP', R)
}

F_Funct_Dev_Large_k <- function(A_cur, UD) {
    .Call('_FastGaSP_F_Funct_Dev_Large_k', PACKAGE = 'FastGaSP', A_cur, UD)
}

Get_B_U_V_Large_k <- function(A_cur, UD) {
    .Call('_FastGaSP_Get_B_U_V_Large_k', PACKAGE = 'FastGaSP', A_cur, UD)
}

Y_Funct <- function(A_cur, B_U_V, tau) {
    .Call('_FastGaSP_Y_Funct', PACKAGE = 'FastGaSP', A_cur, B_U_V, tau)
}

F_Funct_Dev <- function(A_cur, G) {
    .Call('_FastGaSP_F_Funct_Dev', PACKAGE = 'FastGaSP', A_cur, G)
}

Get_B_U_V <- function(A_cur, G) {
    .Call('_FastGaSP_Get_B_U_V', PACKAGE = 'FastGaSP', A_cur, G)
}

Optimization_Stiefel_Manifold <- function(A_ini, G, max_iter) {
    .Call('_FastGaSP_Optimization_Stiefel_Manifold', PACKAGE = 'FastGaSP', A_ini, G, max_iter)
}

A_t_times_x_particle <- function(output, A_all_v, num_neighbors_vec, D_y, N_tilde) {
    .Call('_FastGaSP_A_t_times_x_particle', PACKAGE = 'FastGaSP', output, A_all_v, num_neighbors_vec, D_y, N_tilde)
}

A_times_x_particle <- function(output, A_all_v, num_neighbors_vec, D, N) {
    .Call('_FastGaSP_A_times_x_particle', PACKAGE = 'FastGaSP', output, A_all_v, num_neighbors_vec, D, N)
}

IKF_CG_particle <- function(param, kernel_type, delta_x_all, output, A_all_v, sort_d_all_ix, num_neighbors_vec, tilde_nu, D, N, tol = 1e-6, maxIte = 1000L) {
    .Call('_FastGaSP_IKF_CG_particle', PACKAGE = 'FastGaSP', param, kernel_type, delta_x_all, output, A_all_v, sort_d_all_ix, num_neighbors_vec, tilde_nu, D, N, tol, maxIte)
}

IKF_CG_particle_two_interact <- function(param1, param2, kernel_type1, kernel_type2, delta_x_all1, delta_x_all2, A_all_v1, A_all_v2, sort_d_all_ix1, sort_d_all_ix2, num_neighbors_vec1, num_neighbors_vec2, output, tilde_nu, D, N, tol = 1e-6, maxIte = 1000L) {
    .Call('_FastGaSP_IKF_CG_particle_two_interact', PACKAGE = 'FastGaSP', param1, param2, kernel_type1, kernel_type2, delta_x_all1, delta_x_all2, A_all_v1, A_all_v2, sort_d_all_ix1, sort_d_all_ix2, num_neighbors_vec1, num_neighbors_vec2, output, tilde_nu, D, N, tol, maxIte)
}

IKF_CG_particle_cell <- function(param, kernel_type, delta_x_all, output, A_all_v, sort_d_all_ix, sigma_2_vec, num_neighbors_vec, tilde_nu, D, n_t_record, tol = 1e-6, maxIte = 1000L) {
    .Call('_FastGaSP_IKF_CG_particle_cell', PACKAGE = 'FastGaSP', param, kernel_type, delta_x_all, output, A_all_v, sort_d_all_ix, sigma_2_vec, num_neighbors_vec, tilde_nu, D, n_t_record, tol, maxIte)
}

\name{show.particle.data}
\alias{show.particle.data}
\alias{show,particle.data-method}
\title{Show method for particle data class}
\description{
Display method for objects of class \code{particle.data}, which prints a summary of the key parameters and characteristics of the particle system based on its data type (simulation or experimental).
}
\usage{
\S4method{show}{particle.data}(object)
}
\arguments{
\item{object}{An object of class \code{particle.data}.}
}
\details{
This method displays essential information about the particle system. The output varies based on the data type:

For simulation data:
\itemize{
\item Type of particle interaction model
\item Number of time steps
\item Number of particles in the system
\item Noise variance parameter (sigma_0)
\item Interaction radius between particles
}

For experimental data:
\itemize{
\item Number of time steps
\item Average number of particles across all time steps
}
}
\references{
Vicsek, T., Czirok, A., Ben-Jacob, E., Cohen, I., & Shochet, O. (1995). \emph{Novel type of phase transition in a system of self-driven particles}, \emph{Physical Review Letters}, \bold{75}(6), 1226.

Chat'{e}, H., Ginelli, F., Gr'{e}goire, G., Peruani, F., & Raynaud, F. (2008). \emph{Modeling collective motion: variations on the Vicsek model}, \emph{The European Physical Journal B}, \bold{64}(3), 451-456.

Fang, X., & Gu, M. (2024). \emph{The inverse Kalman filter}. arXiv:2407.10089.
}
\seealso{
\code{\link{particle.data-class}} for the full class description,
\code{\link{simulate_particle}} for creating simulation data objects,
\code{\link{trajectory_data}} for creating experimental data objects
}
\keyword{methods}

\name{fit}
\alias{fit}
\title{Fit Particle Interaction Models}
\description{
Generic function for fitting particle interaction models to data.
}
\usage{
fit(object, ...)
}
\arguments{
\item{object}{Object containing data to be fit}
\item{...}{Additional arguments passed to fitting methods}
}
\seealso{
\code{\link{fit.particle.data}} for fitting particle data models
}
\keyword{methods}

\name{Construct_W_matern_3_2}
\alias{Construct_W_matern_3_2}
\title{ 
The conditional covariance matrix for  matern covariance with roughness parameter 1.5
}
\description{
The conditional covariance matrix of the state in the dynamic linear model when kernel is the matern covariance with roughness parameter 1.5.
}
\usage{
Construct_W_matern_3_2(sigma2,delta_x,lambda,W0)
}
\arguments{
  \item{sigma2}{the variance parameter.
}
  \item{delta_x}{the distance between the sorted input.
}
  \item{lambda}{the transformed range parameter. }
    \item{W0}{the covariance matrix of the stationary distribution of the state. }

}


\value{
W matrix. 
}

\references{
%% ~put references to the literature/web site here ~

Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models}.  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.

M. Gu, Y. Xu (2019), \emph{fast nonseparable Gaussian stochastic process with application to methylation level interpolation}.  \emph{Journal of Computational and Graphical Statistics}, In Press, arXiv:1711.11501.

Campagnoli P, Petris G, Petrone S. (2009), \emph{Dynamic linear model with R}. Springer-Verlag New York.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}

\keyword{internal}
%\note{
%%  ~~further notes~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~

%\seealso{
%% ~~objects to See Also as \code{\link{help}}, ~~~
%}

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
%\keyword{ ~kwd1 }
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line

\name{fit.fmou}
\alias{fit.fmou}
\alias{fit.fmou,fmou-method}
\title{ 
The fast EM algorithm of multivariate Ornstein-Uhlenbeck processes}


\description{
This function implements an efficient EM algorithm to estimate the parameters in the FMOU model, a latent factor model with a fixed or estimated orthogonal factor loading matrix, where each latent factor is modeled as an O-U (Ornstein-Uhlenbeck) process.
}


\usage{
\S4method{fit.fmou}{fmou}(object, M=50, threshold=1e-4,
         track_iterations=FALSE,track_neg_log_lik=FALSE,
         U_init=NULL, rho_init=NULL, sigma2_init=NULL, d_ub=NULL)
}


\arguments{
  \item{object}{an objecft of  class \code{fmou.}}
  \item{M}{number of iterations in the EM algorithm, default is 50.}
  \item{threshold}{stopping criteria with respect to predictive mean of observations, default is 1e-4.}
  \item{track_iterations}{a bool value, default is \code{FALSE}. If \code{TRUE}, the estimations in each EM iteration will be recorded and returned. }
  \item{track_neg_log_lik}{a bool value, default is \code{FALSE}. If \code{TRUE}, the negative log marginal likelihood of the output in each EM iteration will be recorded and returned.}
  \item{U_init}{user-specified initial factor loading matrix in the EM algorithm. Default is \code{NULL}. The dimension is k*d, where k is the length of observations at each time step and d is the number of latent factors.}
  \item{rho_init}{user-specified initial correlation parameters in the EM algorithm. Default is \code{NULL}. The length is equal to the number of latent factors.}
  \item{sigma2_init}{user-specified initial variance parameters in the EM algorithm. Default is \code{NULL}. The length is equal to the number of latent factors.}
  \item{d_ub}{upper bound of d when d is estimated. Default is null.}
}

\value{
  \item{output}{the observation matrix.}
  \item{U}{the estimated (or fixed) factor loading matrix.}
  \item{post_z_mean}{the posterior mean of latent factors.}
  \item{post_z_var}{the posterior variance of latent factors.}
  \item{post_z_cov}{the posterior covariance between two consecutive time steps of a latent process.}
  \item{mean_obs}{the predictive mean of the observations.}
  \item{mean_obs_95lb}{the lower bound of the 95\% posterior credible intervals of predictive mean.}
  \item{mean_obs_95ub}{the upper bound of the 95\% posterior credible intervals of predictive mean.}
  \item{sigma0_2}{estimated variance of noise.}
  \item{rho}{estimated correlation parameters.}
  \item{sigma2}{estimated variance parameters}
  \item{num_iterations}{number of iterations in the EM algorithm.}
  \item{d}{the estimated (or fixed) number of latent factors.}
  \item{record_sigma0_2}{estimated variance of noise in each iteration.}
  \item{record_rho}{estimated correlation parameters in each iteration.}
  \item{record_sigma2}{estimation variance parameters in each iteration.}
}

\references{

Lin, Y., Liu, X., Segall, P., & Gu, M. (2025). Fast data inversion for high-dimensional dynamical systems from noisy measurements. arXiv preprint arXiv:2501.01324.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}


\examples{

## generate simulated data
library(FastGaSP)
library(rstiefel)

d = 5  # number of latent factors
k = 20 # length of observation at each time step
n = 500 # number time step
noise_level = 1 # variance of noise

U = rustiefel(k, k) # factor loading matrix
z = matrix(NA, d, n)
sigma_2 = runif(d, 0.5, 1)
rho = runif(d, 0.95, 1)
for(l in 1:d){
  R = matrix(NA, n, n)
  diag(R) = 1
  for(ir in 1:n){
    for(ic in 1:n){
      R[ir, ic] = rho[l]^(abs(ir-ic)) * R[ir, ir]
    }
  }
  R = (sigma_2[l]/(1-rho[l]^2) )* R
  z[l, ] = t(chol(R)) \%*\% rnorm(n)
}

signal = U[,1:d] \%*\% z
y = signal + matrix(rnorm(n*k,mean=0,sd=sqrt(noise_level)),k,n)

##constucting the fmou.model
fmou.model=fmou(output=y, d=d, est_U0=TRUE, est_sigma0_2=TRUE)

## estimate the parameters
em_alg <- fit.fmou(fmou.model, M=500)

## root mean square error (RMSE) of predictive mean of observations
sqrt(mean((em_alg$mean_obs-signal)^2))

## standard deviation of (truth) mean of observations
sd(signal)

## estimated variance of noise
em_alg$sigma0_2

}

\name{Get_L_inv_y}
\alias{Get_L_inv_y}
\title{ 
The multiplication of the inverse of L with y
}
\description{
This function computes the product of the inverse of the L matrix and the output vector, where the L matrix is the Cholesky decomposition of the correlation matrix. Instead of computing the Cholesky matrix, we compute it using the forward filtering algorithm. 
}
\usage{
Get_L_inv_y(GG,Q,K,output)
}
\arguments{
  \item{GG}{a list of matrices defined in the dynamic linear model. 
}
%%     ~~Describe \code{response} here~~
    \item{Q}{a vector defined in the dynamic linear model.    }
    \item{K}{a matrix defined in the filtering algorithm for the dynamic linear model.  }
    \item{output}{a vector of output.}

}


\value{
A vector representing the product of the inverse of the L matrix and the output vector, where the L matrix is the Cholesky decomposition of the correlation matrix. 
}

\references{
%% ~put references to the literature/web site here ~

Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models}.  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.

M. Gu, Y. Xu (2019), \emph{fast nonseparable gaussian stochastic process with application to methylation level interpolation}.  \emph{Journal of Computational and Graphical Statistics}, In Press, arXiv:1711.11501.

Campagnoli P, Petris G, Petrone S. (2009), \emph{Dynamic linear model with R}. Springer-Verlag New York.


}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}
\seealso{\code{\link{Get_Q_K}} for more details about \eqn{Q} vector and \eqn{K} matrix,
\code{\link{Get_L_t_y}} for \eqn{L^T y} computation, 
\code{\link{Get_L_y}} for \eqn{L y} computation, 
\code{\link{Get_L_t_inv_y}} for \eqn{(L^T)^{-1}y} computation.
%% ~~objects to See Also as \code{\link{help}}, ~~~
}

\keyword{internal}
%\note{
%%  ~~further notes~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~

%\seealso{
%% ~~objects to See Also as \code{\link{help}}, ~~~
%}

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
%\keyword{ ~kwd1 }
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line

\name{Get_log_det_S2}
\alias{Get_log_det_S2}

\title{ 
the natural logarithm of the determinant of the correlation matrix and the estimated sum of squares in the exponent of the profile likelihood
}
\description{
This function computes the natural logarithm of the determinant of the correlation matrix and the estimated sum of squares for computing the profile likelihood. 
}
\usage{
Get_log_det_S2(param,have_noise,delta_x,output,kernel_type)
}
\arguments{
  \item{param}{a vector of parameters. The first parameter is the natural logarithm of the inverse range parameter in the kernel function. If the data contain noise, the second parameter is the logarithm of the nugget-variance ratio parameter.
}
  \item{have_noise}{a bool value. If it is true, it means the model contains a noise. }
%%     ~~Describe \code{response} here~~
    \item{delta_x}{a vector with dimension (num_obs-1) x 1 for the differences between the sorted input locations.}
    \item{output}{a  vector with dimension num_obs x 1 for the observations at the sorted input locations.}
    \item{kernel_type}{A \code{character} specifying the type of kernel.}


}


\value{
A list where the first value is the natural logarithm of the determinant of the correlation matrix and the second value is the estimated sum of squares.
}

\references{
%% ~put references to the literature/web site here ~

Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models},  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.

M. Gu, Y. Xu (2017), \emph{Nonseparable Gaussian stochastic process: a unified
view and computational strategy}, arXiv:1711.11501.

M. Gu, X. Wang and J.O. Berger (2018), \emph{Robust Gaussian Stochastic Process Emulation}, \emph{Annals of Statistics}, \bold{46}, 3038-3066.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}
\seealso{\code{\link{log_lik}} for more details about the profile likelihood.
%% ~~objects to See Also as \code{\link{help}}, ~~~
}
%\keyword{documentation}
%\concept{internal}
\keyword{internal}
%\concept{internal}
%\docType{data}

%\note{
%%  ~~further notes~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~

%\seealso{
%% ~~objects to See Also as \code{\link{help}}, ~~~
%}

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
%\keyword{ ~kwd1 }
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line

\name{trajectory_data}
\alias{trajectory_data}
\title{Convert experimental particle tracking data to particle.data object}
\description{
Processes experimental particle tracking data and creates a standardized particle.data object. This function handles time series data with varying numbers of particles across time steps and maintains particle identity tracking between consecutive frames.
}
\usage{
trajectory_data(particle_data)
}
\arguments{
\item{particle_data}{A data frame containing particle tracking data with the following required columns:
\itemize{
\item \code{px}, \code{py}: Particle positions in x and y coordinates
\item \code{vx}, \code{vy}: Particle velocities in x and y directions
\item \code{time}: Time step identifier (integer). Must be consecutive integers starting from the minimum time value
\item \code{particleID}: Unique particle identifier for tracking across frames
}
}
}
\value{
Returns an S4 object of class \code{\linkS4class{particle.data}} with:
  \describe{
    \item{px_list, py_list}{Lists of particle x and y positions at each time step}
    \item{vx_list, vy_list}{Lists of particle x and y velocities at each time step}
    \item{theta_list}{List of particle angles computed from velocities}
    \item{particle_tracking}{List of data frames containing particle mappings between consecutive frames}
    \item{data_type}{"experiment"}
    \item{n_particles}{Vector recording number of particles at each time step}
    \item{T_time}{Total number of time steps}
    \item{D_y}{Dimension of the output space (set to 1)}
  }
}

\examples{
# Create sample tracking data
sample_data <- data.frame(time = rep(1:3, each = 5),
                          particleID = rep(1:5, 3),
                          px = rnorm(15),py = rnorm(15),
                          vx = rnorm(15),vy = rnorm(15)
)
# Convert to particle.data object
traj <- trajectory_data(sample_data)
# Display summary
show(traj)
}
\seealso{
\code{\link{particle.data-class}} for details on the particle.data class structure
}

\name{predictobj.fgasp-class}
\alias{predictobj.fgasp-class}
\alias{predictobj.fgasp}
\docType{class}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Predictive results for the Fast GaSP class }
\description{S4 class for prediction for a Fast GaSP model with or without a noise.}

\section{Objects from the Class}{
Objects of this class are created and initialized with the function \code{\link{predict}} that computes the prediction and the uncertainty quantification.}

\section{Slots}{
	 \describe{
	     \item{\code{num_testing}:}{object of class \code{integer}. Number of testing inputs.}
    \item{\code{testing_input}:}{object of class \code{vector}. The testing input locations.}
      \item{param}{a vector of parameters. The first parameter is the natural logarithm of the inverse range parameter in the kernel function. If the data contain noise, the second parameter is the logarithm of the nugget-variance ratio parameter.
}
    \item{\code{mean}:}{object of class \code{vector}. The predictive mean at testing inputs.}
        \item{\code{var}:}{object of class \code{vector}. The predictive variance at testing inputs. If the \code{var_data} is true, the predictive variance of the data is calculated. Otherwise, the predictive variance of the mean is calculated. }
        \item{\code{var_data}:}{object of class \code{logical}. If the \code{var_data} is true, the predictive variance of the data is calculated for \code{var}. Otherwise, the predictive variance of the mean is calculated for \code{var}. }
        \item{\code{index_x_all}:}{object of class \code{vector} or \code{NULL}. A vector of all index locations. }
        \item{\code{delta_x_all}:}{object of class \code{vector} or \code{NULL}. A vector of distances between all locations. }
        \item{\code{testing_loc}:}{object of class \code{vector} or \code{NULL}. A vector of location for  the test inputs in the combined sorted inputs. }
        \item{\code{input_all_sorted_ix}:}{object of class \code{vector} or \code{NULL}. A vector of sorted indices for all inputs. }
        \item{\code{testing_input_sorted_ix}:}{object of class \code{vector} or \code{NULL}. A vector of sorted indiced for test inputs. }

    }
}




\references{

Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models},  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.

M. Gu, Y. Xu (2017), \emph{Nonseparable Gaussian stochastic process: a unified
view and computational strategy}, arXiv:1711.11501.

M. Gu, X. Wang and J.O. Berger (2018), \emph{Robust Gaussian Stochastic Process Emulation}, \emph{Annals of Statistics}, \bold{46}, 3038-3066.


}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}


%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{\code{\link{predict.fgasp}} for more details about how to do prediction for a \code{fgasp} object.
%% ~~objects to See Also as \code{\link{help}}, ~~~
}

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{classes}
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line

\name{Get_R_y}
\alias{Get_R_y}
\title{ 
The multiplication of R with y
}
\description{
This function computes the product of the R matrix and the output vector, where R is the correlation matrix for a dynamic linear model (DLM). Instead of explicitly forming the Cholesky decomposition of R, this function computes the product as \eqn{L (L^T y)}, where \eqn{L} is the Cholesky decomposition of R. This is achieved using the forward filtering algorithm for efficient computation.
}
\usage{
Get_R_y(GG, Q, K, output)
}
\arguments{
  \item{GG}{a list of matrices defined in the dynamic linear model.}
  
  \item{Q}{a vector defined in the dynamic linear model.} % Each element is the variance of the corresponding state.
  
  \item{K}{a matrix defined in the filtering algorithm for the dynamic linear model.}
  
  \item{output}{a vector of observations.}
}
\value{
A vector representing the product of the R matrix and the output vector, where \eqn{R} is the correlation matrix for a dynamic linear model.
}
\references{
Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal Gaussian process regression models}. \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop}, 379-384.

Fang, X., & Gu, M. (2024). \emph{The inverse Kalman filter}. arXiv:2407.10089.

M. Gu, Y. Xu (2019), \emph{Fast nonseparable Gaussian stochastic process with application to methylation level interpolation}. \emph{Journal of Computational and Graphical Statistics}, In Press, arXiv:1711.11501.

Campagnoli P, Petris G, Petrone S. (2009), \emph{Dynamic linear model with R}. Springer-Verlag New York.

}
\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}
\seealso{
\code{\link{Get_Q_K}} for more details on \eqn{K} and \eqn{Q} matrices, 
\code{\link{Get_L_inv_y}} for \eqn{L^{-1}y} computation, 
\code{\link{Get_L_t_y}} for \eqn{L^T y} computation, 
\code{\link{Get_L_y}} for \eqn{L y} computation, 
\code{\link{Get_L_t_inv_y}} for \eqn{(L^T)^{-1}y} computation.
}
\keyword{internal}

\name{show.fgasp}
\alias{show,fgasp-method}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
%%  ~~function to do ... ~~
Show an \code{fgasp} object.
}
\description{
%%  ~~ A concise (1-5 lines) description of what the function does. ~~
Function to print the \code{fgasp} object. 
}
\usage{
\S4method{show}{fgasp}(object)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{object}{an object of  class \code{fgasp}.}
}

%\details{
%%  ~~ If necessary, more details than the description above ~~
%Provide here some details.
%}

\references{
Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models},  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.

M. Gu, Y. Xu (2017), \emph{Nonseparable Gaussian stochastic process: a unified
view and computational strategy}, arXiv:1711.11501.

M. Gu, X. Wang and J.O. Berger (2018), \emph{Robust Gaussian Stochastic Process Emulation}, \emph{Annals of Statistics}, \bold{46}, 3038-3066.
}

\author{
%%  ~~who you are~~
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}

\examples{

#-------------------------------------
# Example: a simple example with noise 
#-------------------------------------

y_R<-function(x){
  cos(2*pi*x)
}

###let's test for 2000 observations
set.seed(1)
num_obs=2000
input=runif(num_obs)

output=y_R(input)+rnorm(num_obs,mean=0,sd=0.1)

##constucting the fgasp.model
fgasp.model=fgasp(input, output)
show(fgasp.model)


}

%\note{
%%  ~~further notes~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~

%\seealso{
%% ~~objects to See Also as \code{\link{help}}, ~~~
%}


% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
%\keyword{ ~kwd1 }% use one of  RShowDoc("KEYWORDS")
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line

\name{IKF_CG_particle}
\alias{IKF_CG_particle}
\title{IKF-CG algorithm for one-interaction physical model with 1D output}
\description{
  Implements the IKF-CG (Inverse Kalman Filter - Conjugate Gradient) algorithm for the one-interaction physical model with 1D output. This function provides an efficient computational method for large-scale particle systems.
}

\usage{
IKF_CG_particle(param, kernel_type, delta_x_all, output, A_all_v, 
                sort_d_all_ix, num_neighbors_vec, tilde_nu,
                D, N, tol = 1e-6, maxIte = 1000)
}

\arguments{
  \item{param}{A numeric vector containing model parameters. The first element is the log of beta (inverse range parameter), and the second element is the log of tau (variance ratio parameter).}
  
  \item{kernel_type}{A string specifying the kernel type. Must be either "matern_5_2" or "exp".}
  
  \item{delta_x_all}{A numeric vector of successive differences of sorted inputs.}
  
  \item{output}{A numeric vector representing the output values.}
  
  \item{A_all_v}{A numeric vector containing the non-zero entries in matrix A.}
  
  \item{sort_d_all_ix}{An integer vector containing sorted indices for distances.}
  
  \item{num_neighbors_vec}{An integer vector specifying the number of neighbors for each particle.}
  
  \item{tilde_nu}{A numeric value representing the stabilizing parameter.}
  
  \item{D}{An integer specifying the dimension of the output vector per particle.}
  
  \item{N}{An integer specifying the total dimension of the output vector.}
  
  \item{tol}{A numeric value specifying the convergence tolerance. Default is 1e-6.}
  
  \item{maxIte}{An integer specifying the maximum number of iterations. Default is 1000.}
}


\value{
  Returns a list containing three elements:
  \item{x}{A numeric vector containing the estimated state variables.}
  \item{resid}{A numeric vector containing the residuals at each iteration.}
  \item{ite}{An integer indicating the number of iterations performed.}
}



\references{
Fang, X., & Gu, M. (2024). \emph{The inverse Kalman filter}. arXiv:2407.10089.
}



\keyword{internal}

\name{extract_time_window}
\alias{extract_time_window}
\title{Extract time window from particle data}
\description{
Extracts a specified time window from a particle.data object while preserving all relevant tracking information and parameters. Works with both simulation and experimental data.
}
\usage{
extract_time_window(data_obj, first_frame, last_frame)
}
\arguments{
\item{data_obj}{An object of class \code{\linkS4class{particle.data}}.}
\item{first_frame}{Integer specifying the first frame to include (must be >= 1).}
\item{last_frame}{Integer specifying the last frame to include (must be less than the total number of frames).}
}
\value{
Returns a new \code{\linkS4class{particle.data}} object containing:
\describe{
\item{px_list, py_list}{Position data for the selected time window}
\item{vx_list, vy_list}{Velocity data for the selected time window}
\item{theta_list}{Angle data if present in original object}
\item{particle_tracking}{Tracking information for the selected frames (experimental data)}
\item{data_type}{Original data type ("simulation" or "experiment")}
\item{n_particles}{Particle counts (constant for simulation, time series for experimental)}
\item{T_time}{Number of time steps in the extracted window}
\item{model, sigma_0, radius}{Original model parameters (for simulation data)}
\item{D_y}{Original dimension of the output space}
}
}

\name{particle.est-class}
\alias{particle.est-class}
\alias{particle.est}
\docType{class}
\title{Particle interaction estimation class}
\description{S4 class for storing estimated parameters and predictions for particle interaction models. }

\section{Objects from the Class}{
Objects of this class are created by the \code{\link{fit.particle.data}} (via \code{fit}) method when applied to \code{\linkS4class{particle.data}} objects to estimate interaction parameters and make predictions.
}

\section{Slots}{
    \describe{
        \item{\code{data_type}:}{Object of class \code{character}. Specifies the type of data ("simulation" or "experiment").}
        \item{\code{model}:}{Object of class \code{characterOrNULL}. Specifies the model type for simulation data (e.g., "Vicsek" or "two_interactions_Vicsek"). NULL for experimental data.}
        \item{\code{D_y}:}{Object of class \code{numeric}. Dimension of the output space.}
        \item{\code{num_interaction}:}{Object of class \code{numeric}. Number of interactions.}
        \item{\code{parameters}:}{Object of class \code{numeric}. Vector of estimated parameters with length 2*D_y + 1:
          \itemize{
          \item First D_y elements: beta (inverse range parameters)
          \item Next D_y elements: tau (variance-noise ratios)
          \item Last element: interaction radius
          }
        }
        \item{\code{sigma_2_0_est}:}{Object of class \code{numeric}. Estimated noise variance.}
        \item{\code{predictions}:}{object of class \code{listOrNULL}. Contains predicted means and 95\% confidence intervals (lower and upper bounds) for the particle interactions if testing inputs are given.}
        \item{\code{training_data}:}{Object of class \code{list}. Contains the training data used in the GP model, obtained using the estimated interaction radius.}
        \item{\code{gp_weights}:}{Object of class \code{matrix}. Contains the weights from the GP computation (A^T_j Sigma_y^(-1) y) used for prediction, with each column corresponding to a type of interaction j.}
    }
}

\section{Methods}{
    \describe{
        \item{show:}{Method for displaying summary information about the estimated parameters.}
    }
}

\references{
Fang, X., & Gu, M. (2024). \emph{The inverse Kalman filter}. arXiv:2407.10089.
}

\seealso{
  \code{\link{fit.particle.data}} for more details about how to create a \code{particle.est} object.
  \code{\link{particle.data-class}} for the input data structure
}

\keyword{classes}

\name{IKF}
\alias{IKF}
\title{
Inverse Kalman Filter - The multiplication of R with y with given kernel type
}
\description{
This function computes the product of the R matrix and the output vector for a given kernel type, where R is the correlation matrix for a dynamic linear model (DLM). Instead of explicitly forming the Cholesky decomposition of R, this function computes the product as \eqn{L (L^T y)}, where \eqn{L} is the Cholesky decomposition of R. This is achieved using the forward filtering algorithm for efficient computation.
}
\usage{
IKF(beta, tilde_nu, delta_x, output, kernel_type='matern_5_2')
}

\arguments{
  \item{beta}{A scalar representing the inverse range parameter in the kernel function.}
  \item{tilde_nu}{A numerical value representing the nugget or the stabilizing parameter.}
  \item{delta_x}{A numeric vector of successive differences of inputs.}
  \item{output}{The output vector for which the product with R is computed.}
  \item{kernel_type}{A string specifying the kernel type. Must be either "matern_5_2" or "exp".}
}

\value{
A vector representing the product of the R matrix and the output vector
}
\references{
Fang, X., & Gu, M. (2024). \emph{The inverse Kalman filter}. arXiv:2407.10089.
}
\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}

\seealso{
\code{\link{Get_R_y}} for computing the product of R and y with G, Q, and K matrices.
}
\examples{
# Helper function for Matern 2.5 correlation
matern25_correlation <- function(d, beta) {
  #' Compute Matern 2.5 correlation matrix given distance matrix and beta parameter
  #'
  #' @param d Distance matrix or array of any dimensions
  #' @param beta Inverse range parameter 
  #'
  #' @return Correlation matrix with same dimensions as d
  #'
  #' @details
  #' Formula: k(d) = (1 + sqrt(5)*beta*d + 5*beta^2*d^2/3) * exp(-sqrt(5)*beta*d)
  
  # Compute sqrt(5)*beta*d
  sqrt5_beta_d <- sqrt(5) * beta * d
  
  # Compute the correlation
  R <- (1 + sqrt5_beta_d + (5 * beta^2 * d^2) / 3) * exp(-sqrt5_beta_d)
  
  return(R)
}

# Example: Comparing IKF with direct matrix computation
n <- 2000                           # number of observations
input <- sort(runif(n))             # sorted input points
delta_x <- input[-1] - input[-n]    # consecutive differences
u <- rnorm(n)                       # random output vector
beta <- 10                          # inverse range parameter

# Test 1: Noise-free scenario
# Non-robust IKF (no stabilization parameter)
x_non_robust_IKF <- IKF(beta = beta, tilde_nu = 0, delta_x = delta_x, 
                        output = u, kernel_type = 'matern_5_2')

# Robust IKF (with stabilization parameter)
tilde_nu <- 0.1  # stabilization parameter
x_IKF <- IKF(beta = beta, tilde_nu = tilde_nu, delta_x = delta_x, 
             output = u, kernel_type = 'matern_5_2') - tilde_nu * u

# Direct matrix computation for comparison
R0 <- abs(outer(input, input, '-'))  # distance matrix
R <- matern25_correlation(R0, beta)  # correlation matrix
x_direct <- R \%*\% u          

# Compare results (should be nearly identical)
cat("Maximum difference (non-robust IKF vs direct):", 
    max(abs(x_direct - x_non_robust_IKF)), "\n")
cat("Maximum difference (robust IKF vs direct):", 
    max(abs(x_direct - x_IKF)), "\n")

# Test 2: With noise
V <- 0.2  # noise variance
x_IKF_noisy <- IKF(beta = beta, tilde_nu = V, delta_x = delta_x, 
                   output = u, kernel_type = 'matern_5_2')

# Direct computation with noise
x_direct_noisy <- (R + V * diag(n)) \%*\% u

# Compare results
cat("Maximum difference (IKF vs direct):", 
    max(abs(x_direct_noisy - x_IKF_noisy)), "\n")

}

\name{A_t_times_x_particle}
\alias{A_t_times_x_particle}
\title{Transpose matrix-vector multiplication for particle systems}
\description{
  Performs the matrix-vector multiplication A^T*x for particle systems, where A is a sparse matrix stored only the non-zero entries
}
\usage{
A_t_times_x_particle(output, A_all_v, num_neighbors_vec, D_y, N_tilde)
}
\arguments{
  \item{output}{A numeric vector containing the input vector for multiplication.}
  
  \item{A_all_v}{A numeric vector containing the interaction matrices in vectorized form.}
  
  \item{num_neighbors_vec}{An integer vector specifying the number of neighbors for each particle.}
  
  \item{D_y}{An integer specifying the dimension of the output vector per particle.}
  
  \item{N_tilde}{An integer specifying the total dimension of the output vector.}
}
\value{
  Returns a numeric vector containing the result of the matrix-vector multiplication.
}
\references{
Fang, X., & Gu, M. (2024). \emph{The inverse Kalman filter}. arXiv:2407.10089.
}
\author{
\packageAuthor{FastGaSP}
Maintainer: \packageMaintainer{FastGaSP}
}
\keyword{internal}

\name{show.particle.est}
\alias{show,particle.est-method}
\title{Show method for particle estimation class}
\description{
  Display method for objects of class \code{particle.est}, which prints a summary of the estimated parameters for the particle interaction model.
}
\usage{
  \S4method{show}{particle.est}(object)
}
\arguments{
  \item{object}{An object of class \code{particle.est}.}
}
\details{
This method displays essential information about the estimated model parameters, including:
  \itemize{
    \item Data type (simulation or experiment)
    \item Model type (for simulation data only)
    \item Dimension of output space
    \item Estimated parameters:
      \itemize{
      \item beta parameters (inverse range)
      \item tau parameters (variance-noise ratio)
      \item interaction radius
    }
  }
}
\references{
Fang, X., & Gu, M. (2024). \emph{The inverse Kalman filter}. arXiv:2407.10089.
}

\seealso{
\code{\link{particle.est-class}} for details of the particle estimation class
}

\name{fgasp-class}
\docType{class}
\alias{fgasp-class}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Fast GaSP class}
\description{S4 class for fast computation of the Gaussian stochastic process (GaSP) model with the Matern kernel function with or without a noise.}

\section{Objects from the Class}{
Objects of this class are created and initialized with the function \code{\link{fgasp}} that computes the calculations needed for setting up the estimation and prediction.}

\section{Slots}{
	 \describe{
    \item{\code{num_obs}:}{object of class \code{integer}. The number of experimental observations.}
    \item{\code{have_noise}:}{object of class \code{logical} to specify whether the the model has a noise or not. "TRUE" means the model contains a noise and "FALSE" means the model does not contain a noise.}
    \item{\code{kernel_type}:}{a \code{character} to specify the type of kernel to use.The current version supports kernel_type to be "matern_5_2", "matern_3_2", or "exp", meaning that the matern kernel with roughness parameter being 2.5, 1.5, or 0.5 (exponent kernel), respectively. }
    \item{\code{input}:}{object of class \code{vector} with dimension num_obs x 1 for the sorted input locations.}
        \item{\code{delta_x}:}{object of class \code{vector} with dimension (num_obs-1) x 1 for the differences between the sorted input locations.}
    \item{\code{output}:}{object of class \code{vector} with dimension num_obs x 1 for the observations at the sorted input locations.}
        \item{\code{input_sorted_ix}:}{object of class \code{vector} with dimension num_obs x 1 with sorted index.}

    }
}


\section{Methods}{
  \describe{
    \item{show}{Prints the main slots of the object. }
    \item{predict}{See \code{\link[=predict.fgasp]{predict}}.}
	 }
}


\references{

Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models},  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.

M. Gu, Y. Xu (2017), \emph{Nonseparable Gaussian stochastic process: a unified
view and computational strategy}, arXiv:1711.11501.

M. Gu, X. Wang and J.O. Berger (2018), \emph{Robust Gaussian Stochastic Process Emulation}, \emph{Annals of Statistics}, \bold{46}, 3038-3066.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}


%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{ \code{\link{fgasp}} for more details about how to create a \code{fgasp} object.
%% ~~objects to See Also as \code{\link{help}}, ~~~
}

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{classes}
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line

\name{particle.data-class}
\alias{particle.data-class}
\alias{particle.data}
\docType{class}
\title{Particle trajectory data class}
\description{
S4 class for storing and analyzing particle trajectory data from both simulations and experimental observations. This class supports different models including Vicsek and can handle both position and velocity data along with optional angle information and particle tracking capabilities for experimental data.
}
\section{Objects from the Class}{
Objects of this class can be created in two ways:
\itemize{
\item For simulation data: Using \code{\link{simulate_particle}} that computes particle trajectories under physical models
\item For experimental data: Using \code{\link{trajectory_data}} to save particle trajectories while handling varying numbers of particles between time steps
}
}
\section{Slots}{
    \describe{
        \item{\code{px_list}:}{Object of class \code{list}. List of x-positions at each time step.}
        \item{\code{py_list}:}{Object of class \code{list}. List of y-positions at each time step.}
        \item{\code{vx_list}:}{Object of class \code{list}. List of x-velocities at each time step.}
        \item{\code{vy_list}:}{Object of class \code{list}. List of y-velocities at each time step.}
        \item{\code{theta_list}:}{Object of class \code{listOrNULL}. Optional list of particle velocity angles at each time step.}
        \item{\code{particle_tracking}:}{Object of class \code{listOrNULL}. List of data frames containing particle mappings between consecutive frames (primarily for experimental data).}
        \item{\code{data_type}:}{Object of class \code{character}. Type of data: either "simulation" or "experiment".}
        \item{\code{n_particles}:}{Object of class \code{numeric}. Number of particles (constant for simulation data, or a vector recording the number of particles at each time step for experimental data).}
        \item{\code{T_time}:}{Object of class \code{numeric}. Total number of time steps.}
        \item{\code{D_y}:}{Object of class \code{numeric}. Dimension of the output space.}
        \item{\code{model}:}{Object of class \code{characterOrNULL}. Type of particle interaction model (e.g., "Vicsek"). NULL for experimental data.}
        \item{\code{sigma_0}:}{Object of class \code{numericOrNULL}. Noise variance parameter used in the model. NULL for experimental data.}
        \item{\code{radius}:}{Object of class \code{numericOrNULL}. Interaction radius between particles. NULL for experimental data.}
    }
}

\section{Methods}{
\describe{
  \item{show:}{Method for displaying summary information about the particle.data object.}
  \item{fit:}{Method for fitting the latent factor model to data using the IKF-CG algorithm, which returns a \code{particle.est} object containing estimated parameters and predictions. See \code{\link{fit.particle.data}} for detailed documentation.}
}
}



\references{
Vicsek, T., Czirok, A., Ben-Jacob, E., Cohen, I., & Shochet, O. (1995). \emph{Novel type of phase transition in a system of self-driven particles}, \emph{Physical Review Letters}, \bold{75}(6), 1226.

Chat'{e}, H., Ginelli, F., Gr'{e}goire, G., Peruani, F., & Raynaud, F. (2008). \emph{Modeling collective motion: variations on the Vicsek model}, \emph{The European Physical Journal B}, \bold{64}(3), 451-456.
}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}

\seealso{
\code{\link{simulate_particle}} for simulating particle trajectories,
\code{\link{trajectory_data}} for saving experimantal particle trajectories,
\code{\link{fit.particle.data}} for model fitting methods
}
\keyword{classes}

\name{fit.gppca}
\alias{fit.gppca}
\alias{fit.gppca,gppca-method}
\title{ 
Parameter estimation for generalized probabilistic principal component analysis of correlated data.}

\description{
This function estimates the parameters for generalized probabilistic principal component analysis of correlated data.
}

\usage{
\S4method{fit.gppca}{gppca}(object, sigma0_2=NULL, d_ub=NULL)
}

\arguments{
  \item{object}{an object of  class \code{gppca.}}
  \item{sigma0_2}{variance of noise. Default is \code{NULL}. User should specify a value when it is known in real data.}
  \item{d_ub}{upper bound of d when d is estimated. Default is \code{NULL}.}
}

\value{
  \item{est_A}{the estimated factor loading matrix.}
  \item{est_beta}{the estimated inversed range parameter.}
  \item{est_sigma0_2}{the estimated variance of noise.}
  \item{est_sigma2}{the estimated variance parameter.}
  \item{mean_obs}{the predictive mean of the observations.}
  \item{mean_obs_95lb}{the lower bound of the 95\% posterior credible intervals of predictive mean.}
  \item{mean_obs_95ub}{the upper bound of the 95\% posterior credible intervals of predictive mean.}
}

\references{

Gu, M., & Shen, W. (2020), Generalized probabilistic principal component analysis of correlated data, \emph{Journal of Machine Learning Research, 21}(13), 1-41.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}


\examples{

library(FastGaSP)
library(rstiefel)

matern_5_2_funct <- function(d, beta_i) {
  cnst <- sqrt(5.0)
  matOnes <- matrix(1, nrow = nrow(d), ncol = ncol(d))
  result <- cnst * beta_i * d
  res <- (matOnes + result + (result^2) / 3) * exp(-result)
  return(res)
}

n=200
k=8
d=4

beta_real=0.01
sigma_real=1
sigma_0_real=sqrt(.01)
input=seq(1,n,1)
R0_00=as.matrix(abs(outer(input,input,'-')))
R_r = matern_5_2_funct(R0_00, beta_real)
L_sample = t(chol(R_r))
kernel_type='matern_5_2'


input=sort(input)
delta_x=input[2:length(input)]-input[1:(length(input)-1)]
A=rustiefel(k, d)  ##sample from Stiefel manifold
Factor=matrix(0,d,n)
for(i in 1: d){
  Factor[i,]=sigma_real^2*L_sample\%*\%rnorm(n)
}
output=A\%*\%Factor+matrix(rnorm(n*k,mean=0,sd=sigma_0_real),k,n)
  
##constucting the gppca.model
gppca_obj <- gppca(input, output, d, shared_params = TRUE,est_d=FALSE)
## estimate the parameters
gppca_fit <- fit.gppca(gppca_obj)

## MSE between predictive mean of observations and true mean
Y_mean=A\%*\%Factor
mean((gppca_fit$mean_obs-Y_mean)^2)
sd(Y_mean)

## coverage of 95\% posterior credible intervals
sum(gppca_fit$mean_obs_95lb<=Y_mean & gppca_fit$mean_obs_95ub>=Y_mean)/(n*k)

## length of 95\% posterior credible intervals
mean(abs(gppca_fit$mean_obs_95ub-gppca_fit$mean_obs_95lb))

}

\name{simulate_particle}
\alias{simulate_particle}
\title{Simulate particle trajectories}
\description{
  Simulates particle trajectories using either the standard Vicsek model or a two-interaction variation of the Vicsek model.
}
\usage{
  simulate_particle(v_abs, n_t = 100, T_sim = 5, h = 0.1,
    cut_r = 0.5, sigma_0 = 0.1,
    noise_type = "Gaussian", model = "Vicsek")
}
\arguments{
  \item{v_abs}{Absolute velocity magnitude for all particles.}
  \item{n_t}{Number of particles (default: 100).}
  \item{T_sim}{Total simulation time steps (default: 5).}
  \item{h}{Time step size for numerical integration (default: 0.1).}
  \item{cut_r}{Radius of interaction between particles (default: 0.5).}
  \item{sigma_0}{Standard deviation of noise (default: 0.1).}
  \item{noise_type}{Distribution of noise: "Gaussian" or "Uniform" (default: "Gaussian").}
  \item{model}{Type of interaction model: "Vicsek", "unnormalized_Vicsek", or "two_interactions_Vicsek" (default: "Vicsek").}
}
\value{
Returns an S4 object of class \code{\linkS4class{particle.data}}. See \code{\link{particle.data-class}} for details of the returned object structure.
}

\references{
Vicsek, T., Czirok, A., Ben-Jacob, E., Cohen, I., & Shochet, O. (1995). \emph{Novel type of phase transition in a system of self-driven particles}, \emph{Physical Review Letters}, \bold{75}(6), 1226.

Chat'{e}, H., Ginelli, F., Gr'{e}goire, G., Peruani, F., & Raynaud, F. (2008). \emph{Modeling collective motion: variations on the Vicsek model}, \emph{The European Physical Journal B}, \bold{64}(3), 451-456.

Fang, X., & Gu, M. (2024). \emph{The inverse Kalman filter}. arXiv:2407.10089.
}

\examples{
#--------------------------------------------------
# Example: Simulate using standard Vicsek model
#--------------------------------------------------
vx_abs=0.5
vy_abs=0.5
v_abs=sqrt(vx_abs^2+vy_abs^2)
sim1 <- simulate_particle(v_abs=v_abs)

#--------------------------------------------------
# Example: Simulate using unnormalized variation
#--------------------------------------------------
sim2 <- simulate_particle(v_abs=v_abs, model = 'unnormalized_Vicsek')

#--------------------------------------------------
# Example: Simulate using two-interaction variation
#--------------------------------------------------
sim3 <- simulate_particle(v_abs=v_abs, model = 'two_interactions_Vicsek')
}
\seealso{
\code{\link{particle.data-class}} for details on the particle.data class structure
}

\name{log_lik}
\alias{log_lik}
\title{ 
Natural logarithm of profile likelihood by the fast computing algorithm}
\description{
This function computes the natural logarithm of the profile likelihood for the range and nugget parameter (if there is one) after plugging the closed form maximum likelihood estimator for the variance parameter. 
}

\usage{
log_lik(param, object)
}

\arguments{
  \item{param}{a vector of parameters. The first parameter is the natural logarithm of the inverse range parameter in the kernel function. If the data contain noise, the second parameter is the logarithm of the nugget-variance ratio parameter.}
  \item{object}{an object of  class \code{fgasp}.}
}

\value{
The numerical value of natural logarithm of the profile likelihood.
}

\references{
%% ~put references to the literature/web site here ~

Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models},  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.

M. Gu, Y. Xu (2017), \emph{Nonseparable Gaussian stochastic process: a unified
view and computational strategy}, arXiv:1711.11501.

M. Gu, X. Wang and J.O. Berger (2018), \emph{Robust Gaussian Stochastic Process Emulation}, \emph{Annals of Statistics}, \bold{46}, 3038-3066.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}

%\note{
%%  ~~further notes~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~

%\seealso{
%% ~~objects to See Also as \code{\link{help}}, ~~~
%}

\examples{
library(FastGaSP)
#--------------------------------------------------------------------------
# Example 1: comparing the fast and slow algorithms to compute the likelihood 
# of the Gaussian stochastic process for data with noises
#--------------------------------------------------------------------------

y_R<-function(x){
  sin(2*pi*x)
}

###let's test for 1000 observations
set.seed(1)
num_obs=1000
input=runif(num_obs)
output=y_R(input)+rnorm(num_obs,mean=0,sd=0.1)

##constucting the fgasp.model
fgasp.model=fgasp(input, output)

##range and noise-variance ratio (nugget) parameters 
param=c( log(1),log(.02))
## the log lik
log_lik(param,fgasp.model)
##time cost to compute the likelihood
time_cost=system.time(log_lik(param,fgasp.model))
time_cost[1]

##now I am comparing to the one with straightforward inversion

matern_5_2_kernel<-function(d,beta){  
  res=(1+sqrt(5)*beta*d + 5*beta^2*d^2/3 )*exp(-sqrt(5)*beta*d)
  res
}

##A function for computing the likelihood by the GaSP in a straightforward way
log_lik_GaSP_slow<-function(param,have_noise=TRUE,input,output){
  n=length(output)
  beta=exp(param[1])
  eta=0
  if(have_noise){
    eta=exp(param[2])
  }
  R00=abs(outer(input,input,'-'))
  R=matern_5_2_kernel(R00,beta=beta)
  R_tilde=R+eta*diag(n)
  #use Cholesky and one backsolver and one forward solver so it is more stable
  L=t(chol(R_tilde))
  output_t_R.inv= t(backsolve(t(L),forwardsolve(L,output )))
  S_2=output_t_R.inv\%*\%output
  
  -sum(log(diag(L)))-n/2*log(S_2)
}



##range and noise-variance ratio (nugget) parameters 
param=c( log(1),log(.02))
## the log lik
log_lik(param,fgasp.model)
log_lik_GaSP_slow(param,have_noise=TRUE,input=input,output=output)

##time cost to compute the likelihood
##More number of inputs mean larger differences
time_cost=system.time(log_lik(param,fgasp.model))
time_cost

time_cost_slow=system.time(log_lik_GaSP_slow(param,have_noise=TRUE,input=input,output=output))
time_cost_slow


#--------------------------------------------------------------------------
# Example 2: comparing the fast and slow algorithms to compute the likelihood 
# of the Gaussian stochastic process for data without a noise
#--------------------------------------------------------------------------
## Here is a function in the Sobolev Space with order 3
y_R<-function(x){
  j_seq=seq(1,200,1)
  record_y_R=0
  for(i_j in 1:200){
    record_y_R=record_y_R+2*j_seq[i_j]^{-2*3}*sin(j_seq[i_j])*cos(pi*(j_seq[i_j]-0.5)*x)

  }
  record_y_R
}


##generate some data without noise
num_obs=50
input=seq(0,1,1/(num_obs-1))

output=y_R(input)


##constucting the fgasp.model
fgasp.model=fgasp(input, output,have_noise=FALSE)

##range and noise-variance ratio (nugget) parameters 
param=c( log(1))
## the log lik
log_lik(param,fgasp.model)
log_lik_GaSP_slow(param,have_noise=FALSE,input=input,output=output)


}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
%\keyword{ ~kwd1 }
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line

\name{f_Vicsek_variation}
\alias{f_Vicsek_variation}
\title{Modified Vicsek Interaction Function}
\description{
  A modified interaction function for the Vicsek model that defines the interaction strength between particles based on their distance. This function is used as one of the interactions in the two-interaction Vicsek model.
}
\usage{
f_Vicsek_variation(r, a = 0.02, b = 1, r_min = 0.01, r_max = 0.8, beta = 20)
}
\arguments{
  \item{r}{numeric, the distance between particles}
  \item{a}{numeric, strength parameter for the short-range interaction term. Default is 0.02}
  \item{b}{numeric, strength parameter for the linear term. Default is 1}
  \item{r_min}{numeric, minimum distance parameter for the interaction term. Default is 0.01}
  \item{r_max}{numeric, maximum distance parameter. Default is 0.8}
  \item{beta}{numeric, scaling parameter for the overall interaction. Default is 20}
}
\details{
  The function implements a modified Vicsek interaction with three components:
  \itemize{
    \item A short-range term (-a/(r+r_min))
    \item A linear term (-b*(r-r_max))
    \item A constant offset term (a/r_max)
  }
  The final value is scaled by the beta parameter.
}
\value{
  Returns a numeric value representing the interaction strength at distance r.
}
\examples{
# Calculate interaction values at various distances
r_seq <- seq(0.01, 1, by = 0.01)
interaction <- f_Vicsek_variation(r_seq)
plot(r_seq, interaction, type = "l", 
     xlab = "Distance", ylab = "Interaction Strength",
     main = "Vicsek Variation Interaction Function")
}

\references{
Chat'{e}, H., Ginelli, F., Gr'{e}goire, G., Peruani, F., & Raynaud, F. (2008). \emph{Modeling collective motion: variations on the Vicsek model}, \emph{The European Physical Journal B}, \bold{64}(3), 451-456.

Fang, X., & Gu, M. (2024). \emph{The inverse Kalman filter}. arXiv:2407.10089.

}

\keyword{internal}

\name{Sample_KF_post}
\alias{Sample_KF_post}
\title{ 
Sample the posterior distribution of the process using the backward smoothing algorithm  
}
\description{
This function samples the posterior distribution of the process using the backward smoothing algorithm. 
}
\usage{
Sample_KF_post(index_obs, C_R_K_Q,W0,GG,W,VV,output,kernel_type,sample_type)
}
\arguments{
    \item{index_obs}{a vector  where the entries with 1 have observations and entries with 0 have no observation.}
        \item{C_R_K_Q}{a list of matrices to compute the inverse covariance matrix in the dynamic linear model.}

  \item{GG}{a list of matrices defined in the dynamic linear model. 
}
  \item{W}{a list of coefficient matrices defined in the dynamic linear model. 
}
  \item{VV}{a numerical value of the variance of the nugget parameter. }
  \item{output}{a vector of the output. }

  \item{kernel_type}{a \code{character} to specify the type of kernel to use. The current version supports kernel_type to be "matern_5_2", "matern_3_2" or "exp", meaning that the matern kernel with roughness parameter being 2.5, 1.5, or 0.5 (exponent kernel), respectively. }
    \item{sample_type}{a integer to specify the type of sample we need. 0 means the states. 1 means the first value of each state vector. 2 means the noisy observations. }

}


\value{
A matrix of the posterior samples. 
}

\references{
%% ~put references to the literature/web site here ~

Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models}.  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.

M. Gu, Y. Xu (2019), \emph{fast nonseparable gaussian stochastic process with application to methylation level interpolation}.  \emph{Journal of Computational and Graphical Statistics}, In Press, arXiv:1711.11501.

Campagnoli P, Petris G, Petrone S. (2009), \emph{Dynamic linear model with R}. Springer-Verlag New York.


}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}
\seealso{\code{\link{Sample_KF_post}} for more details about sampling from the posterior distribution.
%% ~~objects to See Also as \code{\link{help}}, ~~~
}

\keyword{internal}
%\note{
%%  ~~further notes~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~

%\seealso{
%% ~~objects to See Also as \code{\link{help}}, ~~~
%}

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
%\keyword{ ~kwd1 }
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line

\name{unnormalized_Vicsek}
\alias{unnormalized_Vicsek}
\title{Unnormalized Vicsek model simulation}
\description{
  Simulates particle movement according to the unnormalized Vicsek model, where particles align their velocities with neighboring particles within a specified radius, subject to noise. The model implements both Gaussian and uniform noise options.
}
\usage{
unnormalized_Vicsek(p0,v0,n_t,T_sim,h,cut_r,
                    sigma_0,noise_type='Gaussian')
}
\arguments{
  \item{p0}{A numeric vector of initial positions for all particles, structured as (x1, y1, x2, y2, ..., xn, yn).}
  \item{v0}{A numeric vector of initial velocities for all particles, structured similarly to p0.}
  \item{n_t}{An integer specifying the number of particles.}
  \item{T_sim}{An integer specifying the number of time steps to simulate.}
  \item{h}{A numeric value specifying the time step size.}
  \item{cut_r}{A numeric value specifying the interaction radius within which particles align.}
  \item{sigma_0}{A numeric value specifying the noise strength.}
  \item{noise_type}{A character string specifying the type of noise: either "Gaussian" (default) or "Uniform".}
}
\value{
  Returns a list with three components:
  \item{pos}{A matrix of dimension (2*n_t) x (T_sim+1) containing particle positions at each time step.}
  \item{v}{A matrix of dimension (2*n_t) x (T_sim+1) containing particle velocities at each time step.}
}

\keyword{internal}

\name{FastGaSP-package}
\alias{FastGaSP-package}
\alias{FastGaSP}
\docType{package}
\title{
\packageTitle{FastGaSP}
}
\description{
\packageDescription{FastGaSP}
}
\details{

The DESCRIPTION file:
\packageDESCRIPTION{FastGaSP}
\packageIndices{FastGaSP}
Fast computational algorithms for Gaussian stochastic process with Matern kernels 
by the forward filtering and backward smoothing algorithm.
}
\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}
\references{

Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models},  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.


M. Gu and Y. Xu (2020), \emph{Nonseparable Gaussian stochastic process: a unified
view and computational strategy}, \emph{Fast Nonseparable Gaussian Stochastic Process With Application to Methylation Level Interpolation}, \emph{Journal of Computational and Graphical Statistics}, \bold{29}, 250-260.

M. Gu and W. Shen (2020), \emph{Generalized probabilistic principal component analysis of correlated data}, \emph{Journal of Machine Learning Research}, \bold{21}, 13-1.

M. Gu, X. Wang and J.O. Berger (2018), \emph{Robust Gaussian Stochastic Process Emulation}, \emph{Annals of Statistics}, \bold{46}, 3038-3066.

}
%~~ Optionally other standard keywords, one per line, from file KEYWORDS in the R documentation ~~
%~~ directory ~~
\keyword{Gaussian stochastic process}
\keyword{Kalman filter}
\keyword{profile likelihood}
\keyword{prediction}


\seealso{
\code{\link[FastGaSP]{FastGaSP}}
}
\examples{

library(FastGaSP)

#------------------------------------------------------------------------------
# Example 1 : fast computation algorithm for noisy data
#------------------------------------------------------------------------------

y_R<-function(x){
  sin(2*pi*x)
}

###let's test for 2000 observations
set.seed(1)
num_obs=2000
input=runif(num_obs)
output=y_R(input)+rnorm(num_obs,mean=0,sd=0.1)

##constucting the fgasp.model
fgasp.model=fgasp(input, output)

##range and noise-variance ratio (nugget) parameters 
param=c( log(1),log(.02))
## the log lik
log_lik(param,fgasp.model)
##time cost to compute the likelihood
time_cost=system.time(log_lik(param,fgasp.model))
time_cost[1]

##consider a nonparametric regression setting 
##estimate the parameter by maximum likelihood estimation

est_all<-optim(c(log(1),log(.02)),log_lik,object=fgasp.model,method="L-BFGS-B",
              control = list(fnscale=-1))  

##estimated log inverse range parameter and log nugget
est_all$par

##estimate variance 
est.var=Get_log_det_S2(est_all$par,fgasp.model@have_noise,fgasp.model@delta_x,
                          fgasp.model@output,fgasp.model@kernel_type)[[2]]/fgasp.model@num_obs
est.var

###1. Do some interpolation test 
num_test=5000
testing_input=runif(num_test) ##there are the input where you don't have observations
pred.model=predict(param=est_all$par,object=fgasp.model,testing_input=testing_input)

lb=pred.model@mean+qnorm(0.025)*sqrt(pred.model@var)
ub=pred.model@mean+qnorm(0.975)*sqrt(pred.model@var)

## calculate lb for the mean function
pred.model2=predict(param=est_all$par,object=fgasp.model,testing_input=testing_input,var_data=FALSE)
lb_mean_funct=pred.model2@mean+qnorm(0.025)*sqrt(pred.model2@var)
ub_mean_funct=pred.model2@mean+qnorm(0.975)*sqrt(pred.model2@var)

## plot the prediction
min_val=min(lb,output)
max_val=max(ub,output)

plot(pred.model@testing_input,pred.model@mean,type='l',col='blue',
     ylim=c(min_val,max_val),
     xlab='x',ylab='y')
polygon(c(pred.model@testing_input,rev(pred.model@testing_input)),
        c(lb,rev(ub)),col = "grey80", border = FALSE)
lines(pred.model@testing_input,pred.model@mean,type='l',col='blue')
lines(pred.model@testing_input,y_R(pred.model@testing_input),type='l',col='black')
lines(pred.model2@testing_input,lb_mean_funct,col='blue',lty=2)
lines(pred.model2@testing_input,ub_mean_funct,col='blue',lty=2)
lines(input,output,type='p',pch=16,col='black',cex=0.4) #one can plot data

legend("bottomleft", legend=c("predictive mean","95\% predictive interval","truth"),
       col=c("blue","blue","black"), lty=c(1,2,1), cex=.8)

#--------------------------------------------------------------
# Example 2: example that one does not have a noise in the data
#--------------------------------------------------------------
## Here is a function in the Sobolev Space with order 3
y_R<-function(x){
  j_seq=seq(1,200,1)
  record_y_R=0
  for(i_j in 1:200){
    record_y_R=record_y_R+2*j_seq[i_j]^{-2*3}*sin(j_seq[i_j])*cos(pi*(j_seq[i_j]-0.5)*x)

  }
  record_y_R
}


##generate some data without noise
num_obs=50
input=seq(0,1,1/(num_obs-1))

output=y_R(input)


##constucting the fgasp.model
fgasp.model=fgasp(input, output,have_noise=FALSE)

##range and noise-variance ratio (nugget) parameters 
param=c( log(1))
## the log lik
log_lik(param,fgasp.model)



#if one does not have noise one may need to give a lower bound or use a penalty 
#(e.g. induced by a prior) to make the estimation more robust
est_all<-optimize(log_lik,interval=c(0,10),maximum=TRUE,fgasp.model)
  
##Do some interpolation test for comparison
num_test=1000
testing_input=runif(num_test) ##there are the input where you don't have observations

pred.model=predict(param=est_all$maximum,object=fgasp.model,testing_input=testing_input)



#This is the 95 posterior credible interval for the outcomes which contain the estimated 
#variance of the noise
#sometimes there are numerical instability is one does not have noise or error
lb=pred.model@mean+qnorm(0.025)*sqrt(abs(pred.model@var))
ub=pred.model@mean+qnorm(0.975)*sqrt(abs(pred.model@var))

## plot the prediction
min_val=min(lb,output)
max_val=max(ub,output)

plot(pred.model@testing_input,pred.model@mean,type='l',col='blue',
     ylim=c(min_val,max_val),
     xlab='x',ylab='y')
polygon( c(pred.model@testing_input,rev(pred.model@testing_input)),
         c(lb,rev(ub)),col = "grey80", border = FALSE)
lines(pred.model@testing_input,pred.model@mean,type='l',col='blue')
lines(pred.model@testing_input,y_R(pred.model@testing_input),type='l',col='black')
lines(input,output,type='p',pch=16,col='black')
legend("bottomleft", legend=c("predictive mean","95\% predictive interval","truth"),
       col=c("blue","blue","black"), lty=c(1,2,1), cex=.8)


##mean square error for all inputs
mean((pred.model@mean- y_R(pred.model@testing_input))^2)

}

\name{fmou}
\alias{fmou}
\alias{fmou-method}
\title{ 
Setting up the FMOU model}
\description{
Creating an \code{fmou} class for fmou, a latent factor model with a fixed or estimated orthogonal factor loading matrix, where each latent factor is modeled as an O-U (Ornstein-Uhlenbeck) process.
}

\usage{
  fmou(output, d, est_d=FALSE, est_U0=TRUE, est_sigma0_2=TRUE, U0=NULL, sigma0_2=NULL)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
    \item{output}{a k*n observation matrix, where k is the length of observations at each time step and n is the number of time steps.}
    \item{d}{number of latent factors.}
    \item{est_d}{a bool value, default is \code{FALSE}. If \code{TRUE}, d will be estimated by either variance matching (when noise level is given) or information criteria (when noise level is unknown). Otherwise, d is fixed, and users must assign a value to argument d.}
    \item{est_U0}{a bool value, default is \code{TRUE}. If \code{TRUE}, the factor loading matrix (U0) will be estimated. Otherwise, U0 is fixed.}
    \item{est_sigma0_2}{a bool value, default is \code{TRUE} . If \code{TRUE}, the variance of the noise will be estimated. Otherwise, it is fixed.}
    \item{U0}{the fixed factor loading matrix. Users should assign a k*d matrix to it when \code{est_U0=FALSE}. }
    \item{sigma0_2}{variance of noise. User should assign a value to it when \code{est_sigma0_2=FALSE}.}

}

\value{
\code{fmou} returns an S4 object of class \code{fmou}.
}

\references{

Lin, Y., Liu, X., Segall, P., & Gu, M. (2025). Fast data inversion for high-dimensional dynamical systems from noisy measurements. arXiv preprint arXiv:2501.01324.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}

\examples{



## generate simulated data
library(FastGaSP)
library(rstiefel)

d = 5  # number of latent factors
k = 20 # length of observation at each time step
n = 100 # number time step
noise_level = 1 # variance of noise

U = rustiefel(k, k) # factor loading matrix
z = matrix(NA, d, n)
sigma_2 = runif(d, 0.5, 1)
rho = runif(d, 0.95, 1)
for(l in 1:d){
  R = matrix(NA, n, n)
  diag(R) = 1
  for(ir in 1:n){
    for(ic in 1:n){
      R[ir, ic] = rho[l]^(abs(ir-ic)) * R[ir, ir]
    }
  }
  R = (sigma_2[l]/(1-rho[l]^2) )* R
  z[l, ] = t(chol(R)) \%*\% rnorm(n)
}

signal = U[,1:d] \%*\% z
y = signal + matrix(rnorm(n*k,mean=0,sd=sqrt(noise_level)),k,n)

##constucting the fmou.model
fmou.model=fmou(output=y, d=d, est_U0=TRUE, est_sigma0_2=TRUE)
}

\name{fit.particle.data}
\alias{fit.particle.data}
\alias{fit,particle.data-method}
\title{Fit method for particle data}
\description{
Estimates interaction parameters for particle systems using trajectory data with the IKF-CG (Inverse Kalman Filter - Conjugate Gradient) approach. Supports both simulation and experimental data.
}
\usage{
  \S4method{fit}{particle.data}(
    object, param, cut_r_max=1, est_param = TRUE, nx=NULL, ny=NULL,
    kernel_type = "matern_5_2", tilde_nu = 0.1, tol = 1e-6,
    maxIte = 1000, output = NULL, ho_output = NULL, 
    testing_inputs=NULL, compute_CI = TRUE, num_interaction = (length(param)-1)/2,
    data_type = object@data_type, model = object@model, 
    apolar_vicsek = FALSE, direction = NULL
  )
}
\arguments{
  \item{object}{An object of class \code{particle.data} containing the trajectory data.}
  \item{param}{Numeric vector of parameters. Should contain 2*num_interaction + 1 elements: first num_interaction elements are log of inverse range parameters (beta), next num_interaction elements are log of variance-noise ratios (tau), and the final element is log(radius/(cut_r_max-radius)) where radius is the interaction radius.}
  \item{cut_r_max}{Numeric value specifying the maximum interaction radius to consider during estimation (default: 1).}
  \item{est_param}{If TRUE, param is used as initial values for parameter optimization. If FALSE, param is treated as fixed parameters for prediction (default: TRUE).}
  \item{nx}{An integer specifying the number of grid points along the x-axis (horizontal direction). If NULL, automatically calculated as floor((px_max-px_min)/cut_r_max), where px_max and px_min represent the maximum and minimum x-coordinates of all particles.}  
  \item{ny}{An integer specifying the number of grid points along the y-axis (vertical direction). If NULL, automatically calculated as floor((py_max-py_min)/cut_r_max), where py_max and py_min represent the maximum and minimum y-coordinates of all particles.}
  \item{kernel_type}{Character string specifying the kernel type: 'matern_5_2' (default) or 'exp'.}
  \item{tilde_nu}{Numeric value for stabilizing the IKF computation (default: 0.1).}
  \item{tol}{Numeric value specifying convergence tolerance for the conjugate gradient algorithm (default: 1e-6).}
  \item{maxIte}{Integer specifying maximum iterations for the conjugate gradient algorithm (default: 1000).}
  \item{output}{Numerical vector (default = NULL). Used for residual bootstrap when different outputs but same inputs are needed.}
  \item{ho_output}{Numerical vector (default = NULL). Used for residual bootstrap when different hold-out outputs but same inputs are needed.}
  \item{testing_inputs}{Matrix of inputs for prediction (NULL if only performing parameter estimation). Each column represents testing inputs for one interaction.}
  \item{compute_CI}{When TRUE, computes the 95\% credible interval for testing_inputs (default: TRUE).}
  \item{num_interaction}{Integer specifying number of interactions to predict (default: (length(param_ini)-1)/2).}
  \item{data_type}{Character string indicating data type ("simulation" or "experiment").}
  \item{model}{Character string specifying the model type (e.g., "unnormalized_Vicsek").}
  \item{apolar_vicsek}{When TRUE, considers only neighboring particles moving in the same direction (default: FALSE).}
  \item{direction}{Modeling direction ('x' or 'y') for experimental data analysis.}
}
\value{
Returns an object of class \code{\linkS4class{particle.est}}. See \code{\link{particle.est-class}} for details.
}
\references{
Fang, X., & Gu, M. (2024). \emph{The inverse Kalman filter}. arXiv:2407.10089.
}
\examples{
# Simulate data
vx_abs <- 0.5
vy_abs <- 0.5
v_abs <- sqrt(vx_abs^2+vy_abs^2)
sim <- simulate_particle(v_abs=v_abs, model = 'unnormalized_Vicsek')
show(sim)

# Set up testing inputs and initial parameters
testing_n <- 200
testing_inputs <- matrix(as.numeric(seq(-1,1,length.out=testing_n)),nr=1)
cut_r_max=1.5
param_ini <- log(c(0.3,1000,0.3/(cut_r_max-0.3)))  # Initial parameter values

# Fit model to simulation data
est <- fit(sim,param=param_ini,cut_r_max=1.5, testing_inputs = testing_inputs)
show(est)
}
\keyword{methods}

\name{fmou-class}
\docType{class}
\alias{fmou-class}
\title{ FMOU class}
\description{
An S4 class for fast parameter estimation in the FMOU model, a latent factor model with a fixed or estimated orthogonal factor loading matrix, where each latent factor is modeled as an O-U (Ornstein-Uhlenbeck) process.
}

\section{Objects from the Class}{
Objects of this class are created and initialized using the \code{\link{fmou}} function to set up the estimation.
}

\section{Slots}{
	 \describe{
    \item{\code{output}:}{object of class \code{matrix}. The observation matrix.}
    \item{\code{d}}{object of class \code{integer} to specify the number of latent factors.}
    \item{\code{est_d}}{object of class \code{logical}, default is \code{FALSE}. If \code{TRUE}, d will be estimated by either variance matching (when noise level is given) or information criteria (when noise level is unknown). Otherwise, d is fixed, and users must assign a value to \code{d}.}
    \item{\code{est_U0}}{object of class \code{logical}, default is \code{TRUE}. If \code{TRUE}, the factor loading matrix (U0) will be estimated. Otherwise, U0 is fixed.}
    \item{\code{est_sigma0_2}}{object of class \code{logical}, default is \code{TRUE} . If \code{TRUE}, the variance of the noise will be estimated. Otherwise, it is fixed.}
    \item{\code{U0}}{object of class \code{matrix}. The fixed factor loading matrix. Users should assign a k*d matrix to it when \code{est_U0=False}. Here k is the length of observations at each time step.}
    \item{\code{sigma0_2}}{object of class \code{numeric}. Variance of noise. User should assign a value to it when \code{est_sigma0_2=False}.}
    }
}

\section{Methods}{
  \describe{
    \item{fit.fmou}{See \code{\link[=fit.fmou]{fit.fmou}}.}
    \item{predict.fmou}{See \code{\link[=predict.fmou]{predict.fmou}}.}
	 }
}

\references{

Lin, Y., Liu, X., Segall, P., & Gu, M. (2025). Fast data inversion for high-dimensional dynamical systems from noisy measurements. arXiv preprint arXiv:2501.01324.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}

\seealso{ \code{\link{fmou}} for more details about how to create a \code{fmou} object.}

\name{IKF_CG_particle_cell}
\alias{IKF_CG_particle_cell}
\title{Inverse Kalman Filter with Conjugate Gradient for Particle Systems}
\description{
  Implements the IKF-CG (Inverse Kalman Filter - Conjugate Gradient) algorithm for the one-interaction physical model with 1D output, different particle numbers for different time, and non-identity diagonal noise. 
}
\usage{
IKF_CG_particle_cell(param, kernel_type, delta_x_all, output, A_all_v, 
                     sort_d_all_ix, sigma_2_vec, num_neighbors_vec, 
                     tilde_nu, D, n_t_record, tol = 1e-6, maxIte = 1000)
}
\arguments{
  \item{param}{A numeric vector containing model parameters. The first element is the log of beta (inverse range parameter), and the second element is the log of tau (variance ratio parameter).}
  \item{kernel_type}{A string specifying the covariance kernel type: "matern_5_2" or "exp".}
  \item{delta_x_all}{A numeric vector of successive differences of sorted inputs.}
  \item{output}{A numeric vector of observations.}
  \item{A_all_v}{A numeric vector containing the non-zero entries in matrix A.}
  \item{sort_d_all_ix}{An integer vector of indices for sorting distances.}
  \item{sigma_2_vec}{A numeric vector of variances for each time point.}
  \item{num_neighbors_vec}{An integer vector specifying number of neighbors for each observation.}
  \item{tilde_nu}{A numeric value for numerical stabilization.}
  \item{D}{An integer specifying the dimension of observations.}
  \item{n_t_record}{An integer vector containing number of particles at each time point.}
  \item{tol}{A numeric value specifying convergence tolerance (default: 1e-6).}
  \item{maxIte}{An integer specifying maximum number of iterations (default: 1000).}
}
\value{
  Returns a list containing:
  \item{x}{A numeric vector of solved coefficients.}
  \item{resid}{A numeric vector of residuals at each iteration.}
  \item{ite}{An integer indicating the number of iterations performed.}
}
\references{
  Fang, X., & Gu, M. (2024). \emph{The inverse Kalman filter}. arXiv:2407.10089.
}

\keyword{internal}

\name{predict.gppca}
\alias{predict.gppca}
\alias{predict.gppca,gppca-method}
\title{ 
Prediction and uncertainty quantification on the future observations using GPPCA.}

\description{
This function predicts the future observations using a GPPCA model. Uncertainty quantification is available. 
}

\usage{
\S4method{predict.gppca}{gppca}(object, param, A_hat, step=1, interval=FALSE, interval_data=TRUE)
}

\arguments{
  \item{object}{an object of  class \code{gppca.}}
  \item{param}{estimated parameters (beta, sigma2, sigma0_2).}
  \item{A_hat}{estimated factor loading matrix.}
  \item{step}{a vector. Number of steps to be predicted. Default is 1. } 
  \item{interval}{a bool value, default is \code{FALSE}. If \code{TRUE}, the 95\% predictive intervals are computed.}
  \item{interval_data}{a bool value, default is \code{TRUE}. If \code{TRUE}, the 95\% predictive intervals of the observations are computed. Otherwise, the 95\% predictive intervals of the mean of the observation are computed.} 
}

\value{
  \item{pred_mean}{the predictive mean. }
  \item{pred_interval_95lb}{the 95\% lower bound of the interval.}
  \item{pred_interval_95ub}{the 95\% upper bound of the interval.}
}

\references{

Gu, M., & Shen, W. (2020), Generalized probabilistic principal component analysis of correlated data, \emph{Journal of Machine Learning Research, 21}(13), 1-41.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}


\examples{

library(FastGaSP)
library(rstiefel)

matern_5_2_funct <- function(d, beta_i) {
  cnst <- sqrt(5.0)
  matOnes <- matrix(1, nrow = nrow(d), ncol = ncol(d))
  result <- cnst * beta_i * d
  res <- (matOnes + result + (result^2) / 3) * exp(-result)
  return(res)
}

n=200
k=8
d=4

beta_real=0.01
sigma_real=1
sigma_0_real=sqrt(.01)
input=seq(1,n,1)
R0_00=as.matrix(abs(outer(input,input,'-')))
R_r = matern_5_2_funct(R0_00, beta_real)
L_sample = t(chol(R_r))
kernel_type='matern_5_2'


input=sort(input)
delta_x=input[2:length(input)]-input[1:(length(input)-1)]
A=rustiefel(k, d)  ##sample from Stiefel manifold
Factor=matrix(0,d,n)
for(i in 1: d){
  Factor[i,]=sigma_real^2*L_sample\%*\%rnorm(n)
}
output=A\%*\%Factor+matrix(rnorm(n*k,mean=0,sd=sigma_0_real),k,n)
  
##constucting the gppca.model
gppca_obj <- gppca(input, output, d, shared_params = TRUE,est_d=FALSE)
## estimate the parameters
gppca_fit <- fit.gppca(gppca_obj)

## two-step-ahead prediction
param <- c(gppca_fit$est_beta, gppca_fit$est_sigma2, gppca_fit$est_sigma0_2)
gppca_pred <- predict.gppca(gppca_obj, param, gppca_fit$est_A, step=1:3)
gppca_pred$pred_mean

}

\name{Kalman_smoother}
\alias{Kalman_smoother}
\title{ 
the predictive mean and predictive variance by Kalman Smoother
}
\description{
This function computes the predictive mean and predictive variance on the sorted input and testing input  by the Kalman Smoother.
}
\usage{
Kalman_smoother(param,have_noise,index_obs,delta_x_all,output,sigma_2_hat,kernel_type)
}
\arguments{
  \item{param}{a vector of parameters. The first parameter is the natural logarithm of the inverse range parameter in the kernel function. If the data contain noise, the second parameter is the logarithm of the nugget-variance ratio parameter.
}
  \item{have_noise}{a bool value. If it is true, it means the model contains a noise. 
%%     ~~Describe \code{response} here~~
}
    \item{index_obs}{a vector  where the entries with 1 have observations and entries with 0 have no observation.}

    \item{delta_x_all}{a vector for the differences between the sorted input and testing input locations.}
    
    \item{output}{a  vector with dimension num_obs x 1 for the observations at the sorted input locations.}
        \item{sigma_2_hat}{a numerical value of variance parameter of the covariance function.}
     \item{kernel_type}{A \code{character} specifying the type of kernel.}

}


\value{
A list where the first item is the the predictive mean and the second item is predictive variance on the sorted input and testing input  by the Kalman Smoother.
}

\references{
%% ~put references to the literature/web site here ~

Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models},  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.

M. Gu, Y. Xu (2017), \emph{Nonseparable Gaussian stochastic process: a unified
view and computational strategy}, arXiv:1711.11501.

M. Gu, X. Wang and J.O. Berger (2018), \emph{Robust Gaussian Stochastic Process Emulation}, \emph{Annals of Statistics}, \bold{46}, 3038-3066.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}
\seealso{\code{\link{predict}} for more details about the prediction on the testing input by the Fast GaSP class.
%% ~~objects to See Also as \code{\link{help}}, ~~~
}

\keyword{internal}
%\note{
%%  ~~further notes~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~

%\seealso{
%% ~~objects to See Also as \code{\link{help}}, ~~~
%}

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
%\keyword{ ~kwd1 }
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line

\name{gppca}
\alias{gppca}
\alias{gppca-method}
\title{Setting up the GPPCA model}
\description{Creating an \code{gppca} class for generalized probabilistic 
principal component analysis of correlated data.}

\usage{
gppca(input,output, d, est_d=FALSE, shared_params=TRUE, kernel_type="matern_5_2")
}
%- maybe also 'usage' for other objects documented here.
\arguments{
    \item{input}{a vector for the sorted input locations. The length is equal to the number of observations.}
    \item{output}{a k*d matrix for the observations at the sorted input locations. Here k is the number of locations and n is the number of observations.}
    \item{d}{number of latent factors.}
    \item{est_d}{a bool value, default is \code{FALSE}. If \code{TRUE}, d will be estimated by either variance matching (when noise level is given) or information criteria (when noise level is unknown). Otherwise, d is fixed, and users must assign a value to argument d.}
    \item{shared_params}{a bool value, default is \code{TRUE}. If \code{TRUE}, the latent processes share the correlation and variance parameters. Otherwise, each latent process has distinct parameters.}
    \item{kernel_type}{a \code{character} to specify the type of kernel to use. The current version supports kernel_type to be "matern_5_2" or "exponential", meaning that the matern kernel with roughness parameter being 2.5 or 0.5 (exponent kernel), respectively. }
    % description of inputs of the gppca() function
}

\value{
\code{gppca} returns an S4 object of class \code{gppca}.
}

\references{

Gu, M., & Shen, W. (2020), Generalized probabilistic principal component analysis of correlated data, \emph{Journal of Machine Learning Research, 21(13)}, 1-41.

}


\author{
\packageAuthor{FastGaSP}
Maintainer: \packageMaintainer{FastGaSP}
}

\examples{
library(FastGaSP)
library(rstiefel)

matern_5_2_funct <- function(d, beta_i) {
  cnst <- sqrt(5.0)
  matOnes <- matrix(1, nrow = nrow(d), ncol = ncol(d))
  result <- cnst * beta_i * d
  res <- (matOnes + result + (result^2) / 3) * exp(-result)
  return(res)
}

n=200
k=8
d=4

beta_real=0.01
sigma_real=1
sigma_0_real=sqrt(.01)
input=seq(1,n,1)
R0_00=as.matrix(abs(outer(input,input,'-')))
R_r = matern_5_2_funct(R0_00, beta_real)
L_sample = t(chol(R_r))
kernel_type='matern_5_2'


input=sort(input)
delta_x=input[2:length(input)]-input[1:(length(input)-1)]
A=rustiefel(k, d)  ##sample from Stiefel manifold
Factor=matrix(0,d,n)
for(i in 1: d){
  Factor[i,]=sigma_real^2*L_sample\%*\%rnorm(n)
}
output=A%*%Factor+matrix(rnorm(n*k,mean=0,sd=sigma_0_real),k,n)
  
##constucting the gppca.model
gppca_obj <- gppca(input, output, d, shared_params = TRUE,est_d=FALSE)

}

\name{fgasp}
\alias{fgasp}
%\alias{show.rgasp}
\alias{fgasp-method}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{ Setting up the Fast GaSP model
%%  ~~function to do ... ~~
}
\description{
%%  ~~ A concise (1-5 lines) description of what the function does. ~~
Creating an \code{fgasp} class for a GaSP model with matern covariance. 
%The range and noise-variance ratio parameters are given and/or have been estimated.
}
\usage{
  fgasp(input, output, have_noise=TRUE, kernel_type='matern_5_2')
}
%- maybe also 'usage' for other objects documented here.
\arguments{
    \item{input}{a vector with dimension num_obs x 1 for the sorted input locations.}
    \item{output}{a  vector with dimension n x 1 for the observations at the sorted input locations.}

  \item{have_noise}{a bool value. If it is true, it means the model contains a noise. 
%%     ~~Describe \code{response} here~~
}

  \item{kernel_type}{a \code{character} to specify the type of kernel to use. The current version supports kernel_type to be "matern_5_2", "matern_3_2" or "exp", meaning that the matern kernel with roughness parameter being 2.5, 1.5, or 0.5 (exponent kernel), respectively. 
%%     ~~Describe \code{trend} here~~
}

}
%\details{
%%  ~~ If necessary, more details than the description above ~~
%expand here the details.
%}

\value{
%%  ~Describe the value returned
%%  If it is a LIST, use
%%  \item{comp1 }{Description of 'comp1'}
%%  \item{comp2 }{Description of 'comp2'}
%% ...
\code{fgasp} returns an S4 object of class \code{fgasp} (see \code{fgasp}).

%If there is an emulator (i.e. simul_type is 0), \code{rcalibration} returns a list with 
%\item{rcalibration}{an S4 object of class \code{rcalibration} (see \code{rcalibration-class}.}
%\item{emulator}{an S4 object of class \code{rgasp} produced by RobustGaSP R Package .}


}

\references{
%% ~put references to the literature/web site here ~

Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models},  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.

M. Gu, Y. Xu (2017), \emph{Nonseparable Gaussian stochastic process: a unified
view and computational strategy}, arXiv:1711.11501.

M. Gu, X. Wang and J.O. Berger (2018), \emph{Robust Gaussian Stochastic Process Emulation}, \emph{Annals of Statistics}, \bold{46}, 3038-3066.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}

%\note{
%%  ~~further notes~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~

%\seealso{
%% ~~objects to See Also as \code{\link{help}}, ~~~
%}

\examples{

library(FastGaSP)

#-------------------------------------
# Example 1: a simple example with noise 
#-------------------------------------

y_R<-function(x){
  cos(2*pi*x)
}

###let's test for 2000 observations
set.seed(1)
num_obs=2000
input=runif(num_obs)

output=y_R(input)+rnorm(num_obs,mean=0,sd=0.1)

##constucting the fgasp.model
fgasp.model=fgasp(input, output)
show(fgasp.model)

#------------------------------------------
# Example 2: a simple example with no noise 
#------------------------------------------

y_R<-function(x){
  sin(2*pi*x)
}


##generate some data without noise
num_obs=50
input=seq(0,1,1/(num_obs-1))

output=y_R(input)


##constucting the fgasp.model
fgasp.model=fgasp(input, output,have_noise=FALSE)

show(fgasp.model)

}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
%\keyword{ ~kwd1 }
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line

\name{get_consecutive_data}
\alias{get_consecutive_data}
\title{Extract consecutive time steps data from particle trajectories}
\description{
  Extracts paired data from consecutive time steps for a specified variable in particle trajectory data. This function is particularly useful for experimental data, where it uses particle tracking information to maintain particle identity between frames.
}
\usage{
get_consecutive_data(data_obj, variable = c("vx", "vy", "px", "py", "theta"))
}
\arguments{
  \item{data_obj}{An object of class \code{\linkS4class{particle.data}} containing particle trajectory data.}
  \item{variable}{A character string specifying which variable to extract. Must be one of "vx" (x-velocity), "vy" (y-velocity), "px" (x-position), "py" (y-position), or "theta" (angle).}
}
\value{
  Returns a list with two components:
  \item{start}{A list of length T_time containing the data at each time step t.}
  \item{end}{A list of length T_time containing the corresponding data at time step t+1.}
  
  For each time t, start[[t]] and end[[t]] contain paired measurements for the same particles at consecutive time steps.
}
\examples{
\dontrun{
vx_pairs = get_consecutive_data(data_obj, "vx")
vx_list = vx_pairs$start
vx_end_list = vx_pairs$end
}
}

\keyword{internal}

\name{predict.fmou}
\alias{predict.fmou}
\alias{predict.fmou,fmou-method}
\title{ 
Prediction and uncertainty quantification on the future observations using a FMOU model.}


\description{
This function predicts the future observations using a FMOU model. Uncertainty quantification is available. 
}


\usage{
\S4method{predict.fmou}{fmou}(object, param_est, step=1, interval=FALSE, interval_data=TRUE)
}


\arguments{
  \item{object}{an objecft of  class \code{fmou.}}
  \item{param_est}{estimated parameters in the FMOU model. Obtained by the results of \code{fit.fmou()}. }
  \item{step}{a vector. Number of steps to be predicted. Default is 1. } 
  \item{interval}{a bool value, default is \code{FALSE}. If \code{TRUE}, the 95\% predictive intervals are computed.}
  \item{interval_data}{a bool value, default is \code{TRUE}. If \code{TRUE}, the 95\% predictive intervals of the observations are computed. Otherwise, the 95\% predictive intervals of the mean of the observation are computed.} 
}

\value{
  \item{pred_mean}{the predictive mean. }
  \item{pred_interval_95lb}{the 95\% lower bound of the interval.}
  \item{pred_interval_95ub}{the 95\% upper bound of the interval.}
}

\references{

Lin, Y., Liu, X., Segall, P., & Gu, M. (2025). Fast data inversion for high-dimensional dynamical systems from noisy measurements. arXiv preprint arXiv:2501.01324.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}


\examples{

## generate simulated data
library(FastGaSP)
library(rstiefel)

d = 5  # number of latent factors
k = 20 # length of observation at each time step
n = 500 # number time step
noise_level = 1 # variance of noise

U = rustiefel(k, k) # factor loading matrix
z = matrix(NA, d, n)
sigma_2 = runif(d, 0.5, 1)
rho = runif(d, 0.95, 1)
for(l in 1:d){
  R = matrix(NA, n, n)
  diag(R) = 1
  for(ir in 1:n){
    for(ic in 1:n){
      R[ir, ic] = rho[l]^(abs(ir-ic)) * R[ir, ir]
    }
  }
  R = (sigma_2[l]/(1-rho[l]^2) )* R
  z[l, ] = t(chol(R)) \%*\% rnorm(n)
}

signal = U[,1:d] \%*\% z
y = signal + matrix(rnorm(n*k,mean=0,sd=sqrt(noise_level)),k,n)

##constucting the fmou.model
fmou.model=fmou(output=y, d=d, est_U0=TRUE, est_sigma0_2=TRUE)

## estimate the parameters
em_alg <- fit.fmou(fmou.model, M=500)

## two-step-ahead prediction
pred_2step <- predict.fmou(fmou.model,em_alg, step=c(1:2))

}

\name{gppca-class}
\docType{class}
\alias{gppca-class}
\title{GPPCA class}
\description{
An S4 class for generalized probabilistic principal component analysis of correlated data.
}

\section{Objects from the Class}{
Objects of this class are created and initialized using the \code{\link{gppca}} function to set up the estimation.
}

\section{Slots}{
	 \describe{
	  \item{\code{input}:}{object of class \code{vector}, the length is equivalent to the number of observations.}
    \item{\code{output}:}{object of class \code{matrix}. The observation matrix.}
    \item{\code{d}:}{object of class \code{integer} to specify the number of latent factors.}
    \item{\code{est_d}:}{object of class \code{logical}, default is \code{FALSE}. If \code{TRUE}, d will be estimated by either variance matching (when noise level is given) or information criteria (when noise level is unknown). Otherwise, d is fixed, and users must assign a value to \code{d}.}
    \item{\code{shared_params}:}{object of class \code{logical}, default is \code{TRUE}. If \code{TRUE}, the latent processes share the correlation and variance parameters. Otherwise, each latent process has distinct parameters.}
    \item{\code{kernel_type}:}{a \code{character} to specify the type of kernel to use. The current version supports kernel_type to be "matern_5_2" or "exponential", meaning that the matern kernel with roughness parameter being 2.5 or 0.5 (exponent kernel), respectively. }
    }
}

\section{Methods}{
  \describe{
    \item{fit.gppca}{See \code{\link[=fit.gppca]{fit.gppca}} for details.}
    \item{predict.gppca}{See \code{\link[=predict.gppca]{predict.gppca}} for details.}
	 }
}

\references{

Gu, M., & Shen, W. (2020), Generalized probabilistic principal component analysis of correlated data, \emph{Journal of Machine Learning Research, 21}(13), 1-41.

}


\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}

\seealso{ \code{\link{gppca}} for more details about how to create a \code{gppca} object.}

\name{predict}
\alias{predict}
\alias{predict.fgasp}
\alias{predict,fgasp-method}
\title{ 
Prediction and uncertainty quantification on the testing input using a GaSP model.
}
\description{
This function computes the predictive mean and variance on the given testing input using a GaSP model.
}
\usage{
\S4method{predict}{fgasp}(param,object, testing_input=NULL, var_data=TRUE, sigma_2=NULL,
                          predict_input=F,index_return=F)
}
\arguments{
  \item{param}{a vector of parameters. The first parameter is the natural logarithm of the inverse range parameter in the kernel function. If the data contain noise, the second parameter is the logarithm of the nugget-variance ratio parameter.
}
  \item{object}{an object of  class \code{fgasp}.
}
  \item{testing_input}{a vector of testing input for prediction.
}

  \item{var_data}{a bool value to decide whether the noise of the data is included for computing the posterior predictive variance.
}
   \item{sigma_2}{a numerical value specifying the variance of the kernel function. If given, the package uses this parameter for prediction.}
      \item{predict_input}{a bool value. If true, it makes predictions on inputs.}
   \item{index_return}{a bool value. If true, some additional indices relevant to sorting will be returned.}

}

\value{

The returned value is a S4 Class \code{predictobj.fgasp}. 

}

\references{
%% ~put references to the literature/web site here ~

Hartikainen, J. and Sarkka, S. (2010). \emph{Kalman filtering and smoothing solutions to temporal gaussian process regression models},  \emph{Machine Learning for Signal Processing (MLSP), 2010 IEEE International Workshop},  379-384.

M. Gu, Y. Xu (2017), \emph{Nonseparable Gaussian stochastic process: a unified
view and computational strategy}, arXiv:1711.11501.

M. Gu, X. Wang and J.O. Berger (2018), \emph{Robust Gaussian Stochastic Process Emulation}, \emph{Annals of Statistics}, \bold{46}, 3038-3066.

}

\author{
\packageAuthor{FastGaSP}

Maintainer: \packageMaintainer{FastGaSP}
}

%\note{
%%  ~~further notes~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~

%\seealso{
%% ~~objects to See Also as \code{\link{help}}, ~~~
%}

\examples{

library(FastGaSP)

#------------------------------------------------------------------------------
# Example 1: a simple example with noise to show fast computation algorithm 
#------------------------------------------------------------------------------

y_R<-function(x){
  cos(2*pi*x)
}

###let's test for 2000 observations
set.seed(1)
num_obs=2000
input=runif(num_obs)

output=y_R(input)+rnorm(num_obs,mean=0,sd=0.1)

##constucting the fgasp.model
fgasp.model=fgasp(input, output)

##range and noise-variance ratio (nugget) parameters 
param=c( log(1),log(.02))
## the log lik
log_lik(param,fgasp.model)
##time cost to compute the likelihood
time_cost=system.time(log_lik(param,fgasp.model))
time_cost[1]

##consider a nonparametric regression setting 
##estimate the parameter by maximum likelihood estimation

est_all<-optim(c(log(1),log(.02)),log_lik,object=fgasp.model,method="L-BFGS-B",
              control = list(fnscale=-1))  

##estimated log inverse range parameter and log nugget
est_all$par

##estimate variance 
est.var=Get_log_det_S2(est_all$par,fgasp.model@have_noise,fgasp.model@delta_x,
                          fgasp.model@output,fgasp.model@kernel_type)[[2]]/fgasp.model@num_obs
est.var

###1. Do some interpolation test 
num_test=5000
testing_input=runif(num_test) ##there are the input where you don't have observations
pred.model=predict(param=est_all$par,object=fgasp.model,testing_input=testing_input)

lb=pred.model@mean+qnorm(0.025)*sqrt(pred.model@var)
ub=pred.model@mean+qnorm(0.975)*sqrt(pred.model@var)

## calculate lb for the mean function
pred.model2=predict(param=est_all$par,object=fgasp.model,testing_input=testing_input,var_data=FALSE)
lb_mean_funct=pred.model2@mean+qnorm(0.025)*sqrt(pred.model2@var)
ub_mean_funct=pred.model2@mean+qnorm(0.975)*sqrt(pred.model2@var)

## plot the prediction
min_val=min(lb,output)
max_val=max(ub,output)

plot(pred.model@testing_input,pred.model@mean,type='l',col='blue',
     ylim=c(min_val,max_val),
     xlab='x',ylab='y')
polygon( c(pred.model@testing_input,rev(pred.model@testing_input)),
    c(lb,rev(ub)),col = "grey80", border = FALSE)
lines(pred.model@testing_input,pred.model@mean,type='l',col='blue')
lines(pred.model@testing_input,y_R(pred.model@testing_input),type='l',col='black')
lines(pred.model2@testing_input,lb_mean_funct,col='blue',lty=2)
lines(pred.model2@testing_input,ub_mean_funct,col='blue',lty=2)
lines(input,output,type='p',pch=16,col='black',cex=0.4) #one can plot data

legend("bottomleft", legend=c("predictive mean","95\% predictive interval","truth"),
       col=c("blue","blue","black"), lty=c(1,2,1), cex=.8)


##mean square error for all inputs
mean((pred.model@mean- y_R(pred.model@testing_input))^2)
##coverage for the mean
length(which(y_R(pred.model@testing_input)>lb_mean_funct &
               y_R(pred.model@testing_input)<ub_mean_funct))/pred.model@num_testing
##average length of the invertal for the mean
mean(abs(ub_mean_funct-lb_mean_funct))
##average length of the invertal for the data
mean(abs(ub-lb))

#---------------------------------------------------------------------------------
# Example 2: numerical comparison with the GaSP by inverting the covariance matrix
#---------------------------------------------------------------------------------
##matern correlation with smoothness parameter being 2.5
matern_5_2_kernel<-function(d,beta){  
  res=(1+sqrt(5)*beta*d + 5*beta^2*d^2/3 )*exp(-sqrt(5)*beta*d)
  res
}

##A function for computing the likelihood by the GaSP in a straightforward way
log_lik_GaSP_slow<-function(param,have_noise=TRUE,input,output){
  n=length(output)
  beta=exp(param[1])
  eta=0
  if(have_noise){
    eta=exp(param[2])
  }
  R00=abs(outer(input,input,'-'))
  R=matern_5_2_kernel(R00,beta=beta)
  R_tilde=R+eta*diag(n)
  #use Cholesky and one backsolver and one forward solver so it is more stable
  L=t(chol(R_tilde))
  output_t_R.inv= t(backsolve(t(L),forwardsolve(L,output )))
  S_2=output_t_R.inv\%*\%output
  
  -sum(log(diag(L)))-n/2*log(S_2)
}


pred_GaSP_slow<-function(param,have_noise=TRUE,input,output,testing_input){
  beta=exp(param[1])
  R00=abs(outer(input,input,'-'))
  eta=0
  if(have_noise){
    eta=exp(param[2])
  }
  R=matern_5_2_kernel(R00,beta=beta)
  R_tilde=R+eta*diag(length(output))
  
  ##I invert it here but one can still use cholesky to make it more stable
  R_tilde_inv=solve(R_tilde)

  r0=abs(outer(input,testing_input,'-'))
  r=matern_5_2_kernel(r0,beta=beta)

  S_2=t(output)\%*\%(R_tilde_inv\%*\%output)

  sigma_2_hat=as.numeric(S_2/num_obs)

  pred_mean=t(r)\%*\%(R_tilde_inv\%*\%output)
  pred_var=rep(0,length(testing_input))
  
  
  for(i in 1:length(testing_input)){
    pred_var[i]=-t(r[,i])\%*\%R_tilde_inv\%*\%r[,i]
  }
  pred_var=pred_var+1+eta
  list=list()
  list$mean=pred_mean
  list$var=pred_var*sigma_2_hat
  list
}

##let's test sin function
y_R<-function(x){
  sin(2*pi*x)
}


###let's test for 500 observations
set.seed(1)
num_obs=500
input=runif(num_obs)
output=y_R(input)+rnorm(num_obs,mean=0,sd=0.1)


##constucting the fgasp.model
fgasp.model=fgasp(input, output)

##range and noise-variance ratio (nugget) parameters 
param=c( log(1),log(.02))
## the log lik
log_lik(param,fgasp.model)
log_lik_GaSP_slow(param,have_noise=TRUE,input=input,output=output)

##time cost to compute the likelihood
##More number of inputs mean larger differences
time_cost=system.time(log_lik(param,fgasp.model))
time_cost

time_cost_slow=system.time(log_lik_GaSP_slow(param,have_noise=TRUE,input=input,output=output))
time_cost_slow

est_all<-optim(c(log(1),log(.02)),log_lik,object=fgasp.model,method="L-BFGS-B",
              control = list(fnscale=-1))  
##estimated log inverse range parameter and log nugget
est_all$par


##Do some interpolation test for comparison
num_test=200
testing_input=runif(num_test) ##there are the input where you don't have observations

##one may sort the testing_input or not, and the prediction will be on the sorted testing_input
testing_input=sort(testing_input)

## two ways of prediction
pred.model=predict(param=est_all$par,object=fgasp.model,testing_input=testing_input)
#pred_slow=pred_GaSP_slow(param=est_all$par,have_noise=TRUE,input,output,sort(testing_input))
## look at the difference
#max(abs(pred.model@mean-pred_slow$mean))
#max(abs(pred.model@var-pred_slow$var))


}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
%\keyword{ ~kwd1 }
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line
